# Какие бывают виртуальные функции и как их искать

В своем бестселлере «Фундаментальные основы хакерства», увидевшем свет более 15 лет назад, Крис Касперски поделился с читателями секретами дизассемблирования и исследования программ. Мы продолжаем публиковать отрывки из обновленного издания его книги. Сегодня мы поговорим о виртуальных функциях, их особенностях и о хитростях, которые помогут отыскать их в коде.

## Идентификация чисто виртуальных функций

Если функция объявляется в базовом, а реализуется в производном классе, она называется чисто виртуальной функцией, а класс, содержащий хотя бы одну такую функцию, --- абстрактным классом. Язык C++ запрещает создание экземпляров абстрактного класса, да и как они могут создаваться, если по крайней мере одна из функций класса не определена?

В стародавние времена компилятор в виртуальной таблице замещал вызов чисто виртуальной функции указателем на библиотечную функцию purecall, потому что на стадии компиляции программы он не мог гарантированно отловить все попытки вызова чисто виртуальных функций. И если такой вызов происходил, управление получала заранее подставленная сюда purecall, которая «ругалась» на запрет вызова чисто виртуальных функций и завершала работу приложения.

Однако в современных реалиях дело обстоит иначе. Компилятор отлавливает вызовы чисто виртуальных функций и банит их во время компиляции. Таким образом, он даже не создает таблицы виртуальных методов для абстрактных классов.

| ![Реализация вызова виртуальных функций](./images/08/call_vir_func.png)
|:--:|
| *Реализация вызова виртуальных функций* |

В этом нам поможет убедиться следующий пример (листинг примера PureCall):

```cpp
#include <stdio.h>

class Base {
public:
    virtual void demo(void) = 0;
};

class Derived :public Base {
public:
    virtual void demo(void) {
        printf("DERIVED\n");
    }
};
int main()
{
    Base *p = new Derived;
    p‐>demo();
    delete p; // Хотя статья не о том, как писать код на C++,
              // будем правильными до конца
}
```

Результат его компиляции в общем случае должен выглядеть так:

```asm
main proc near
 push rbx
 sub  rsp, 20h
 mov  ecx, 8 ; size
 ; Выделение памяти для нового экземпляра объекта
 call operator new(unsigned __int64)
 mov  rbx, rax
 lea  rax, const Derived::`vftable'
 mov  rcx, rbx ; this
 mov  [rbx], rax
 ; Вызов метода
 call cs:const Derived::`vftable'
 mov  edx, 8 ; __formal
 mov  rcx, rbx ; block
 ; Очищаем выделенную память, попросту удаляем объект
 call operator delete(void *,unsigned __int64)
 xor  eax, eax
 add  rsp, 20h
 pop  rbx
 retn
main endp
```

Чтобы узнать, какой метод вызывается инструкцией `call cs:const Derived::'vftable'`, надо сначала перейти в таблицу виртуальных методов класса **Derived**(нажав Enter):

```dump
const Derived::`vftable' dq offset Derived::demo(void)
```

а отсюда уже в сам метод:

```asm
public: virtual void Derived::demo(void) proc near
    lea rcx, _Format    ; "DERIVED\n"
    jmp printf
public: virtual void Derived::demo(void) endp
```

В дизассемблерном листинге для x86 IDA сразу подставляет правильное имя вызываемого метода:

```asm
call Derived::demo(void)
```

Это мы выяснили. И никакого намека на purecall.

Хочу также обратить твое внимание на следующую деталь. Старые компиляторы вставляли код проверки и обработки ошибок выделения памяти непосредственно после операции выделения памяти, тогда как современные компиляторы перенесли эту заботу внутрь оператора `new`:

```asm
void * operator new(unsigned __int64) proc near
  push rbx
  sub  rsp, 20h
  mov  rbx, rcx
  jmp  short loc_14000110E ; После пролога выполняется
                           ; безусловный переход
loc_1400010FF:
  mov  rcx, rbx
  call _callnewh_0         ; Вторая попытка выделения памяти
  test eax, eax
  jz   short loc_14000111E ; Если память снова не удалось выделить
                           ; переходим в конец функции, где вызываем функции
                           ; обработки ошибок
  mov  rcx, rbx
loc_14000110E:
  call malloc_0            ; Первая попытка выделения памяти
  test rax, rax            ; Проверка успешности выделения памяти
  jz   short loc_1400010FF ; Если rax == 0, значит, произошла
                           ; ошибка и память не
                           ; выделена, тогда совершаем переход
                           ; и делаем еще попытку
  add  rsp, 20h
  pop  rbx
  retn
loc_14000111E:
  cmp  rbx, 0FFFFFFFFFFFFFFFFh
  jz   short loc_14000112A
  call __scrt_throw_std_bad_alloc(void)
  align 2
loc_14000112A:
  call scrt_throw_std_bad_array_new_length(void)
  align 10h
void * operator new(unsigned __int64) endp
```

После пролога функции командой jmp short loc_14000110E выполняется безусловный переход на код для выделения памяти: call malloc_0. Проверяем результат операции: test rax, rax. Если выделение памяти провалилось, переходим на метку jz short loc_1400010FF, где еще раз пытаемся зарезервировать память:

```asm
mov rcx, rbx
call _callnewh_0
test eax, eax
```

Если эта попытка тоже проваливается, нам ничего не остается, как перейти по метке `jz short loc_14000111E`, обработать ошибки и вывести соответс‐ твующее ругательство.

## СОВМЕСТНОЕ ИСПОЛЬЗОВАНИЕ ВИРТУАЛЬНОЙ ТАБЛИЦЫ НЕСКОЛЬКИМИ ЭКЗЕМПЛЯРАМИ КЛАССА

Сколько бы экземпляров класса (другими словами, объектов) ни существовало, все они пользуются одной и той же виртуальной таблицей. Виртуальная таблица принадлежит самому классу, но не экземпляру (экземплярам) этого класса. Впрочем, из этого правила существуют исключения.

| ![Все экземпляры класса используют одну и ту же виртуальную таблицу](./images/08/virtual_table.png)
|:--:|
| *Все экземпляры класса используют одну и ту же виртуальную таблицу* |

Для демонстрации совместного использования одной копии виртуальной таблицы несколькими экземплярами класса рассмотрим следующий пример (листинг примера UsingVT):

```cpp
#include <stdio.h>
class Base {
public:
    virtual void demo()
    {
        printf("Base\n");
    }
};

class Derived : public Base{
public:
    virtual void demo()
    {
        printf("Derived\n");
    }
};

int main(){

    Base *obj1 = new Derived;
    Base *obj2 = new Derived;
    obj1‐ > demo();
    obj2‐ > demo();

    delete obj1;
    delete obj2;
}
```

Результат его компиляции в общем случае должен выглядеть так:

```asm
main proc near
  mov    [rsp+arg_0], rbx
  mov    [rsp+arg_8], rsi
  push   rdi
  sub    rsp, 20h
  mov    ecx, 8 ; size
  ; Выделяем память под первый экземпляр класса
  call   operator new(unsigned int64)
  ; В созданный объект копируем виртуальную таблицу класса Derived
  lea    rsi, const Derived::`vftable'
  mov    ecx, 8 ; size
  mov    rdi, rax
  ; RAX теперь указывает на первый экземпляр
  mov    [rax], rsi
  ; Выделяем память под второй экземпляр класса
  call   operator new(unsigned int64)
  ; В RDI — указатель на виртуальную таблицу класса Derived
  mov    rcx, rdi
  mov    rbx, rax
  ; В RSI находится первый объект
  mov    [rax], rsi
  ; Берем указатель на виртуальную таблицу методов
  mov    r8, [rdi]
  ; Для первого объекта, скопированного в RAX, вызываем метод
  ; по указателю в виртуальной таблице
  call   qword ptr [r8]
  ; В RBX — указатель на виртуальную таблицу класса Derived
  mov    r8, [rbx]
  mov    rcx, rbx
  ; Вызываем метод по указателю в этой же самой таблице
  call   qword ptr [r8]
  mov    edx, 8 ; formal
  mov    rcx, rdi ; block
  call   operator delete(void *,unsigned int64)
  mov    edx, 8 ; formal
  mov    rcx, rbx ; block
  call   operator delete(void *,unsigned int64)
  mov    rbx, [rsp+28h+arg_0]
  xor    eax, eax
  mov    rsi, [rsp+28h+arg_8]
  add    rsp, 20h
  pop    rdi
  retn
main endp
```

Виртуальная таблица класса Derived выглядит так:

```dump
const Derived::`vftable' dq offset Derived::demo(void), 0
```

Обрати внимание: виртуальная таблица одна на все экземпляры класса.

## КОПИИ ВИРТУАЛЬНЫХ ТАБЛИЦ

Окей, для успешной работы, понятное дело, вполне достаточно и одной вир‐ туальной таблицы, однако на практике приходится сталкиваться с тем, что исследуемый файл прямо‐таки кишит копиями этих виртуальных таблиц. Что же это за напасть такая, откуда она берется и как с ней бороться?

Если программа состоит из нескольких файлов, компилируемых в самос‐ тоятельные obj‐модули (а такой подход используется практически во всех мало‐мальски серьезных проектах), компилятор, очевидно, должен помес‐ тить в каждый obj свою собственную виртуальную таблицу для каждого используемого модулем класса. В самом деле, откуда компилятору знать о существовании других obj и наличии в них виртуальных таблиц?

Вот так и возникают никому не нужные дубли, отъедающие память и зат‐ рудняющие анализ. Правда, на этапе компоновки линкер может обнаружить копии и удалить их, да и сами компиляторы используют различные эвристи‐ ческие приемы для повышения эффективности генерируемого кода. Наибольшую популярность завоевал следующий алгоритм: виртуальная таб‐ лица помещается в тот модуль, в котором содержится реализация первой невстроенной невиртуальной функции класса.

Обычно каждый класс реализуется в одном модуле, и в большинстве слу‐ чаев такая эвристика срабатывает. Хуже, если класс состоит из одних вир‐ туальных или встраиваемых функций. В этом случае компилятор «ложится» и начинает запихивать виртуальные таблицы во все модули, где этот класс используется. Последняя надежда на удаление «мусорных» копий — линкер, но и он не панацея. Собственно, эти проблемы должны больше заботить раз‐ работчиков программы (если их волнует, сколько памяти занимает прог‐ рамма), для анализа лишние копии всего лишь досадная помеха, но отнюдь не непреодолимое препятствие!

## СВЯЗАННЫЙ СПИСОК

В большинстве случаев виртуальная таблица — это обыкновенный массив, но некоторые компиляторы представляют ее в виде связанного списка. Каж‐ дый элемент виртуальной таблицы содержит указатель на следующий эле‐ мент, а сами элементы не размещены вплотную друг к другу, а рассеяны по всему исполняемому файлу.

На практике подобное, однако, попадается крайне редко, поэтому не будем подробно на этом останавливаться — достаточно лишь знать, что такое бывает. Если ты встретишься со списками (впрочем, это вряд ли) — разберешься по обстоятельствам, благо это несложно.

## ВЫЗОВ ЧЕРЕЗ ШЛЮЗ

Будь также готов и к тому, чтобы встретить в виртуальной таблице указатель не на виртуальную функцию, а на код, который модифицирует этот указатель, занося в него смещение вызываемой функции. Этот прием был предложен самим разработчиком языка C++ Бьерном Страуструпом, позаимствовавшим его из ранних реализаций алгола‐60. В алголе код, корректирующий ука‐ затель вызываемой функции, называется шлюзом (thunk), а сам вызов — вызовом через шлюз. Вполне справедливо употреблять эту терминологию и по отношению к C++.

Однако в настоящее время вызов через шлюз чрезвычайно мало рас‐ пространен и не используется практически ни одним компилятором. Нес‐ мотря на то что он обеспечивает более компактное хранение виртуальных таблиц, модификация указателя приводит к излишним накладным расходам на процессорах с конвейерной архитектурой (а все современные процес‐ соры как раз и построены на основе такой архитектуры). Поэтому исполь‐ зование шлюзовых вызовов оправданно лишь в программах, критических к размеру, но не к скорости.

Подробнее обо всем этом можно прочесть в руководстве по алголу‐60 (шутка) или у Бьерна Страуструпа в «Дизайне и эволюции языка C++».

## СЛОЖНЫЙ ПРИМЕР НАСЛЕДОВАНИЯ

До сих пор мы рассматривали лишь простейшие примеры использования виртуальных функций. В жизни же порой встречается такое... Рассмотрим сложный случай наследования с конфликтом имен, демонстрирующий помещение невиртуальных функций в виртуальные таблицы (листинг примера HardSample):

```cpp
#include <stdio.h>

class A {
public:
    virtual void f()
    {

        printf("A_F\n");
    }
};

class B {
public:
    virtual void f()
    {

        printf("B_F\n");
    }

    virtual void g()
    {

        printf("B_G\n");
    }
};

class C : public A, public B {
public:
    void f()
    {

        printf("C_F\n");
    }
};

int main()
{
    A* a = new A;
    B* b = new B;
    C* c = new C;
    a‐> f();

    b‐> f();

    b‐> g();

    c‐> f();
    delete a;
    delete b;
    delete c;
}
```

Как будет выглядеть виртуальная таблица класса C? Так, давай подумаем: раз класс C — производный от классов A и B, то он наследует функции обоих, но виртуальная функция f() класса B перекрывает одноименную виртуаль‐ ную функцию класса A, поэтому из класса А она не наследуется. Далее, пос‐ кольку невиртуальная функция f() присутствует и в производном классе C, она перекрывает виртуальную функцию класса B (да, именно так, а вот невир‐ туальная функция невиртуальную не перекрывает, и она всегда вызывается из базового, а не производного класса). Таким образом, виртуальная таб‐ лица класса C должна содержать только один элемент — указатель на вир‐ туальную функцию g(), унаследованную от B, а невиртуальная функция f() вызывается как обычная C‐функция. Правильно? Нет!

Из‐за подобных сложностей, возникающих при использовании множес‐ твенного наследования классов, более современные по сравнению с C++ языки (например, такие как Java, C#, Swift) его не реализуют. Тем не менее в них допускается множественная реализация интерфейсов (или протоколов в Swift). А последние, как известно, не содержат реализации объявленных методов, поэтому таких неопределенностей не возникает. В C++ для избе‐ гания таких ситуаций используются абстрактные классы, но в реальности вся‐ кое приходится анализировать.

Здесь мы видим как раз тот случай, когда невиртуальная функция вызыва‐ ется через указатель как виртуальная функция. Более того, виртуальная таб‐ лица класса будет содержать не два, а три элемента! Третий элемент — это ссылка на виртуальную функцию f(), унаследованную от B, но тут же замещенная компилятором на «переходник» к C::f(). Уф... Как все непросто! Может, после изучения дизассемблерного листинга это станет понятнее?

```asm
main proc near
    mov [rsp+arg_0], rbx
    mov [rsp+arg_8], rsi
    push rdi
    sub rsp, 20h
    mov ecx, 8 ; size
```

Выделяем память для объекта A:

```asm
call operator new(unsigned int64)
mov rsi, rax
mov ecx, 8 ; size
```

Помещаем в объект A указатель на его виртуальную таблицу:

```asm
lea rax, const A::`vftable'
mov [rsi], rax
```

Выделяем память для объекта B:

```asm
call operator new(unsigned int64)
mov rdi, rax
mov ecx, 10h ; size
```

Помещаем в объект B указатель на его виртуальную таблицу:

```asm
lea rax, const B::`vftable'
mov [rdi], rax
```

Выделяем память для объекта C:

```asm
call operator new(unsigned int64)
mov rbx, rax
mov rcx, rsi
```

Помещаем в объект C указатель на виртуальную таблицу класса A:

```asm
lea rax, const C::`vftable'{for `A'}
mov [rbx], rax
```

Помещаем в объект C указатель на виртуальную таблицу класса B, то есть в результате объект C будет содержать два указателя на две виртуальные таблицы базовых классов.

```asm
lea rax, const C::`vftable'{for `B'}
mov [rbx+8], rax
```

Далее следует код, в котором для каждого созданного объекта в соответс‐ твии со смещением в виртуальной таблице происходит вызов метода. По дизассемблерному листингу можно разобраться, кто и что вызывает.

```asm
mov r8, [rsi]
call qword ptr [r8]
```

Первым делом надо определить, что находится в регистре RSI. Сделать это не составляет труда, надо лишь взглянуть в окно IDA и подвести курсор к обозначенному выше регистру. После чего он подсветится во всем лис‐ тинге (см. рисунок ниже).

Таким нехитрым образом определяем, что в регистре RSI находится объ‐ ект A со своей виртуальной таблицей методов. Копируем RSI в R8 и вызываем первый метод в таблице объекта A. Чтобы узнать, какой же это метод, надо перейти в виртуальную таблицу класса A, дважды щелкнув на `A::'vftable'`.

В итоге обнаружим

```dump
const A::`vftable' dq offset A::f(void), 0
```

То есть первый и единственный метод **f**. Подведем к нему курсор ввода, нажмем Enter, чтобы перейти в сам метод:

```asm
public: virtual void A::f(void) proc near
    lea rcx, _Format ; "A_F\n"
    jmp printf
public: virtual void A::f(void) endp
```

Он, собственно, выводит строку A_F, что и требовалось доказать.

## ПЕРЕИМЕНОВАНИЕ ИНСТРУКЦИЙ

Мы рассмотрели только начало анализа вызываемых методов из функции main. Вернемся в ее дебри.

Смотрим, что хранится в регистре RDI. Поднимаем глазки чуть выше, видим, что в RDI находится объект B. Значит, в следующем фрагменте кода происходит вызов двух его методов: первый — без смещения в виртуальной таблице `call qword ptr [r8]`, а второй — со смещением — `call qword ptr [r8+8]`.

```asm
mov r8, [rdi]
mov rcx, rdi
call qword ptr [r8]
mov r8, [rdi]
mov rcx, rdi
call qword ptr [r8+8]
```

> ### На заметку
>
> Если строчка `call qword ptr [r8]` портит картину и охота привести листинг к более ясному виду, заменив неявный вызов фактическим названием метода, поступи следующим образом: подведи курсор мыши к строке `call qword ptr [r8]`, вызови контекстное меню, в нем выбери пункт Manual... и в строке ввода появившегося окна вбей удобное название, которое хочешь присвоить, например `A::f`, как в нашем случае. Жми ОК. Вызов примет жела‐ емый вид. Чтобы проделать обратную операцию, то есть вернуть прежнюю абракадабру, достаточно таким же образом вызвать диалоговое окно Manual operand и в нем очистить поле ввода.

Смотрим дальше. В регистре RBX находится объект C и происходит вызов его метода:

```asm
mov r8, [rbx]
mov rcx, rbx
call qword ptr [r8]
```

Удаление объектов:

```asm
...
  mov edx, 8 ; formal
  mov rcx, rsi ; block
  call operator delete(void *,unsigned int64)
  mov edx, 8 ; formal
  mov rcx, rdi ; block
  call operator delete(void *,unsigned int64)
  mov edx, 10h ; formal
  mov rcx, rbx ; block
  call operator delete(void *,unsigned int64)
  mov rbx, [rsp+28h+arg_0]
  xor eax, eax
  mov rsi, [rsp+28h+arg_8]
  add rsp, 20h
  pop rdi
  retn
main endp
```

| ![Дизассемблированный листинг из IDA Pro, с подсвеченным регистром RSI](./images/08/ida_list_rsi.png)
|:--:|
| *Дизассемблированный листинг из IDA Pro, с подсвеченным регистром RSI* |

Рассмотрим виртуальные таблицы классов B и C:

```asm
const B::`vftable' dq
offset B::f(void), ; 1‐й элемент
offset B::g(void), ; 2‐й элемент
offset const A::`RTTI Complete Object Locator'
```

Так как класс B включает два виртуальных метода, его таблица виртуальных методов содержит два элемента: `B::f` и `B::g`. Виртуальная таблица класса C:

```asm
dq offset const C::`RTTI Complete Object Locator'{for `B'}
const C::`vftable'{for `B'} dq
offset [thunk]:C::f`adjustor{8}' (void), ; 1‐й элемент
offset B::g(void) dq                     ; 2‐й элемент
offset const C::`RTTI Complete Object Locator'{for `A'}
const C::`vftable'{for `A'} dq
offset C::f(void),                       ; 3‐й элемент
offset const B::`RTTI Complete Object Locator'
```

Как мы и предполагали, таблица содержит три элемента! Виртуальная таб‐ лица класса C, скопированная компилятором из класса B, первоначально состояла из двух указателей на функции `f()` и `g()`, но еще на стадии ком‐ пиляции компилятор разобрался в конфликте имен и заменил указатель на `B::f()` указателем на переходник к `C::f()`. При этом он распознал вызов через шлюз, а IDA отметила это: `[thunk]`, следом за которым идет ключевое слово adjustor (корректор).

Напоследок посмотрим на сами методы. У класса B они выглядят вполне ожидаемо:

```asm
public: virtual void B::f(void) proc near
    lea rcx, aBF ; "B_F\n"
    jmp printf
public: virtual void B::f(void) endp

public: virtual void B::g(void) proc near
    lea rcx, aBG ; "B_G\n"
    jmp printf
public: virtual void B::g(void) endp
```

У класса C вызов `B::g(void)`, как и положено, восходит к методу класса B (см. выше). Вызов `C::f(void)` обращается к невиртуальному методу:

```asm
public: virtual void C::f(void) proc near
    lea rcx, aCF ; "C_F\n"
    jmp printf
public: virtual void C::f(void) endp
```

В свою очередь, `C::f'adjustor{8}'` вызывает «сквозной» метод, передающий управление на метод `C::f(void)`.

```dump
[thunk]:public: virtual void C::f`adjustor{8}' (void) proc near
    sub rcx, 8
    jmp C::f(void) ; Передаем управление методу C::f(void)
[thunk]:public: virtual void C::f`adjustor{8}' (void) endp
```

Как видим, таблица виртуальных методов вполне может содержать указатель на невиртуальный метод!

На самом деле виртуальная таблица производного класса включает вир‐ туальные таблицы всех базовых классов (во всяком случае всех, откуда она наследует виртуальные методы). В нашем примере виртуальная таблица класса C содержит указатель на собственный невиртуальный метод и вир‐ туальную таблицу класса B. Задача: как определить, что функция C::f() невиртуальная? И как найти все базовые классы класса C?

Начнем с последнего. Современные версии компилятора и дизассембле‐ ра, идя рука об руку, явно показывают в виртуальной таблице и в коммента‐ рии к реализации метода, от какого класса он наследуется:

```dump
; DATA XREF: .rdata:const C::`vftable'{for `A'}?o
```

Следовательно, класс C — производный от A. С другой стороны, а что, если класс не наследует ни одной функции из базового класса? Тогда возникает другой вопрос: так ли необходимо устанавливать «родителей», от которых не наследуется ни одной функции? (Если хоть одна функция наследуется, никаких сложностей в поиске, как мы видели, не возникает.) В общем‐то, для анализа это действительно некритично, но чем точнее будет восстанов‐ лен исходный код программы, тем нагляднее он будет и тем легче в нем разобраться.

Теперь перейдем к невиртуальной функции f(). Подумаем, что было бы, будь она на самом деле виртуальной. Тогда она перекрыла бы одноименную функцию базовых классов и никакой «дикости» наподобие «переходников» в откомпилированной программе не встретилось бы. А так они свидетель‐ ствуют о том, что тут не все гладко и функция невиртуальная, хоть и стремится казаться такой. Опять‐таки умный компилятор теоретически может выкинуть переходник и дублирующийся элемент виртуальной таблицы класса C, но на практике этой интеллектуальности не наблюдается...

| ![Вывод консольного приложения HardSample](./images/08/hard_sample.png)
|:--:|
| *Вывод консольного приложения HardSample* |

## СТАТИЧЕСКИЕ ОБЪЕКТЫ

Есть ли разница, как создавать экземпляр класса — `MyClass zzz;` или `MyClass *zzz = new MyClass?` Разумеется, в первом случае компилятор может определить адреса виртуальных функций еще на стадии компиляции, тогда как во втором это приходится вычислять во время выполнения прог‐ раммы. Другое различие: статические объекты размещаются в стеке (сег‐ менте данных, первый случай), а динамические — в куче (второй случай).

Старые компиляторы упорно создавали таблицы виртуальных функций в обоих случаях, а при вызове каждой функции (включая невиртуальные) под‐ готавливали указатель `this` (как правило, помещаемый в один из регистров общего назначения), который содержит адрес экземпляра класса. Таким образом, если мы встречали вызываемую непосредственно по ее смещению функцию, которая в то же время присутствует в виртуальной таблице класса, можно было с уверенностью утверждать, что это виртуальная функция статич‐ ного экземпляра класса. Между тем современные компиляторы генерируют абсолютно отличный код. Чтобы увидеть это, рассмотрим следующий пример вызова функции статического объекта (листинг примера StaticObjects):

```cpp
#include <stdio.h>

class Base {
public:
    virtual void demo(void)
    {
        printf("BASE DEMO\n");
    }

    virtual void demo_2(void)
    {
        printf("BASE DEMO 2\n");
    }

    void demo_3(void)
    {
        printf("Non virtual BASE DEMO 3\n");
    }
};

class Derived : public Base {
public:
    virtual void demo(void)
    {
        printf("DERIVED DEMO\n");
    };

    virtual void demo_2(void)
    {
        printf("DERIVED DEMO 2\n");
    }

    void demo_3(void)
    {
        printf("Non virtual DERIVED DEMO 3\n");
    }
};

int main()
{
    Base p;
    p.demo();

    p.demo_2();

    p.demo_3();

    Derived d;
    d.demo();

    d.demo_2();

    d.demo_3();
}
```

| ![Вывод приложения StaticObjects](./images/08/static_object.png)
|:--:|
| *Вывод приложения StaticObjects* |

В результате компиляции мы увидим такой дизассемблерный листинг:

```asm
main proc near
    sub rsp, 28h
    lea rcx, _Format ; "BASE DEMO\n"
    call printf
    lea rcx, aBaseDemo2 ; "BASE DEMO 2\n"
    call printf
    lea rcx, aNonVirtualBase ; "Non virtual BASE DEMO 3\n"
    call printf
    lea rcx, aDerivedDemo ; "DERIVED DEMO\n"
    call printf
    lea rcx, aDerivedDemo2 ; "DERIVED DEMO 2\n"
    call printf
    lea rcx, aNonVirtualDeri ; "Non virtual DERIVED DEMO 3\n"
    call printf
    xor eax, eax
    add rsp, 28h
    retn
main endp
```

Вот так новость! Компилятор оптимизировал весь код в одну функцию! Выкинув при этом объекты, их методы, виртуальные таблицы. Программа берет из секции данных строковую константу, помещает ее в регистр RCX, а затем вызывает функцию `printf`. Потом это действие повторяется еще пять раз, и на этом работа приложения завершается. Ни малейшего упоминания о виртуальных, невиртуальных функциях нет!

## ИДЕНТИФИКАЦИЯ ВИРТУАЛЬНЫХ ТАБЛИЦ

Теперь, основательно освоившись с виртуальными таблицами и функциями, рассмотрим очень коварный вопрос: всякий ли массив указателей на фун‐ кции есть виртуальная таблица? Разумеется, нет! Ведь косвенный вызов фун‐ кции через указатель — частое дело в практике программиста. Массив ука‐ зателей на функции... хм, конечно, типичным его не назовешь, но и такое в жизни встречается.

Рассмотрим следующий пример — кривой и наигранный, конечно, — но, чтобы продемонстрировать ситуацию, где массив указателей жизненно необ‐ ходим, пришлось бы написать не одну сотню строк кода (листинг примера VT):

```cpp
#include <stdio.h>

void demo_1(void)
{
    printf("Demo 1\n");
}

void demo_2(void)
{
    printf("Demo 2\n");
}

void call_demo(void** x)
{
    ((void (*)(void))x[0])();
    ((void (*)(void))x[1])();
}

int main()
{
    static void* x[2] =
    {

        (void*)demo_1, (void*)demo_2

    };

    call_demo(&x[0]);
}

```

| ![Вывод приложения VT](./images/08/vt.png)
|:--:|
| *Вывод приложения VT* |

А теперь посмотрим, сможем ли мы отличить рукотворную таблицу указателей от настоящей виртуальной таблицы:

```asm
main proc near
    sub rsp, 28h
    call cs:x
    call cs:off_140003040
    xor eax, eax
    add rsp, 28h retn
main endp
```

Первое, что бросается в глаза, — это нетипичный для виртуальной функции вызов: `call cs:x`. Как мы знаем, перед вызовом виртуальной функции происходит копирование виртуальной таблицы класса ее объекта, затем в дело вмешивается указатель `this`. И только после этого вызывается виртуальная функция по смещению от начала таблицы виртуальных функций. Здесь же вызов происходит по смещению в сегменте кода. Второй вызов: `call cs: off_140003040` — аналогичен. То есть этот код слишком тривиален, тогда как вызов виртуальной функции сопряжен с большой избыточностью.

Поставив курсор на строку `call cs:x`, нажмем Enter для перехода в виртуальную таблицу, которая в действительности ей не является и только на беглый, нетренированный взгляд на нее похожа:

```dump
x             dq offset demo_1(void)
off_140003040 dq offset demo_2(void)
```

На самом же деле это массив указателей, и он лишь с первого впечатления смахивает на виртуальную таблицу. В глаза сразу бросается отсутствие име‐ ни класса и ключевого слова `vftable`.

Перейдем в тело функции `demo_1`:

```asm
void demo_1(void) proc near
    lea rcx, _Format ; "Demo 1\n"
    jmp printf
void demo_1(void) endp

void demo_2(void) proc near
    lea rcx, aDemo2 ; "Demo 2\n"
    jmp printf
void demo_2(void) endp
```

Здесь мы видим отсутствие модификатора доступа, ключевого слова virtual и имени класса. Все это подтверждает невиртуальные корни функции.

## ВТОРОЙ ВАРИАНТ ИНИЦИАЛИЗАЦИИ МАССИВА

Мы можем инициализировать массив не при его объявлении, а по ходу прог‐ раммы. Это можно сделать, заменив конструкцию

```cpp
static void* x[2] =
{
    (void*)demo_1,
    (void*)demo_2
};
```

вот таким вариантом:

```cpp
static void* x[2];
x[0] = (void*)demo_1;
x[1] = (void*)demo_2;
```

В этом случае компилятор сгенерирует адекватный код, но смещения функций будут заноситься в регистры прямо перед вызовом непосредственно в функции `main`.

```asm
main proc near
    sub rsp, 28h
    lea rax, demo_1(void)
    mov cs:x, rax
    lea rax, demo_2(void)
    mov cs:qword_140003638,
    rax call demo_1(void)
    call cs:qword_140003638
    xor eax, eax
    add rsp, 28h retn
main endp
```

В результате на месте «виртуальной таблицы» будет совсем непохожий на нее код:

```asm
x               dq ?
qword_140003638 dq ?
```

К тому же инициализация при объявлении (как в первом случае) помещает уже готовые указатели в сегмент данных, и это смахивает на настоящую виртуальную таблицу (и экономит такты процессора).

Обобщая выводы, повторим основные признаки подделки еще раз:

* слишком тривиальный код, минимум используемых регистров и никакой избыточности, обращение к виртуальным таблицам происходит куда витиеватее;
* указатель на виртуальную функцию заносится в экземпляр объекта;
* отсутствует указатель this, всегда подготавливаемый перед вызовом виртуальной функции;
* виртуальные функции и статические переменные располагаются в раз‐ личных местах сегмента данных, поэтому сразу можно отличить одни от других.

А можно ли так организовать вызов функции по ссылке, чтобы компиляция программы давала код, идентичный вызову виртуальной функции? Как сказать... Теоретически да, но практически едва ли такое удастся сделать (а уж непреднамеренно — тем более). Код вызова виртуальных функций в связи с большой избыточностью очень специфичен и легко различим на глаз. Нетрудно сымитировать общую технику работы с виртуальными таблицами, но без ассемблерных вставок невозможно воспроизвести ее в точности.

## ЗАКЛЮЧЕНИЕ

Вообще, как мы видим, работа с виртуальными функциями сопряжена с огромной избыточностью и «тормозами», а их анализ связан с большими трудозатратами — приходится постоянно держать в голове множество указателей и помнить, какой из них на что указывает. Но как бы там ни было, никаких принципиально неразрешимых преград перед исследователем не стоит.
