# Учимся идентифицировать конструкторы и деструкторы

В этой статье речь пойдет об известных каждому программисту функциях работы с объектами класса — конструкторах и деструкторах, а также о том, как обнаружить их в коде исследуемой программы.

## ОБЪЕКТЫ В КУЧЕ

Конструктор в силу своего автоматического вызова при создании нового экземпляра объекта — первая по счету функция. Так какие же могут возникнуть сложности в его идентификации? Камень преткновения в том, что конструктор факультативен, то есть может присутствовать в объекте, а может и не присутствовать. Поэтому совсем не факт, что первая вызываемая функция — именно конструктор!

Заглянув в описание языка C++, можно обнаружить, что конструктор не возвращает никакого значения, что нехарактерно для обычных функций, однако все же не настолько редко встречается, чтобы однозначно его идентифицировать. Как же тогда быть?

Выручает то обстоятельство, что по стандарту конструктор не должен автоматически вызывать исключения, даже если выделить память под объект не удалось. Реализовать это требование можно множеством различных способов, но все знакомые автору компиляторы просто помещают перед вызовом конструктора проверку на нулевой указатель, передавая ему управление только при удачном выделении памяти для объекта. Напротив, все остальные функции объекта вызываются всегда, даже при неуспешном выделении памяти. Точнее, пытаются вызываться, но нулевой указатель (возвращаемый в случае ошибки выделения памяти) при первой же попытке обращения вызывает исключение, передавая бразды правления обработчику соответствующей исключительной ситуации.

Таким образом, функция, окольцованная проверкой нулевого указателя, и есть конструктор, а не что‐либо иное. Теоретически, впрочем, подобная проверка может присутствовать и при вызове других функций, конструктором не являющихся, но... Во всяком случае, автору на практике с таким еще не приходилось встречаться.

Деструктор, как и конструктор, факультативен, то есть последняя вызываемая функция объекта может и не быть деструктором. Тем не менее отличить деструктор от любой другой функции очень просто — он вызывается только при результативном создании объекта (то есть успешном выделении памяти) и игнорируется в противном случае. Это документированное свойство языка — следовательно, обязательное к реализации всеми компиляторами. Таким образом, в код помещается такое же «кольцо», как и у конструктора, но никакой путаницы не возникает, так как конструктор вызывается всегда первым (если он есть), а деструктор — последним (опять‐таки если он присутствует).

Особый случай представляет объект, целиком состоящий из одного конструктора (или деструктора), — попробуй‐ка разобраться, с чем мы имеем дело. Но разобраться можно! За вызовом конструктора практически всегда присутствует код, обнуляющий `this` в случае неудачного выделения памяти, а у деструктора этого нет. Далее, деструктор обычно вызывается не непосредственно из материнской процедуры, а из функции‐обертки, вызывающей помимо деструктора и оператор `delete`, который освобождает занятую объектом память. Так что отличить конструктор от деструктора вполне можно.

Между тем современные компиляторы для создания оптимального по быстродействию и размеру кода могут впихнуть все элементы в функцию main. Для лучшего понимания сказанного рассмотрим демонстрацию конструктора и деструктора (пример myClass1):

```cpp
#include <stdio.h>

class MyClass {

public:
    MyClass(void);
    void demo(void);

    ~MyClass(void);
};

MyClass::MyClass()
{
    printf("Constructor\n");
}

MyClass::~MyClass()
{
    printf("Destructor\n");
}

void MyClass::demo(void)
{
    printf("MyClass\n");
}

int main()
{
    MyClass* zzz = new MyClass();
    zzz‐> demo();

    delete zzz;
}
```

| ![Вывод приложения myClass1](./images/09/my_class_1.png)
|:--:|
| *Вывод приложения myClass1* |

В результате компиляции примера myClass1 с помощью VC++’17 с настройкой вывода проекта Release будет выглядеть так:

```asm
main proc near
  push rbx
  sub  rsp, 20h
  mov  ecx, 1           ; size
  call operator new(unsigned int64)
  lea  rcx, _Format     ; "Constructor\n"
  mov  [rsp+28h+arg_0], rax
  mov  rbx, rax call printf
  lea  rcx, aMyclass    ; "MyClass\n"
  call printf
  lea  rcx, aDestructor ; "Destructor\n"
  call printf
  mov  edx, 1           ; formal
  mov  rcx, rbx ; block
  call operator delete(void *,unsigned int64)
  xor  eax, eax
  add  rsp, 20h pop rbx
  retn
main endp
```

Как раз упомянутый случай! Компилятор все сделал достоянием функции `main`, в отдельности не создав ни конструктора, ни деструктора, ни метода класса. После выделения памяти он тут же вызывает функцию для вывода строки символов! То же происходит с очисткой памяти. Вот это магия! Здесь интересным для нас может быть только оператор `new`: его устройство мы подробно рассмотрели в предыдущей статье.

С другой стороны, нам же хочется увидеть, как будет сгенерирован код без оптимизации. Скомпилируем проект в режиме Debug и посмотрим на дизассемблерный листинг в Ida Pro:

Так выглядит функция `main`:

```asm
push  rbp
push  rdi
sub   rsp, 1A8h
lea   rbp, [rsp+20h]
mov   rdi, rsp
mov   ecx, 6Ah
mov   eax, 0CCCCCCCCh
rep   stosd
mov   [rbp+190h+var_28],     0FFFFFFFFFFFFFFFEh
lea   rcx, unk_140023002
call  sub_1400110AA
lea   rcx, aMainDebug            ; "Main_Debug\n"
call  sub_140011230              ; вызов printf
mov   ecx, 1
call  sub_14001119F
mov   [rbp+190h+var_88],     rax
cmp   [rbp+190h+var_88],     0
jz    short                  loc_140011BE1
mov   rcx, [rbp+190h+var_88]
call  sub_140011415                        ; вызов конструктора для создания
                                           ; экземпляра класса MyClass
mov   [rbp+190h+var_18],     rax           ; копируем указатель на выделенную
jmp   short                  loc_140011BEC ; память и делаем безусловный переход
loc_140011BE1:
mov   [rbp+190h+var_18],     0 loc_140011BEC:
mov   rax, [rbp+190h+var_18]               ; возвращаем указатель на созданный
               ; объект в регистр rax
mov   [rbp+190h+var_A8],     rax
mov   rax, [rbp+190h+var_A8]
mov   [rbp+190h+var_188],    rax
mov   rcx, [rbp+190h+var_188]
call  sub_140011393                        ; для экземпляра класса MyClass
               ; вызываем метод demo
mov   rax, [rbp+190h+var_188]
mov   [rbp+190h+var_48],     rax
mov   rax, [rbp+190h+var_48]
mov   [rbp+190h+var_68],     rax
cmp   [rbp+190h+var_68],     0
jz    short                  loc_140011C4B mov edx, 1
mov   rcx, [rbp+190h+var_68]
call  sub_1400111EA                        ; вызов деструктора
mov   [rbp+190h+var_18],     rax
jmp   short                  loc_140011C56 loc_140011C4B:
mov   [rbp+190h+var_18],     0 loc_140011C56:
xor   eax, eax
lea   rsp, [rbp+188h]
pop   rdi
pop   rbp  retn
```

Вызовы конструктора, деструктора, функции `printf`, а так же метода `MyClass::demo()` осуществляются опосредованно, то есть через вызовы промежуточных функций. В результате стечения всех мыслимых обстоятельств на моем компьютере компилятор сгенерировал такие цепочки вызовов:

```dump
sub_140011415 ‐> sub_1400118A0 ‐ constructor
sub_1400111EA ‐> sub_140011980 ‐> sub_14001139D ‐> sub_140011910 ‐ destructor
sub_140011230 ‐> sub_140011CA0 ‐ printf
sub_140011393 ‐> sub_140011A10 ‐ MyClass::demo()
```

У тебя результаты вполне могут быть другими.

Вывод о том, что далее показано тело конструктора класса MyClass, можно сделать путем анализа реализации и его вызова:

```asm
mov  [rsp‐8+arg_0], rcx
push rbp
push rdi
sub  rsp, 0E8h
lea  rbp, [rsp+20h]
mov  rdi, rsp
mov  ecx, 3Ah
mov  eax, 0CCCCCCCCh
rep  stosd
mov  rcx, [rsp+0F0h+arg_8]
lea  rcx, unk_140023002
call sub_1400110AA
lea  rcx, aConstructor ; "Constructor\n"
call sub_140011230 ; вызов printf
mov  rax, [rbp+0D0h+arg_0]
lea  rsp, [rbp+0C8h]
pop  rdi
pop  rbp
retn
```

Сходным образом определяем деструктор:

```asm
mov  [rsp‐8+arg_0], rcx
push rbp
push rdi
sub  rsp, 108h
lea  rbp, [rsp+20h]
mov  rdi, rsp
mov  ecx, 42h
mov  eax, 0CCCCCCCCh
rep  stosd
mov  rcx, [rsp+110h+arg_8]
mov  [rbp+0F0h+var_28], 0FFFFFFFFFFFFFFFEh
lea  rcx, unk_140023002
call sub_1400110AA nop
lea  rcx, aDestructor ; "Destructor\n"
call sub_140011230 ; вызов printf nop
lea  rsp, [rbp+0E8h] pop rdi
pop  rbp
retn
```

И, наконец, метод MyClass::demo():

```asm
mov  [rsp‐8+arg_0], rcx
push rbp
push rdi
sub  rsp, 0E8h
lea  rbp, [rsp+20h]
mov  rdi, rsp
mov  ecx, 3Ah
mov  eax, 0CCCCCCCCh
rep  stosd
mov  rcx, [rsp+0F0h+arg_8]
lea  rcx, unk_140023002
call sub_1400110AA
lea  rcx, aMyclass ; "MyClass\n"
call sub_140011230 ; вызов printf
lea  rsp, [rbp+0C8h]
pop  rdi
pop  rbp
retn
```

Уверен, здесь все понятно без дополнительных комментариев.

### Современные микропроцессоры архитектуры x86-64

Наверняка в предыдущем листинге ты обратил внимание на ассемблерную инструкцию `rep stosd`. Здесь выполняется цикл (команда `rep`), при этом количество повторений предварительно помещено в регистр `ECX` или `RCX` (в зависимости от размера). После каждого прохода счетчик цикла уменьшается на единицу, и это происходит до тех пор, пока значение не станет нулевым. Команда `stosd` (в режиме x64) записывает по адресу, сохраненному в регистре `RDI` или `EDI`, двойное слово, находящееся в регистре `EAX`. Таким образом происходит заполнение целого блока памяти. Даже на языке высокого уровня подобная операция займет не одну строку:

#### Псевдокод

```dump
IF (Doubleword store)
  WHILE ((R|E)CX > 0)
  DO
    THEN
      DEST := EAX;
      THEN
        IF DF = 0
          THEN (R|E)DI := (R|E)DI + 4;
          ELSE (R|E)DI := (R|E)DI – 4;
        FI;
    (R|E)CX := (R|E)CX ‐ 1;
  OD;
FI;
```

Тем не менее на ассемблере операция умещается в одну строчку! Интересный момент. Давай покопаемся в истории ЭВМ.

По мере своего развития, начиная с 1978 года, архитектура x86 соответствовала (и продолжает соответствовать для совместимости с предыдущими моделями микропроцессоров) стандарту CISC — что расшифровывается как complex instruction set computer, или «компьютер со сложным набором команд». В первые годы оптимизирующих компиляторов еще не было. Программисты писали программы на ассемблере. Инженеры из Intel в соответствии с прогрессом своих процессоров расширяли низкоуровневый язык сборки. От такого подхода в ассемблере для x86, как в снежном коме, накапливались разные полезные и бесполезные команды.

До поры до времени компанию Intel не беспокоило происходящее параллельно зарождение и развитие процессорной архитектуры RISC (компьютер с уменьшенным набором команд). В соответствии с названием в ней использовался только короткий набор высокооптимизированных операций. Со временем стало очевидно, что RISC‐архитектура позволяет конструировать более производительные процессоры, нежели CISC. Тем не менее на Intel лежал огромный груз ответственности перед клиентами за совместимость с предыдущими моделями процессоров и софтом, написанным за все эти годы. Тогда корпорация выкатила архитектуру P6, главная особенность которой состояла в динамическом преобразовании CISC‐команд в RISC‐микрооперации. А также их анализ, переупорядочивание для возможности параллельного исполнения независящих от результата команд за один такт и выполнения микрокоманд на RISC‐ядре. На основе архитектуры P6 Intel в 1995 году представила процессор Pentium Pro. Тем самым он стал первым x86‐процессором с RISC‐ядром. На этой же архитектуре впоследствии были построены Pentium II, Pentium III, Celeron. С тех пор в рассматриваемом направлении микропроцессоры почти не изменились.

| ![Микроархитектура ядер современных процессоров x86-64](./images/09/risk.png)
|:--:|
| *Микроархитектура ядер современных процессоров x86-64* |

## ОБЪЕКТЫ В АВТОМАТИЧЕСКОЙ ПАМЯТИ

Если объект размещается в стеке (автоматической памяти), то никаких проверок успешности ее выделения не выполняется и вызов конструктора становится неотличим от вызова остальных функций. Аналогичная ситуация и с деструктором — стековая память автоматически освобождается по завершении функции, а вместе с ней умирает и сам объект безо всякого вызова delete (при этом delete применяется только для удаления объектов из кучи). Посмотрим, появятся ли в дизассемблированном листинге отличия при создании объекта в стеке.

Чтобы увидеть это, модифицируем функцию main нашего предыдущего примера следующим образом (пример myClass2):

```cpp
int main()
{
    printf("Main\n"); // для наглядной идентификации
    MyClass zzz;
    zzz.demo();
}
```

Результат компиляции этого кода в режиме отладки должен выглядеть так:

```asm
push  rbp
push  rdi
sub   rsp,  128h
lea   rbp,  [rsp+20h]
mov   rdi,  rsp
mov   ecx,  4Ah
mov   eax,  0CCCCCCCCh
rep   stosd
mov   [rbp+110h+var_28],           0FFFFFFFFFFFFFFFEh
mov   rax,  cs: __security_cookie
xor   rax,  rbp
mov   [rbp+110h+var_18],           rax
lea   rcx,  unk_140021002
call  sub_140011082
lea   rcx,  aMain                 ; "Main\n"
call  sub_1400111D6               ; вызов printf
lea   rcx,  [rbp+110h+var_10C]
call  sub_14001137A               ; вызов конструктора nop
lea   rcx,  [rbp+110h+var_10C]
call  sub_140011316               ; вызов метода demo nop
lea   rcx,  [rbp+110h+var_10C]
call  sub_14001131B               ; вызов деструктора xor eax, eax
mov   edi,  eax
lea   rcx,  [rbp+110h+var_130]
lea   rdx,  unk_140019CB0
call  sub_140011352
mov eax, edi
mov   rcx,  [rbp+110h+var_18]
xor   rcx,  rbp                   ; StackCookie
call  j___security_check_cookie
lea   rsp,  [rbp+108h]
pop   rdi
pop   rbp
retn
```

Заметных отличий в сравнении с предыдущим примером не наблюдается, и все функции по‐прежнему можно определить путем кропотливого анализа их содержимого.

## ИДЕНТИФИКАЦИЯ КОНСТРУКТОРА/ДЕСТРУКТОРА В ГЛОБАЛЬНЫХ ОБЪЕКТАХ

Глобальные объекты (также называемые статическими объектами) размещаются в сегменте данных еще на стадии компиляции. Стало быть, ошибки выделения памяти в принципе невозможны. Выходит, что по аналогии со стековыми объектами надежно идентифицировать конструктор и деструктор здесь нельзя? А вот и нет!

Глобальный объект в силу своей глобальности доступен из многих мест программы, но его конструктор должен вызываться лишь однажды. Как это обеспечить? Конечно, возможны самые различные варианты реализации, но большинство компиляторов идут по простейшему пути, используя для этой цели глобальную переменную‐флаг, изначально равную нулю, а перед первым вызовом конструктора увеличивающуюся на единицу (в более общем случае она устанавливается в `TRUE`). При повторных итерациях остается проверить, равен ли флаг нулю, и если нет — пропустить вызов конструктора. Таким образом конструктор вновь «окольцовывается» условным переходом, что позволяет безошибочно отличить его от всех остальных функций.

С деструктором еще проще — раз объект глобальный, то он уничтожается только при завершении программы. А кто это может отследить, кроме поддержки времени исполнения? Специальная функция, такая как `_atexit`, принимает на вход указатель на деструктор, запоминает его и затем вызывает в случае необходимости. Интересный момент: функция `_atexit` (или что там используется в вашем конкретном случае) должна быть вызвана лишь однократно (надеюсь, понятно почему?). И чтобы не вводить еще один флаг, она вызывается сразу же после вызова конструктора! На первый взгляд объект может показаться состоящим из сочетания конструктора и деструктора, но это не так. Не забывай, что `_atexit` не передает немедленно управление на код деструктора, а только запоминает его указатель для дальнейшего использования!

Таким образом, конструктор и деструктор глобального объекта очень просто идентифицировать, что и доказывает следующий пример myClass3 с измененной функцией main:

```cpp
int main()
{
    printf("Main\n");
    static MyClass zzz;
    zzz.demo();
}
```

Дизассемблерный листинг этого примера выглядит приблизительно так:

```asm
push  rbp
push  rdi
sub   rsp,  108h
lea   rbp,  [rsp+20h]
mov   rdi,  rsp
mov   ecx,  42h
mov   eax,  0CCCCCCCCh
rep   stosd
mov   [rbp+0F0h+var_28], 0FFFFFFFFFFFFFFFEh
lea   rcx,  unk_140024002
call  sub_14001109B
lea   rcx,  aMain  ; "Main\n"
call  sub_140011208 ; вызов printf
mov   eax, 104h
mov   eax,  eax
mov   ecx,  cs:TlsIndex
mov   rdx,  gs:58h
mov   rcx,  [rdx+rcx*8]
mov   eax,  [rax+rcx]
cmp   cs:dword_14001E180, eax
```

Проверяем, инициализирован ли объект.

```asm
jle short loc_140011B10
```

Если объект инициализирован, осуществляем переход на loc_140011B10. Вызов конструктора при этом пропускается.

```asm
lea   rcx,  dword_14001E180
call  sub_1400112F8
cmp   cs:dword_14001E180, 0FFFFFFFFh
jnz   short loc_140011B10
lea   rcx,  unk_14001E17C
call  sub_1400113D4 ; опосредованный вызов конструктора
                    ; класса MyClass
lea   rcx,  sub_140019520
```

А здесь программа заносит в регистр `RCX` указатель на опосредованный вызов деструктора класса `MyClass`.

```asm
call sub_1400111D1
```

Посредством этого вызова программа передает находящийся в `RCX` указатель на деструктор функции `j__crt_atexit_0`, которая будет вызвана по завершении программы. Иными словами, эта функция вызовет деструктор класса `MyClass`.

```asm
nop
lea   rcx,  dword_14001E180
call  sub_140011253
loc_140011B10:
lea   rcx,  unk_14001E17C
call  sub_140011366  ; опосредованный вызов метода demo
                     ; класса MyClass
xor   eax,  eax
lea   rsp,  [rbp+0E8h] pop rdi
pop   rbp   retn
```

Аналогичный код генерирует Embarcadero C++Builder (myClass3_CB).

| ![Embarcadero C++ Builder](./images/09/embarcadero_builder.png)
|:--:|
| *Embarcadero C++ Builder* |

```asm
sub rsp, 48h
mov [rsp+48h+var_4], 0
lea rcx, aMain ; "Main\n"
call printf
cmp cs:byte_45DAB8, 0
```

Смотрим состояние флага инициализации объекта.

```asm
mov [rsp+48h+var_18], eax
jnz short loc_401510
```

Если объект уже инициализирован, делаем прыжок — не вызываем конструктор.

```asm
lea rcx, byte_45DAB8
call cxa_guard_acquire
cmp eax, 0
```

Проверяем состояние регистра `EAX`, если он равен 0, осуществляем переход — т. е. объект инициализирован.

```asm
jz short loc_401510
lea rcx, stru_45DAB0
```

Перед вызовом конструктора помещаем в регистр `RCX` указатель на класс `MyClass`.

```asm
call MyClass::MyClass(void)
    jmp short $+2 ; переход на 2 строчки вперед
loc_4014F4:
    lea rcx, dtor  ZZ4mainE3zzz
    call atexit
```

В двух строках выше происходит такой же маневр, какой мы видели в предыдущем листинге: сначала в регистр `RCX` помещается указатель на функцию, вызывающую деструктор класса `MyClass`. Затем этот указатель передается в функцию `atexit`. Последняя будет вызвана при уничтожении приложения, а уже она вызывает деструктор `MyClass`.

```asm
lea rcx, byte_45DAB8
    mov [rsp+48h+var_1C], eax
    call cxa_guard_release
loc_401510:
    lea rcx, stru_45DAB0
```

Перед вызовом метода demo класса MyClass в регистр RCX помещаем указатель на этот класс.

```asm
call MyClass::demo(void)
mov eax, [rsp+48h+var_4]
add rsp, 48h
retn
```

### Виртуальный деструктор

Деструктор тоже может быть виртуальным. А почему бы и нет? Это бывает полезно, когда экземпляр производного класса удаляется через указатель на базовый класс. Поскольку виртуальные функции связаны с классом объекта, а не с классом указателя, то вызывается виртуальный деструктор, связанный с типом объекта, а не с типом указателя. Впрочем, эти тонкости относятся к непосредственному программированию, а исследователей в первую очередь интересует, как идентифицировать виртуальный деструктор. О, это просто: виртуальный деструктор совмещает в себе свойства обычного деструктора и виртуальной функции (см. прошлую статью).

### Виртуальный конструктор

Виртуальный конструктор? А что, разве есть такой? Ничего подобного стандартный C++ не поддерживает. Непосредственно не поддерживает. И когда виртуальный конструктор позарез требуется программистам (впрочем, это бывает лишь в весьма экзотических случаях), они прибегают к ручной эмуляции некоторого его подобия. В специально выделенную для этих целей виртуальную функцию (не конструктор!) помещается приблизительно следующий код: `return new имя класса (*this)`. Этот трюк кривее, чем бумеранг, но... он работает. Разумеется, существуют и другие решения.

Подробное их обсуждение выходит далеко за рамки нашей беседы и требует глубокого знания C++ (гораздо более глубокого, чем у рядового разработчика), к тому же это заняло бы слишком много места, но едва ли оказалось бы интересным и полезным.

Итак, идентификация виртуального конструктора в силу отсутствия самого понятия в принципе невозможна. Его эмуляция насчитывает десятки (если не больше) решений, попробуй‐ка перечисли их все! Впрочем, этого и не нужно делать, в большинстве случаев виртуальные конструкторы представляют собой виртуальные функции, принимающие в качестве аргумента указатель `this` и возвращающие указатель на новый объект. Не слишком‐то надежно для идентификации, но все же лучше, чем ничего.

### Конструктор раз, конструктор два...

Количество конструкторов у класса может быть и более одного (и очень часто не только может, но и бывает). Однако это никак не влияет на анализ. Сколько бы конструкторов ни присутствовало, для каждого экземпляра класса всегда вызывается только один, выбранный компилятором в зависимости от формы его вызова. Единственная деталь — различные экземпляры класса могут вызывать различные конструкторы. Будь внимателен!

### Зачем козе баян, или Внимание: пустой конструктор!

Некоторые ограничения конструктора (в частности, отсутствие возвращаемого значения) привели к появлению приема программирования «пустой конструктор». Конструктор умышленно оставляется пустым, а весь код инициализации помещается в специальную функцию, как правило называемую `Init`. Обсуждение сильных и слабых сторон такого стиля — предмет отдельного разговора, никаким образом не относящегося к теме анализа. Исследователям достаточно знать: такой стиль есть и активно используется не только отдельно взятыми программистами, но и крупнейшими компаниями‐гигантами (например, той же Microsoft).

Поэтому, встретив вызов пустого конструктора, не удивляйся, это нормально, и ищи функцию инициализации среди обычных членов класса.
