# Идентифицируем возвращаемое функцией значение

> Се­год­ня мы рас­смот­рим все­воз­можные пути воз­вра­та резуль­тата из выз­ванной фун­кции в вызыва­ющую. Уме­ние опре­делять спо­собы воз­вра­щения зна­чения поз­воля­ет хакеру наибо­лее точ­но вос­ста­новить изна­чаль­ный алго­ритм взла­мыва­емой прог­раммы. А если вдо­бавок удас­тся опре­делить тип переда­ваемо­го зна­чения, то задача еще боль­ше упро­щает­ся.

Тра­дици­онно под воз­вра­щаемым фун­кци­ей зна­чени­ем понима­ется то, что переда­ет опе­ратор `return`. Одна­ко это лишь над­водная часть айсбер­га, не рас­кры­вающая всей кар­тины вза­имо­дей­ствия фун­кций друг с дру­гом. В качес­тве наг­лядной демонс­тра­ции рас­смот­рим типич­ный при­мер, в котором про­исхо­дит воз­вра­щение зна­чения в аргу­мен­те, передан­ном по ссыл­ке:

```cpp
int xdiv(int a, int b, int *c=0) {
    if (!b) return -1;
    if (c) c[0] = a % b;
    return a / b;
}
```

Фун­кция `xdiv` воз­вра­щает резуль­тат целочис­ленно­го деления аргу­мен­та `а` на аргу­мент `b`, но, помимо это­го, она записы­вает оста­ток в перемен­ную `с`, передан­ную по ссыл­ке. Так сколь­ко же зна­чений вер­нула фун­кция? И чем воз­вра­щение резуль­тата по ссыл­ке хуже или «незакон­нее» клас­сичес­кого `return`?

По­пуляр­ные изда­ния склон­ны упро­щать проб­лему иден­тифика­ции воз­вра­щен­ного фун­кци­ей зна­чения, рас­смат­ривая один лишь час­тный слу­чай с опе­рато­ром `return`.

Мы же рас­смот­рим сле­дующие механиз­мы:

* воз­врат зна­чения опе­рато­ром `return`;
* воз­врат зна­чений через аргу­мен­ты, передан­ные по ссыл­ке;
* воз­врат зна­чений через динами­чес­кую память (кучу);
* воз­врат зна­чений через гло­баль­ные перемен­ные;
* воз­врат зна­чений через фла­ги про­цес­сора.

Во­обще‑то к это­му спис­ку не помеша­ло бы добавить воз­врат зна­чений через дис­ковые и про­еци­руемые в память фай­лы, но это выходит за рам­ки обсужда­емой темы (хотя, рас­смат­ривая фун­кцию как «чер­ный ящик» с вхо­дом и выходом, нель­зя не приз­нать, что вывод фун­кци­ей резуль­татов сво­ей работы в файл фак­тичес­ки и есть воз­вра­щаемое ею зна­чение).

## ОПЕРАТОР RETURN

По общепри­нято­му сог­лашению на плат­форме x86-64 зна­чение, воз­вра­щаемое опе­рато­ром `return`, помеща­ется в регистр `RAX` (в `EAX` в 32-раз­рядном режиме). Вещес­твен­ные типы (`float`, `double`) — в регистр `XMM0`.

А как воз­вра­щают­ся типы, занима­ющие более 8 байт? Ска­жем, некая фун­кция воз­вра­щает струк­туру, сос­тоящую из сотен бай­тов, или объ­ект сопос­тавимо­го раз­мера. Ни то ни дру­гое в регис­тры не запих­нешь!

Ока­зыва­ется, если воз­вра­щаемое зна­чение не может быть втис­нуто в регис­тры, ком­пилятор скры­то от прог­раммис­та переда­ет фун­кции неяв­ный аргу­мент — ссыл­ку на локаль­ную перемен­ную, в которую и записы­вает­ся воз­вра­щен­ный резуль­тат. Таким обра­зом, фун­кции `mystuct MyFunc(int a, int b)` и `void MyFunc(mystryct *my, int a, int b)` ком­пилиру­ются в иден­тичный (или близ­кий к тому) код, из‑за чего «вытянуть» из машин­ного кода под­линный про­тотип невоз­можно!

Да­вай про­верим наши пред­положе­ния. Откомпи­лируй сле­дующий код с отклю­чен­ной опти­миза­цией в Visual C++:

```cpp
#include <iostream>
// Структура включает три переменные типа double
struct mystruct {
    double d_var1;
    double d_var2;
    double d_var3;
};

mystruct MyFunc1(double a, double b, double c)
{
    mystruct my;
    my.d_var1 = a;
    my.d_var2 = b;
    my.d_var3 = c;

    return my;
}

void MyFunc2(struct mystruct* my, double a, double b, double c)
{
    my->d_var1 = a;
    my->d_var2 = b;
    my->d_var3 = c;
}

int main()
{
    mystruct my;
    my = MyFunc1(1.001, 2.002, 3.003);
    std::cout << my.d_var1 << " " << my.d_var2
              << " " << my.d_var3 << std::endl;

    MyFunc2(&my, 3.004, 5.005, 6.006);
    std::cout << my.d_var1 << " " << my.d_var2
              << " " << my.d_var3 << std::endl;
}
```

| ![Вывод программы](./images/16/a84c065138129ba6671f87ac5a8d1b83.png)
|:--:|
| *Вывод программы* |

Те­перь открой экзешник в IDA и отка­жись от заг­рузки отла­доч­ной информа­ции, ведь при отладке в боевых усло­виях никакой дебаж­ной инфы не будет.

| ![В этом случае ответить отрицательно](./images/16/f1ddd1726a01d743e9cb3d3b653729f4.png)
|:--:|
| *В этом случае ответить отрицательно* |

На сле­дующей иллюс­тра­ции я при­вел обе фун­кции для срав­нения в одном окне VS Code.

| ![VS Code](./images/16/bcf09e6b70127be27f3ddb32933232bc.png)
|:--:|
| *VS Code* |

Хо­тя по объ­ему фун­кции раз­лича­ются, выпол­няемые ими дей­ствия схо­жи. И без отла­доч­ной информа­ции понять их раз­личия неп­росто. Меж­ду тем, если приг­лядеть­ся, мож­но обна­ружить явные намеки, в какой фун­кции про­исхо­дит работа над ссы­лоч­ным типом, а где выпол­няет­ся воз­врат по зна­чению. Смот­ри:

```asm
sub_140001000 proc near ; CODE XREF: main+33↓p
var_38 = qword ptr -38h
var_30 = qword ptr -30h
var_28 = qword ptr -28h
arg_0  = qword ptr 8
arg_8  = qword ptr 10h
arg_10 = qword ptr 18h
arg_18 = qword ptr 20h
; Инициализация переменных-аргументов
movsd  [rsp+arg_18], xmm3
movsd  [rsp+arg_10], xmm2
movsd  [rsp+arg_8], xmm1
; В RCX первый аргумент целочисленный указатель на структуру
mov    [rsp+arg_0], rcx
; Открываем кадр стека
push   rsi
push   rdi
sub    rsp, 28h
; Заполнение полей структуры
movsd  xmm0, [rsp+38h+arg_8]
movsd  [rsp+38h+var_38], xmm0
movsd  xmm0, [rsp+38h+arg_10]
movsd  [rsp+38h+var_30], xmm0
movsd  xmm0, [rsp+38h+arg_18]
movsd  [rsp+38h+var_28], xmm0
lea    rax, [rsp+38h+var_38]
mov    rdi, [rsp+38h+arg_0]
mov    rsi, rax
mov    ecx, 18h
; Побайтно копируем структуру
rep    movsb
; В RAX возвращаем указатель на структуру
mov    rax, [rsp+38h+arg_0]
; Закрываем кадр стека
add    rsp, 28h
pop    rdi
pop    rsi
retn
sub_140001000 endp
```

В начале фун­кции зна­чения копиру­ются из регис­тров в память. Вид­но, что пер­вый аргу­мент целочис­ленный. Мож­но пред­положить, что это ука­затель на струк­туру. Осталь­ные аргу­мен­ты переда­ются в регис­трах `XMM*`, сле­дова­тель­но, пред­став­ляют чис­ла с пла­вающей запятой. Пос­ле откры­тия кад­ра сте­ка зна­чения раз­меща­ются в смеж­ных областях памяти, это может ука­зывать, что они при­над­лежат обще­му кон­тей­неру (струк­туре, мас­сиву).

Да­лее копиру­ется 24 бай­та (`0x18`) или 192 бита (`0xC0`). Если это чис­ло раз­делить на три, получим 64 бита (`0x40`). Что и тре­бова­лось доказать: один эле­мент `double` — это 64 бита, а в струк­туре их три. Пред­послед­ним дей­стви­ем помеща­ем ука­затель на струк­туру в `RAX`, в который будет воз­вра­щен резуль­тат. Пос­ледним дей­стви­ем зак­рыва­ем кадр сте­ка. В ито­ге нап­рашива­ется вывод, что здесь про­исхо­дит воз­врат зна­чения. Сле­дова­тель­но, про­тотип фун­кции выг­лядит так:

```cpp
mystruct Func(double a, double b, double c)
```

Раз­берем дизас­сем­блер­ный лис­тинг вто­рой фун­кции.

```asm
sub_140001060 proc near ; CODE XREF: main+DD↓p
arg_0  = qword ptr 8
arg_8  = qword ptr 10h
arg_10 = qword ptr 18h
arg_18 = qword ptr 20h
; Инициализация переменных-аргументов
movsd  [rsp+arg_18], xmm3
movsd  [rsp+arg_10], xmm2
movsd  [rsp+arg_8], xmm1
mov    [rsp+arg_0], rcx
; Модификация полей структуры
mov    rax, [rsp+arg_0]
movsd  xmm0, [rsp+arg_8]
movsd  qword ptr [rax], xmm0
mov    rax, [rsp+arg_0]
movsd  xmm0, [rsp+arg_10]
movsd  qword ptr [rax+8], xmm0
mov    rax, [rsp+arg_0]
movsd  xmm0, [rsp+arg_18]
movsd  qword ptr [rax+10h], xmm0
retn
sub_140001060 endp
```

На­чало такое же, как в прош­лый раз. Но в этой фун­кции не откры­вает­ся кадр сте­ка, что может ука­зывать на работу с сущес­тву­ющей струк­турой. Далее одно за дру­гим выпол­няет­ся замеще­ние зна­чений полей струк­туры. В кон­це фун­кция ничего не воз­вра­щает. Отсю­да мож­но вывес­ти такой про­тотип:

```cpp
void Func(struct mystruct* my, double a, double b, double c)
```

Нам уда­лось опре­делить раз­личия фун­кций толь­ко пос­ле их глу­боко­го ана­лиза. Теперь взгля­нем на самоде­ятель­ность ком­пилято­ра C++Builder.

| ![Сравнение функций MyFunc1 и MyFunc2, транслированных компилятором C++Builder 10.4](./images/16/abb294160c783634408a60deaf43a48b.png)
|:--:|
| *Сравнение функций MyFunc1 и MyFunc2, транслированных компилятором C++Builder 10.4* |

В этом слу­чае при бег­лом взгля­де раз­ница сов­сем незамет­на. В обе­их фун­кци­ях при­сутс­тву­ет кадр сте­ка. Меж­ду тем и здесь есть зацеп­ка: пос­ле откры­тия кад­ра сте­ка толь­ко в `MyFunc1` содер­жимое регис­тра `RCX` копиру­ется в `RAX`. А в пер­вом по логике вещей находит­ся целочис­ленное зна­чение или ука­затель. Так­же в `MyFunc1` перед зак­рыти­ем кад­ра сте­ка ука­затель на область памяти `RBP+var_20` копиру­ется в регистр `RAX` явно для воз­вра­та резуль­тата, сле­дова­тель­но, эта фун­кция име­ет такой про­тотип:

mystruct MyFunc1(double a, double b, double c)

### ОПРЕДЕЛЕНИЕ ТИПА ВОЗВРАЩАЕМОГО ЗНАЧЕНИЯ

Тип воз­вра­щаемо­го зна­чения мож­но приб­лизитель­но опре­делить по раз­меру регис­тра, в котором он воз­вра­щает­ся. Если воз­вра­щаемое зна­чение помеща­ется в `EAX`, мож­но пред­положить, что воз­вра­щает­ся `int`, `float` или дру­гой четырех­бай­товый тип. Не исклю­чен вари­ант воз­вра­та мень­шего типа, нап­ример `char`. Так­же для воз­вра­та одно­бай­товых типов может быть исполь­зован регистр `AL` или `AX`. В пос­леднем может быть воз­вра­щено двух­бай­товое зна­чение. Вось­мибай­товые типы воз­вра­щают­ся в регис­тре `RAX`, при этом ник­то не зап­реща­ет ком­пилято­ру вер­нуть в нем зна­чение более мел­кого типа, тут уж как повезет. На 64-бит­ной плат­форме это основное средс­тво воз­вра­та зна­чения из фун­кции.

Ес­ли фун­кция при выходе явно прис­ваивает одно­му из перечис­ленных выше регис­тров некото­рое зна­чение, зна­чит, оно воз­вра­щает­ся в вызыва­ющую фун­кцию. Если же эти регис­тры оста­ются неоп­ределен­ными, то, ско­рее все­го, воз­вра­щает­ся тип `void`, то есть нич­то. Уточ­нить информа­цию помога­ет ана­лиз вызыва­ющей фун­кции, а точ­нее, то, как она обра­щает­ся с регис­тра­ми `RAX [EAX]`. Логич­но пред­положить: если вызыва­ющая фун­кция не исполь­зует зна­чения, оставлен­ного вызыва­емой фун­кци­ей в регис­трах `RAX [EAX]`, ее тип — `void`. Но это пред­положе­ние не всег­да вер­но. Час­тень­ко прог­раммис­ты игно­риру­ют воз­вра­щаемое зна­чение, вво­дя иссле­дова­телей в заб­лужде­ние.

Для зак­репле­ния изу­чен­ного рас­смот­рим сле­дующий при­мер, демонс­три­рующий механизм воз­вра­щения основных типов зна­чений:

```cpp
#include <stdio.h>

// Демонстрация возвращения переменной типа char оператором return
char char_func(char a, char b)
{
    return a + b;
}

// Демонстрация возвращения переменной типа int оператором return
int int_func(int a, int b)
{
    return a + b;
}

// Демонстрация возвращения переменной типа int64 оператором return
__int64 int64_func(__int64 a, __int64 b)
{
    return a + b;
}

// Демонстрация возвращения указателя на int оператором return
// Демонстрация возвращения значения через аргументы, переданные по ссылке
int* near_func(int* a, int* b)
{
    int* c = new int;
    c[0] = a[0] + b[0];

    return c;
}

int main()
{
    int a;
    int b;

    a = 0x666;
    b = 0x777;

    printf("%I64x\n",
        char_func(0x1, 0x2) +
        int_func(0x3, 0x4) +
        int64_func(0x5, 0x6) +
        near_func(&a, &b)[0]);
}
```

Ре­зуль­тат его ком­пиляции в Microsoft Visual C++ 2019 с отклю­чен­ной опти­миза­цией будет выг­лядеть так:

```asm
char char_func(char, char) proc near
; Два аргумента размером в байт
arg_0 = byte ptr 8
arg_8 = byte ptr 10h
; Инициализация переменных значениями из регистров, переданных в параметрах
mov   [rsp+arg_8], dl
mov   [rsp+arg_0], cl
```

Из памяти оба зна­чения копиру­ются в 32-бит­ные регис­тры. Вмес­те с тем они пре­обра­зуют­ся из типа `char` в `int`.

```asm
movsx eax, [rsp+arg_0]
movsx ecx, [rsp+arg_8]
```

Сло­жение зна­чений, находя­щих­ся в регис­трах. Сум­ма накап­лива­ется в регис­тре `EAX`. В нем так­же выпол­няет­ся воз­врат резуль­тата в выз­вавшую фун­кцию.

```asm
add eax, ecx
```

К сожале­нию, дос­товер­но опре­делить тип воз­вра­щаемо­го зна­чения невоз­можно. Он с рав­ным успе­хом может пред­став­лять собой `int` и `char`, при­чем `int` даже более веро­ятен, так как сум­ма двух `char` по сооб­ражени­ям безопас­ности дол­жна помещать­ся в `int`, ина­че воз­можно перепол­нение.

```asm
retn
char char_func(char, char) endp

int int_func(int, int) proc near
; Два аргумента размером в 4 байта
arg_0 = dword ptr 8
arg_8 = dword ptr 10h
; Инициализация переменных значениями из регистров, переданных в параметрах
mov [rsp+arg_8], edx
mov [rsp+arg_0], ecx
; Копирование значений из памяти в 32-битные регистры
mov eax, [rsp+arg_8]
mov ecx, [rsp+arg_0]
; Сложение значений из регистров
add ecx, eax
```

Ко­пиро­вание в регистр `EAX` сум­мы, которая будет воз­вра­щена. По всей веро­ятности, тип воз­вра­щаемо­го зна­чения — `int`.

```asm
mov eax, ecx
retn
int int_func(int, int) endp

__int64 int64_func(__int64, __int64) proc near
; Два аргумента размером в четыре слова (8 байт)
arg_0 = qword ptr 8
arg_8 = qword ptr 10h
; Загружаем в память значения из 64-битных регистров
mov [rsp+arg_8], rdx
mov [rsp+arg_0], rcx
; Копируем значения из памяти в более удобные регистры
mov rax, [rsp+arg_8]
mov rcx, [rsp+arg_0]
; Выполняем суммирование
add rcx, rax
; Переносим результат в регистр RAX, который и будет возвращен
mov rax, rcx
```

Сле­дова­тель­но, воз­вра­щаемый тип име­ет раз­мер 64 бита, то есть `int64`, что и тре­бова­лось доказать.

```asm
retn
__int64 int64_func(__int64, __int64) endp

int * near_func(int *, int *) proc near
; Две переменные размером в четыре слова (8 байт)
var_18 = qword ptr -18h
var_10 = qword ptr -10h
; Два аргумента размером в четыре слова (8 байт)
arg_0  = qword ptr 8
arg_8  = qword ptr 10h
; Загружаем в память значения из 64-битных регистров
mov    [rsp+arg_8], rdx
mov    [rsp+arg_0], rcx
; Открываем кадр стека
sub    rsp, 38h
mov    ecx, 4 ; size
; Выделяем 4 байта из кучи
call   operator new(unsigned __int64)
; Заносим указатель на выделенную память в переменную var_10
mov    [rsp+38h+var_10], rax
mov    rax, [rsp+38h+var_10]
mov    [rsp+38h+var_18], rax
```

Сле­дующий вити­ева­тый код исполь­зует­ся, что­бы опре­делить сме­щение внут­ри мас­сива. Стоп! Но у нас же нет никако­го мас­сива. А как же стро­ка `c[0] = a[0] + b[0];`? Получа­ется аж три мас­сива, ведь ком­пилятор обра­баты­вает ука­зате­ли как мас­сивы и наобо­рот.

```asm
mov  eax, 4
imul rax, 0
mov  ecx, 4
imul rcx, 0
; Готовим аргументы для сложения
mov  rdx, [rsp+38h+arg_0]
mov  eax, [rdx+rax]
mov  rdx, [rsp+38h+arg_8]
; Суммируем
add  eax, [rdx+rcx]
; Снова определяем смещение
mov  ecx, 4
imul rcx, 0
mov  rdx, [rsp+38h+var_18]
mov  [rdx+rcx], eax
; Помещаем в RAX возвращаемое значение
mov  rax, [rsp+38h+var_18]
; Закрываем кадр стека
add  rsp, 38h
retn
int * near_func(int *, int *) endp

main proc near
var_38 = dword ptr -38h
var_30 = qword ptr -30h
var_28 = qword ptr -28h
b      = dword ptr -20h
a      = dword ptr -1Ch
var_18 = qword ptr -18h
; __unwind { // __GSHandlerCheck
; Открываем кадр стека
sub    rsp, 58h
mov    rax, cs:__security_cookie
xor    rax, rsp
; Инициализируем переменные
mov    [rsp+58h+var_18], rax
; В переменную a типа int заносим значение 0x666
mov    [rsp+58h+a], 666h
; В переменную b типа int заносим значение 0x777
mov    [rsp+58h+b], 777h
; В 8-битные регистры помещаем параметры типа char перед вызовом функции
mov    dl, 2 ; b
mov    cl, 1 ; a
```

Вы­зыва­ем фун­кцию `char_func(1,2)`. Как мы пом­ним, у нас были сом­нения в типе воз­вра­щаемо­го ею зна­чения: либо `int`, либо `char`.

```asm
call char_func(char,char)
```

Рас­ширя­ем воз­вра­щен­ное фун­кци­ей зна­чение до `signed int`, сле­дова­тель­но, она воз­вра­тила `signed char`.

```asm
movsx eax, al
; Сохраняем результат в переменной var_38
mov   [rsp+58h+var_38], eax
; На этот раз помещаем параметры в 32-битные регистры, следовательно, их тип int
mov   edx, 4 ; b
mov   ecx, 3 ; a
; Вызываем функцию int_func(3,4), возвращающую значение типа int
call  int_func(int,int)
mov   ecx, [rsp+58h+var_38]
; Прибавляем результат к значению переменной var_38
add   ecx, eax
mov   eax, ecx
```

Пре­обра­зуем двой­ное сло­во, содер­жаще­еся в регис­тре `EAX`, в чет­верное, помеща­емое в регистр `RAX`. Это говорит о том, что тип воз­вра­щен­ного фун­кци­ей зна­чения пре­обра­зует­ся из `int` в `int64`. Пока непонят­но, для чего и зачем.

```asm
cdqe
; Копируем расширенное четверное слово в переменную var_30
mov [rsp+58h+var_30], rax
; Готовим 32-битные параметры
mov edx, 6 ; b
mov ecx, 5 ; a
```

Вы­зыва­ем фун­кцию `int64_func(5,6)`, воз­вра­щающую тип `int64`. Теперь ста­новит­ся понят­но, чем выз­вано рас­ширение пре­дыду­щего резуль­тата.

```asm
call  int64_func(__int64,__int64)
mov   rcx, [rsp+58h+var_30]
; Прибавляем результат, возвращенный функцией int64_func, к четверному слову
add   rcx, rax
mov   rax, rcx
; Сохраняем результат предыдущего сложения в переменной var_28
mov   [rsp+58h+var_28], rax
```

Для переда­чи в качес­тве парамет­ров заг­ружа­ем в регис­тры `RDX` и `RCX` ука­зате­ли на перемен­ные `b` и `a`, зна­чения которых опре­деле­ны в начале фун­кции.

```asm
lea     rdx, [rsp+58h+b] ; b
lea     rcx, [rsp+58h+a] ; a
; Вызываем near_func
call    near_func(int *,int *)
; Определяем смещение
mov     ecx, 4
imul    rcx, 0
; Расширяем двойное слово, на которое указывают два регистра [rax+rcx], в четверное слово
movsxd  rax, dword ptr [rax+rcx]
mov     rcx, [rsp+58h+var_28]
; Складываем два четверных слова
add     rcx, rax
mov     rax, rcx
mov     rdx, rax
; Передаем формат строки вывода
lea     rcx, _Format ; "%I64x\n"
; И, наконец, выводим результат
call    printf
xor     eax, eax
mov     rcx, [rsp+58h+var_18]
xor     rcx, rsp ; StackCookie
call    __security_check_cookie
add     rsp, 58h
retn
main endp
```

Как мы видим, в иден­тифика­ции типа зна­чения, воз­вра­щен­ного опе­рато­ром `return`, ничего хит­рого нет, все про­заич­но. Но не будем спе­шить. Рас­смот­рим сле­дующий при­мер, демонс­три­рующий воз­вра­щение струк­туры по зна­чению. Как ты дума­ешь, что имен­но и в каких регис­трах будет воз­вра­щать­ся?

```cpp
#include <stdio.h>
#include <string.h>

struct XT {
    char s0[4];
    int x;
};

// Функция возвращает значение типа "структура XT" по значению
struct XT MyFunc(const char* a, int b)
{
    struct XT xt;
    strcpy_s(&xt.s0[0], 4, a);
    xt.x = b;
    return xt;
}

int main()
{
    struct XT xt;
    xt = MyFunc("Hello, Sailor!", 0x666);
    printf("%s %x\n", &xt.s0[0], xt.x);
}
```

Вни­мание! Не запус­кай откомпи­лиро­ван­ную прог­рамму. Мало того что она содер­жит ошиб­ку (в буфер вмес­тимостью четыре сим­вола помеща­ется стро­ка раз­мером в 14 сим­волов), так еще некото­рые анти­виру­сы счи­тают такую прог­рамму мал­варью.

Заг­лянем в дизас­сем­бли­рован­ный резуль­тат:

```asm
struct XT MyFunc(char const *, int) proc near
```

Эти локаль­ные перемен­ные на самом деле эле­мен­ты «рас­щеплен­ной» струк­туры `XT`. Как уже ранее говори­лось в статье «[Фун­дамен­таль­ные осно­вы хакерс­тва. Как иден­тифици­ровать струк­туры и объ­екты в чужой прог­рамме](./10.md)», ком­пилятор всег­да стре­мит­ся обра­щать­ся к эле­мен­там струк­туры по их фак­тичес­ким адре­сам, а не через базовый ука­затель. Поэто­му не так‑то прос­то отли­чить струк­туру от не свя­зан­ных меж­ду собой перемен­ных, а под­час это и вов­се невоз­можно!

```asm
Dst    = byte ptr -18h
var_10 = qword ptr -10h
; Функция принимает два аргумента
Src    = qword ptr 8
arg_8  = dword ptr 10h
; Берем значения аргументов из регистров
mov   [rsp+arg_8], edx
mov   [rsp+Src], rcx
; Резервируем место для локальных переменных
sub   rsp, 38h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+38h+var_10],    rax
; Определяем смещение
mov   eax, 1
imul  rax, 0
; Помещаем в RAX указатель на буфер-приемник
lea   rax, [rsp+rax+38h+Dst]
; Подготовка параметров для вызова strcpy_s
mov   r8,[rsp+38h+Src]  ; Src — буфер-источник
mov   edx, 4     ; SizeInBytes
mov   rcx, rax   ; Dst
call  cs:__imp_strcpy_s
; Помещаем в EAX значение второго аргумента — 32-битного числа
mov   eax, [rsp+38h+arg_8]
; Помещаем число из регистра EAX сразу после буфера-приемника в памяти
mov   dword ptr [rsp+38h+Dst+4], eax
```

В `RAX` копиру­ется ука­затель на чет­верное сло­во, начиная с адре­са при­емни­ка. Явная заг­рузка в регистр `RAX` перед выходом из фун­кции ука­зыва­ет на то, что это и есть зна­чение, воз­вра­щаемое фун­кци­ей.

```asm
mov   rax, qword ptr [rsp+38h+Dst]
mov   rcx, [rsp+38h+var_10]
xor   rcx, rsp ; StackCookie
call  __security_check_cookie
add   rsp, 38h
retn
struct XT MyFunc(char const *, int) endp
main   proc near
; Три локальные переменные
var_28 = qword ptr -28h
var_20 = qword ptr -20h
var_18 = qword ptr -18h
; Резервируем 0x48 байт под локальные переменные
sub    rsp, 48h
; Открываем кадр стека
mov    rax, cs:__security_cookie
xor    rax, rsp
mov    [rsp+48h+var_18], rax
; Подготовка параметров для вызова функции
mov    edx, 666h ; b
lea    rcx, a ; "Hello, Sailor!"
call   MyFunc(char const *,int)
```

Воз­вра­щен­ный резуль­тат помеща­ется в перемен­ную `var_28`, а затем в `var_20`. Как мы пом­ним, воз­вра­щен­ный резуль­тат есть ука­затель на область памяти.

```asm
mov  [rsp+48h+var_28], rax
mov  rax, [rsp+48h+var_28]
mov  [rsp+48h+var_20], rax
; Находим смещение
mov  eax, 1
imul rax, 0
; В RAX загружаем указатель на область памяти
lea  rax, [rsp+rax+48h+var_20]
```

Об­рати вни­мание: про­исхо­дит раз­деление памяти, в регистр `R8D` помеща­ется ука­затель на вто­рой кусок выделен­ной памяти, где хра­нит­ся чис­ло.

```asm
mov r8d, dword ptr [rsp+48h+var_20+4]
mov rdx, rax
lea rcx, _Format ; "%s %x\n"
```

В ито­ге получи­лось, что фун­кции `printf` были переда­ны три парамет­ра: в регис­тре `RCX` — фор­матная стро­ка, в регис­тре `RDX` — стро­ка для вывода, в `R8D`, пред­став­ляющем собой 32-бит­ную часть регис­тра `R8`, — чис­ло. Нем­ного пораз­мыслив, мы даже смо­жем вос­ста­новить под­линный про­тотип:

```cpp
struct X{char a[4]; int b} MyFunc(char* c, int d);
```

```asm
call  printf
xor   eax, eax
mov   rcx, [rsp+48h+var_18]
; Закрываем кадр стека
xor   rcx, rsp ; StackCookie
call  __security_check_cookie
add   rsp, 48h
retn
main endp
```

А теперь слег­ка изме­ним струк­туру `XT`, заменив `char s0[4]` `char s0[10]`, что гаран­тирован­но не вле­зет в регистр `RAX`, и пос­мотрим, как изме­нит­ся от это­го код:

```asm
main  proc near
var_58 = byte ptr -58h
result = XT ptr -48h
 var_38 = byte ptr -38h
var_2C = dword ptr -2Ch
var_28 = qword ptr -28h
; Открываем кадр стека
push  rsi
push  rdi
; Выделяем 0x68 байт под локальные переменные
sub   rsp, 68h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+78h+var_28],    rax
; Передаем функции MyFunc крайний правый аргумент — значение 0x666 типа int
mov   r8d, 666h    ; b
; Передаем функции MyFunc второй справа аргумент — указатель на строку
lea   rdx, a       ; "Hello, Sailor!"
; Передаем функции MyFunc указатель на переменную result
lea   rcx, [rsp+78h+result]        ; result
```

Стоп! Это­го аргу­мен­та не было в про­тоти­пе фун­кции! Отку­да же он взял­ся? Вер­но, не было. Его вста­вил ком­пилятор для воз­вра­щения струк­туры по зна­чению. Пос­леднюю фра­зу вооб­ще‑то сто­ило бы зак­лючить в кавыч­ки для при­дания ей иро­нич­ного оттенка: струк­тура, воз­вра­щаемая по зна­чению, в дей­стви­тель­нос­ти воз­вра­щает­ся по ссыл­ке.

```asm
; Вызываем MyFunc
call MyFunc(char const *,int)
; В RCX загружаем указатель на переменную var_58
lea rcx, [rsp+78h+var_58]
; Перемещаем этот указатель в RDI
mov rdi, rcx
```

`MyFunc` воз­вра­тила в `RAX` ука­затель на воз­вра­щен­ную ей по ссыл­ке струк­туру. Помеща­ем этот ука­затель в `RSI`.

```asm
mov     rsi, rax
mov     ecx, 10h
; Копируем 0x10 байт из локации памяти, на которую указывает RSI, в локацию, на которую указывает RDI
rep     movsb
; Следующий код снова копирует 0x10 байт, только на этот раз из переменной var_58 в переменную var_38
lea     rax, [rsp+78h+var_38]
lea     rcx, [rsp+78h+var_58]
mov     rdi, rax
mov     rsi, rcx
mov     ecx, 10h
rep     movsb
; Определяем нулевое смещение
mov     eax, 1
imul    rax, 0
; В RAX помещаем указатель на переменную var_38, чтобы потом передать в качестве второго параметра
lea     rax, [rsp+rax+78h+var_38]
; Подготавливаем параметры для вызова printf
; var_2C явно указывает на 32-битное число
mov     r8d, [rsp+78h+var_2C]
mov     rdx, rax
lea     rcx, _Format                 ; "%s %x\n"
; Запихнув в RCX форматную строку, вызываем функцию вывода данных на консоль
call    printf
xor     eax, eax
mov     rcx, [rsp+78h+var_28]
xor     rcx, rsp                     ; StackCookie
call    __security_check_cookie
add     rsp, 68h
; Закрываем кадр стека
pop     rdi
pop     rsi
retn
main    endp
struct  XT   MyFunc(char const *, int) proc near
Dst    = byte ptr -38h
var_2C = dword ptr -2Ch
var_28 = qword ptr -28h
```

Об­рати вни­мание: фун­кции переда­ются три аргу­мен­та, а не два, как было объ­явле­но в про­тоти­пе:

```asm
arg_0  = qword ptr 8
Src    = qword ptr 10h
arg_10 = dword ptr 18h
mov   [rsp+arg_10], r8d
mov   [rsp+Src],    rdx
mov   [rsp+arg_0],  rcx
; Открываем кадр стека
push  rsi
push  rdi
; Резервируем 0x48 байт под локальные переменные
sub   rsp, 48h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+58h+var_28],    rax
; Определяем смещение
mov   eax, 1
imul  rax, 0
; В RAX загружаем указатель на буфер-приемник
lea   rax, [rsp+rax+58h+Dst]
; Подготавливаем параметры для вызова strcpy_s
; В R8 загружаем указатель на буфер-источник
mov   r8,  [rsp+58h+Src]    ; Src
mov   edx, 0Ah     ; SizeInBytes
mov   rcx, rax     ; Dst
call  cs:__imp_strcpy_s
; Копируем в EAX третий аргумент, очевидно число 0x666
mov   eax, [rsp+58h+arg_10]
; Помещаем это число в переменную var_2C, которая находится «рядом» с Dst
mov   [rsp+58h+var_2C],    eax
; В RAX загружаем указатель на буфер-приемник
lea   rax, [rsp+58h+Dst]
```

В `RDI` помеща­ется ука­затель на аргу­мент `arg_0`. Как мы уже зна­ем, в этом аргу­мен­те ком­пилятор скры­то от прог­раммис­та переда­ет ука­затель на локаль­ную перемен­ную. Имен­но в эту перемен­ную фун­кция помеща­ет струк­туру, воз­вра­щаемую «по зна­чению».

```asm
mov rdi, [rsp+58h+arg_0]
mov rsi, rax
mov ecx, 10h
```

Ко­пиру­ем `0x10` байт из локации памяти, на которую ука­зыва­ет `RSI`, в локацию, на которую ука­зыва­ет `RDI`, а там находит­ся область памяти для воз­вра­та.

```asm
rep movsb
```

Об­рати вни­мание: стро­ка занима­ет `0xA` (10) байт, а в область `RDI` было ско­пиро­вано `0x10` (16) байт. Зна­чит, будет воз­вра­щено 16 байт, но отку­да взя­лись еще два бай­та, если из десяти байт сос­тоит стро­ка и из четырех бай­тов — чис­ло (тип `int`)?

В исходном тек­сте было записа­но `char s0[10]`, но ком­пилятор, вырав­нивая эле­мен­ты струк­туры по адре­сам, крат­ным четырем, перенес вто­рой эле­мент — `int x` — по адре­су `base+0xC`, тем самым соз­дав «дыру» меж­ду кон­цом стро­ки и началом вто­рого эле­мен­та.

Ана­лиз дизас­сем­блер­ного лис­тинга не поз­воля­ет вос­ста­новить истинный вид струк­туры. Единс­твен­ное, что мож­но ска­зать, — дли­на стро­ки `s0` лежит в интерва­ле `[0x9 — 0xC]`. В `RAX` для воз­вра­та помеща­ется ука­затель на `arg_0`, который изна­чаль­но был нас­тро­ен для воз­вра­щения.

```asm
mov  rax, [rsp+58h+arg_0]
mov  rcx, [rsp+58h+var_28]
xor  rcx, rsp                 ; StackCookie
call __security_check_cookie
; Закрываем кадр стека
add  rsp, 48h
pop  rdi
pop  rsi
retn
struct XT MyFunc(char const *, int)  endp
```

Воз­ника­ет воп­рос: а как воз­вра­щают­ся струк­туры, сос­тоящие из сотен и тысяч бай­тов? Ответ: они копиру­ются в локаль­ную перемен­ную, неяв­но передан­ную ком­пилято­ром по ссыл­ке. То есть точ­но так же, как мы видели в прош­лом при­мере со струк­турой, нем­ного пре­вос­ходящей раз­мер 64-бит­ного регис­тра. Что­бы уви­деть это, изме­ним в исходном тек­сте пре­дыду­щего при­мера `char s0[10]` на `char s0[0x666]`. Резуль­тат пов­торной ком­пиляции дол­жен выг­лядеть так (при­веде­на толь­ко фун­кция `MyFunc`, пре­тер­певшая наиболь­шие изме­нения):

```asm
struct  XT MyFunc(char const *, int) proc near
Dst    = byte ptr -698h
var_30 = dword ptr -30h
var_28 = qword ptr -28h
arg_0  = qword ptr 8
Src    = qword ptr 10h
arg_10 = dword ptr 18h
mov     [rsp+arg_10],    r8d
mov     [rsp+Src],       rdx
mov     [rsp+arg_0],     rcx
push    rsi
push    rdi
```

Для локаль­ных перемен­ных выделя­ется зна­читель­но боль­ше памяти. И мы наперед зна­ем, почему.

```asm
sub   rsp, 6A8h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+6B8h+var_28], rax
; Определяем смещение
mov   eax, 1
imul  rax, 0
; В RAX загружаем указатель на буфер-приемник
lea   rax, [rsp+rax+6B8h+Dst]
mov   r8,  [rsp+6B8h+Src] ; Src
mov   edx, 666h   ; SizeInBytes
mov   rcx, rax    ; Dst
call  cs:__imp_strcpy_s
; Копируем в EAX третий аргумент, очевидно число 0x666
mov   eax, [rsp+6B8h+arg_10]
; Помещаем это число в переменную var_30, которая находится «рядом» с Dst
mov   [rsp+6B8h+var_30], eax
; В RAX загружаем указатель на буфер-приемник
lea   rax, [rsp+6B8h+Dst]
```

В `RDI` помеща­ется ука­затель на аргу­мент `arg_0`, пред­став­ляющий ука­затель на воз­вра­щаемую струк­туру.

```asm
mov rdi, [rsp+6B8h+arg_0]
mov rsi, rax
mov ecx, 66Ch
```

Ко­пиру­ем `0x66С` байт памяти из локации, на которую ука­зыва­ет `RSI`, в локацию, куда ука­зыва­ет `RDI`, а там находит­ся область памяти для воз­вра­та.

```asm
rep  movsb
mov  rax, [rsp+6B8h+arg_0]
mov  rcx, [rsp+6B8h+var_28]
xor  rcx, rsp                  ; StackCookie
call __security_check_cookie
add  rsp, 6A8h
pop  rdi
pop  rsi
retn
struct XT MyFunc(char const *, int)  endp
```

### ВОЗВРАЩЕНИЕ ВЕЩЕСТВЕННЫХ ЗНАЧЕНИЙ

В сов­ремен­ных про­цес­сорах, под­держи­вающих тех­нологию SSE2, работа с вещес­твен­ными типами дан­ных мало чем отли­чает­ся от работы с целочис­ленны­ми. Единс­твен­ное раз­личие — это исполь­зование дру­гих регис­тров и нового набора команд. Для под­твержде­ния ска­зан­ного иссле­дуем сле­дующий при­мер:

```cpp
#include <stdio.h>

float MyFunc(float a, float b)
{
    return a + b;
}
int main()
{
    printf("%f\n", MyFunc(6.66f, 7.77f));
}
```

Ре­зуль­тат его ком­пиляции Microsoft Visual C++ дол­жен выг­лядеть приб­лизитель­но так:

```asm
float MyFunc(float, float) proc near
arg_0 = dword ptr 8
arg_8 = dword ptr 10h
```

В память заг­ружа­ются передан­ные в регис­трах аргу­мен­ты с пла­вающей точ­кой оди­нар­ной точ­ности. Если бы вмес­то `float` мы исполь­зовали тип `double`, коман­да заг­рузки зна­чения при­няла бы вид `movd`.

```asm
movss [rsp+arg_8], xmm1
movss [rsp+arg_0], xmm0
movss xmm0, [rsp+arg_0]
```

А вот сло­жение двух зна­чений с пла­вающей точ­кой оди­нар­ной точ­ности. Если бы вмес­то `float` мы исполь­зовали тип `double`, коман­да сло­жения зна­чений при­няла бы вид `addsd`. Воз­вра­щение резуль­тата с пла­вающей точ­кой про­исхо­дит в регис­тре `XMM0`.

```asm
addss xmm0, [rsp+arg_8]
retn
float MyFunc(float, float) endp
main  proc near
; Выделение памяти для локальных переменных
sub   rsp, 28h
```

Ни­же при­веден код заг­рузки в регис­тры `XMM0` и `XMM1` чисел с пла­вающей точ­кой оди­нар­ной точ­ности. Если бы вмес­то `float` мы исполь­зовали тип `double`, коман­да заг­рузки зна­чения при­няла бы вид `movd`.

```asm
movss xmm1, cs:__real@40f8a3d7 ; b
movss xmm0, cs:__real@40d51eb8 ; a
; Вызов функции MyFunc вместе с передачей параметров
call MyFunc(float,float)
```

`MyFunc` воз­вра­щает резуль­тат с пла­вающей точ­кой в регистр `XMM0`. Сле­дующая инс­трук­ция пре­обра­зует зна­чение оди­нар­ной точ­ности в зна­чение двой­ной точ­ности (с такими дан­ными 64-бит­ному про­цес­сору удоб­нее работать). Если бы вмес­то `float` мы исполь­зовали тип `double`, то сле­дующая инс­трук­ция не понадо­билась бы.

```asm
cvtss2sd xmm0, xmm0
```

Ко­пиро­вание двух чет­верных слов, содер­жащих четыре зна­чения с пла­вающей точ­кой оди­нар­ной точ­ности:

```asm
movaps xmm1, xmm0
; Копирование одного четверного слова
movq rdx, xmm1
; Подготовка форматной строки
lea rcx, _Format ; "%f\n"
; Вывод значения на консоль
call printf
xor eax, eax
add rsp, 28h
retn
main endp
```

Оче­вид­но, что работа со зна­чени­ями с пла­вающей точ­кой в сов­ремен­ных реалиях не пред­став­ляет труд­ностей.

## ВОЗВРАЩЕНИЕ ЗНАЧЕНИЙ INLINE-ФУНКЦИЯМИ ASSEMBLER

Соз­датель ассем­блер­ной фун­кции волен воз­вра­щать зна­чения в любых регис­трах, каких ему будет угод­но, одна­ко, пос­коль­ку вызыва­ющие фун­кции язы­ка высоко­го уров­ня ожи­дают уви­деть резуль­тат вычис­лений в стро­го опре­делен­ных регис­трах, писаные сог­лашения при­ходит­ся соб­людать. Дру­гое дело — внут­ренние ассем­блер­ные фун­кции: они могут вооб­ще не при­дер­живать­ся никаких пра­вил. Меж­ду тем, сог­ласно докумен­тации на ком­пилятор, я исполь­зую Visual C++ 2019 с самыми пос­ледни­ми обновле­ниями: ат­рибут `naked` допус­тим толь­ко на плат­формах x86 и ARM и недос­тупен в x64. В новой вер­сии Visual Studio 2022 в этом пла­не ничего не изме­нилось. Поэто­му в сле­дующем при­мере, демонс­три­рующем воз­вра­щение зна­чений встро­енны­ми ассем­блер­ными фун­кци­ями, мы будем исполь­зовать ассем­блер для 32-бит­ных про­цес­соров:

```cpp
#include <stdio.h>
```

Naked-фун­кция не име­ет про­тоти­па, обо всем дол­жен заботить­ся сам прог­раммист. Кро­ме того, «голые» фун­кции не содер­жат ни про­лога, ни эпи­лога.

```asm
__declspec(naked) int MyFunc() {
__asm {
lea ebp,[eax + ecx];
```

Воз­вра­щаем в `EBP` сум­му зна­чений регис­тров `EAX` и `ECX`. Такой трюк допус­тим лишь при усло­вии, что эта фун­кция будет вызывать­ся из ассем­блер­ной фун­кции, зна­ющей, через какие регис­тры переда­ются аргу­мен­ты и через какие воз­вра­щает­ся резуль­тат вычис­лений.

```asm
ret
    }
}
```

```cpp
int main()
{
    int a = 0x666;
    int b = 0x777;
    int c;

    __asm {
      push ebp
      push edi
      mov eax,[a];
      mov ecx,[b];
      lea edi,c
```

Вы­зыва­ем фун­кцию `MyFunc` из ассем­блер­ной фун­кции, переда­вая ей аргу­мен­ты через те регис­тры, которые она «хочет».

```asm
call MyFunc;
```

При­нима­ем воз­вра­щен­ное в `EBP` зна­чение и сох­раня­ем его в локаль­ной перемен­ной.

```cpp
        mov [edi],ebp
        pop edi
        pop ebp
    }

    printf("%x\n", c); // Выводим результат
}
```

В ито­ге получим такой резуль­тат:

| ![Результат ассемблерной функции](./images/16/7507162b5f13d71f948843a8b970bb5e.png)
|:--:|
| *Результат ассемблерной функции* |

Ре­зуль­тат ком­пиляции Microsoft Visual C++ (а дру­гими ком­пилято­рами этот при­мер откомпи­лиро­вать и вов­се не удас­тся, ибо они не под­держи­вают клю­чевое сло­во `naked`) дол­жен выг­лядеть так:

```asm
int __cdecl MyFunc(void)          ; CODE XREF: _main+2A↓p
; Принимаем аргументы через регистры EAX и ECX, возвращая через регистр EBP
; их сумму. Конечно, пример несколько надуман, зато нагляден!
lea    ebp, [eax+ecx]
retn
int    __cdecl MyFunc(void)  endp
_main  proc near       ; CODE XREF: __scrt_common_main_seh+F5↓p
c     = dword ptr -10h
b     = dword ptr -0Ch
a     = dword ptr -8
var_4 = dword ptr -4
argc  = dword ptr 8
argv  = dword ptr 0Ch
envp  = dword ptr 10h
; Открываем кадр стека
push   ebp
mov    ebp, esp
; Резервируем место для локальных переменных
sub    esp, 10h
mov    eax, ___security_cookie
xor    eax, ebp
mov    [ebp+var_4],    eax
; Сохраняем изменяемый регистр
push   edi
; Инициализируем переменные a и b
mov    [ebp+a],        666h
mov    [ebp+b],        777h
; Сохраняем регистры или передаем их функции?
; Пока нельзя ответить однозначно
push   ebp
push   edi
; Загружаем в EAX значение переменной A
mov    eax, [ebp+a]
; В ECX значение переменной b
mov    ecx, [ebp+b]
; В EDI загружаем указатель на переменную C
lea    edi, [ebp+c]
```

Вы­зыва­ем `MyFunc`. Из ана­лиза вызыва­ющей фун­кции не очень понят­но, как ей переда­ются парамет­ры. Может, через стек, а может, и через регис­тры. Толь­ко иссле­дова­ние кода `MyFunc` поз­воля­ет уста­новить, что вер­ным ока­зыва­ется пос­леднее пред­положе­ние. Да, парамет­ры переда­ются через регис­тры!

```asm
call MyFunc(void)
```

Что бы это зна­чило? Ана­лиз одной лишь вызыва­ющей фун­кции не может дать исчерпы­вающе­го отве­та, и толь­ко ана­лиз вызыва­емой фун­кции под­ска­зыва­ет, что через `EBP` пос­ледняя воз­вра­щает резуль­тат вычис­лений.

```asm
mov [edi], ebp
```

Вос­ста­нав­лива­ем изме­нен­ные регис­тры. Это говорит о том, что выше эти регис­тры дей­стви­тель­но сох­ранялись в сте­ке, а не переда­вались фун­кции в качес­тве парамет­ров.

```asm
pop    edi
pop    ebp
; Загружаем в EAX содержимое переменной C
mov    eax,   [ebp+c]
push   eax
push   offset _Format        ; "%x\n"
; Вызываем printf
call   _printf
add    esp,   8
; Восстанавливаем регистр
pop    edi
mov    ecx,   [ebp+var_4]
xor    ecx,   ebp            ; cookie
call   __security_check_cookie(x)
; Закрываем кадр стека
mov    esp,   ebp
pop    ebp
retn
_main  endp
```

## ВОЗВРАТ ЗНАЧЕНИЙ ЧЕРЕЗ АРГУМЕНТЫ, ПЕРЕДАННЫЕ ПО ССЫЛКЕ

Иден­тифика­ция зна­чений, воз­вра­щен­ных через передан­ные по ссыл­ке аргу­мен­ты, тес­но переп­лета­ется с иден­тифика­цией самих аргу­мен­тов. Выделив сре­ди передан­ных фун­кции аргу­мен­тов ука­зате­ли, заносим их в спи­сок кан­дидатов на воз­вра­щаемые зна­чения.

Те­перь поищем, нет ли сре­ди них ука­зате­лей на неини­циали­зиро­ван­ные перемен­ные, оче­вид­но, их ини­циали­зиру­ет сама вызыва­емая фун­кция. Одна­ко не сто­ит вычер­кивать ука­зате­ли на ини­циали­зиро­ван­ные перемен­ные (осо­бен­но рав­ные нулю): они так­же могут воз­вра­щать зна­чения. Уточ­нить ситу­ацию поз­волит ана­лиз вызыва­емой фун­кции — нас будут инте­ресо­вать все опе­рации модифи­кации перемен­ных, передан­ных по ссыл­ке. Толь­ко не путай это с модифи­каци­ей перемен­ных, передан­ных по зна­чению. Пос­ледние авто­мати­чес­ки уми­рают в момент завер­шения фун­кции (точ­нее, вычис­тки аргу­мен­тов из сте­ка). Фак­тичес­ки это локаль­ные перемен­ные фун­кции, и она без­болез­ненно может изме­нять их, как ей взду­мает­ся.

Сле­дующий при­мер демонс­три­рует воз­врат зна­чений через перемен­ные, передан­ные по ссыл­ке:

```cpp
#include <stdio.h>
#include <string.h>

// Функция инвертирования строки src с ее записью в строку dst
void Reverse(char* dst, const char* src)
{
    strcpy_s(dst, 100, src);
    _strrev(dst);
}

// Функция инвертирования строки s (результат записывается в саму же строку s)
void Reverse(char* s)
{
    _strrev(s);
}

// Функция возвращает сумму двух аргументов
int sum(int a, int b)
{
    // Мы можем безболезненно модифицировать аргументы, переданные по значению,
    // обращаясь с ними как с обычными локальными переменными
    a += b;
    return a;
}

int main()
{
    char s0[] = "Hello,Sailor!";
    char s1[100];
    // Инвертируем строку s0, записывая ее в s1
    Reverse(&s1[0], &s0[0]);
    printf("%s\n", &s1[0]);
    // Инвертируем строку s1, перезаписывая ее
    Reverse(&s1[0]);
    printf("%s\n", &s1[0]);
    // Выводим сумму двух чисел
    printf("%x\n", sum(0x666, 0x777));
}
```

| ![Вывод приложения return_ref_vars](./images/16/08ed578f9aa57cff242668bf2a9b1f3f.png)
|:--:|
| *Вывод приложения return_ref_vars* |

Ре­зуль­тат ком­пиляции это­го при­мера с помощью Visual C++ дол­жен выг­лядеть приб­лизитель­но так:

```asm
main  proc near
src    = byte ptr -0A8h
dst    = byte ptr -98h
var_28 = qword ptr -28h
; Открываем кадр стека
push  rsi
push  rdi
; Резервируем память для локальных переменных
sub   rsp, 0B8h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+0C8h+var_28],    rax
lea   rax, [rsp+0C8h+src]
; Загружаем в регистр RCX указатель на строку
lea   rcx, aHelloSailor            ; "Hello,Sailor!"
mov   rdi, rax
; Указатель перемещается в регистр RSI, поскольку регистр RCX понадобится для других
; целей
mov   rsi, rcx
; Число 0xE (14) помещается в регистр ECX...
mov   ecx, 0Eh
; ...столько байтов будет скопировано следующий инструкцией, то есть вся строка "Hello,Sailor!" +
; конечный ноль
rep   movsb
; Определяем нулевое смещение
mov   eax, 1
imul  rax, 0
; В RAX загружаем указатель на буфер-источник
lea   rax, [rsp+rax+0C8h+src]
; Определяем нулевое смещение
mov   ecx, 1
imul  rcx, 0
; В RCX загружаем указатель на буфер-приемник — неинициализированную переменную
lea   rcx, [rsp+rcx+0C8h+dst]      ; dst
; Перекладываем указатель на источник в регистр RDX
mov   rdx, rax    ; src
```

В регис­трах уком­плек­тованы парамет­ры. Переда­ем их вызыва­емой фун­кции `Reverse`. В самом пра­вом парамет­ре переда­ется ука­затель на неини­циали­зиро­ван­ную перемен­ную. Это говорит о том, что, ско­рее все­го, фун­кция воз­вра­тит через нее некото­рое зна­чение. Возь­мем это себе на замет­ку.

```asm
call  Reverse(char    *,char const *)
; Определяем нулевое смещение
mov   eax, 1
imul  rax, 0
; Загружаем в RAX указатель на переменную dst — буфер-приемник, возвращенный
; из функции Reverse
lea   rax, [rsp+rax+0C8h+dst]
; Перекладываем его в регистр RDX
mov   rdx, rax
; В RCX помещаем форматную строку
lea   rcx, _Format    ; "%s\n"
```

Пе­реда­ча фун­кции `printf` во вто­ром парамет­ре ука­зате­ля на перемен­ную `dst`. Пос­коль­ку вызыва­ющая фун­кция не ини­циали­зиро­вала эту перемен­ную, мож­но пред­положить, что вызыва­емая воз­вра­тила через нее свое зна­чение. Воз­можно, фун­кция `Reverse` модифи­циро­вала и перемен­ную `src`, одна­ко об этом нель­зя ска­зать с опре­делен­ностью до тех пор, пока не будет изу­чен ее код. Вызов фун­кции `printf` для вывода стро­ки выг­лядит так:

```asm
call printf
; Определяем нулевое смещение
mov  eax, 1
imul rax, 0
lea  rax, [rsp+rax+0C8h+dst]
```

В регистр `RCX` помеща­ем ука­затель на перемен­ную `dst` (буфер‑при­емник). Это единс­твен­ный параметр для сле­дующей фун­кции `Reverse`.

```asm
mov  rcx, rax ; s
call Reverse(char *)
; Определяем нулевое смещение
mov  eax, 1
imul rax, 0
```

Заг­ружа­ем в `RAX` ука­затель на перемен­ную `dst`. Пос­коль­ку зна­чение, воз­вра­щен­ное фун­кци­ей через регистр `RAX`, не исполь­зует­ся, мож­но пред­положить, что она воз­вра­щает его не через регистр, а в перемен­ной `dst`. Но это пока не более чем пред­положе­ние.

```asm
lea rax, [rsp+rax+0C8h+dst]
mov rdx, rax
lea rcx, aS_0 ; "%s\n"
; Передача функции printf форматной строки и указателя на переменную dst
; Собственно вызов функции печати
call printf
```

Сле­дующи­ми дву­мя инс­трук­циями в регис­тры `ECX` и `EDX` для переда­чи фун­кции `Sum` в качес­тве парамет­ров заг­ружа­ем зна­чения типа `int`.

```asm
mov edx, 777h ; b
mov ecx, 666h ; a
; Вызов самой функции Sum
call sum(int,int)
```

Из­вле­каем 32-бит­ный резуль­тат выпол­нения фун­кции из регис­тра `EAX` и раз­меща­ем его в `EDX` для переда­чи в качес­тве вто­рого парамет­ра фун­кции `printf`.

```asm
mov   edx, eax
; В RCX помещается форматная строка
lea   rcx, asc_140002268        ; "%x\n"
; Вызов функции печати
call  printf
xor   eax, eax
mov   rcx, [rsp+0C8h+var_28]
xor   rcx, rsp  ; StackCookie
call  __security_check_cookie
; Восстанавливаем регистр
add   rsp, 0B8h
; Закрываем кадр стека
pop   rdi
pop   rsi
retn
main  endp

; Функция Reverse с двумя параметрами (прототип функции определен правильно)
void  Reverse(char *, char const *) proc near
Dst = qword ptr 8
Src = qword ptr 10h
; Размещение параметров в памяти
mov   [rsp+Src], rdx
mov   [rsp+Dst], rcx
; Выделение памяти под локальные переменные
sub   rsp, 28h
; Подготавливаем параметры для вызова strcpy_s:
; В R8 поместим указатель на буфер-источник
mov   r8,  [rsp+28h+Src]        ; Src
; В EDX — количество байтов для копирования — 0x64 (100)
mov   edx, 64h  ; 'd' ; SizeInBytes
; В RCX — указатель на буфер-приемник
mov   rcx, [rsp+28h+Dst]        ; Dst
; Собственно вызов strcpy_s
call  cs:__imp_strcpy_s
; Возвращенный результат в переменной Dst копируем в регистр RCX для передачи функции
; strrev в качестве параметра
mov   rcx, [rsp+28h+Dst]        ; Str
; Вызов функции strrev, она инвертирует переданную в параметре строку
call  cs:__imp__strrev
; Восстановление регистра
add   rsp, 28h
; В итоге, очевидно, возвращение значения выполняется в параметре Dst
retn
```

Ни­ког­да не пре­неб­регай ква­лифи­като­ром `const`: он сви­детель­ству­ет, что перемен­ная, на которую ссы­лает­ся дан­ный ука­затель, исполь­зует­ся лишь для чте­ния. Эта информа­ция зна­читель­но облегчит работу с дизас­сем­блер­ным лис­тингом, осо­бен­но ког­да ты вер­нешь­ся к нему спус­тя некото­рое вре­мя, осно­ватель­но под­забыв алго­ритм иссле­дуемой прог­раммы.

```asm
void Reverse(char *, char const *) endp
; Функция Reverse с одним параметром (прототип функции определен правильно)
void Reverse(char *) proc near
Str   = qword ptr 8
; Копируем параметр из регистра в память
mov   [rsp+Str], rcx
; Выделяем память для локальной переменной
sub   rsp, 28h
; В RCX помещаем указатель на переменную Str (буфер в памяти) —
; готовим параметр для вызова функции strrev
mov   rcx, [rsp+28h+Str]    ; Str
; Вызов функции strrev
call  cs:__imp__strrev
; Восстанавливаем регистр
add   rsp, 28h
; В итоге, очевидно, возвращение значения выполняется в параметре Str
retn
void Reverse(char *) endp
; Прототип функции определен правильно, функция sum получает два числа
; и возвращает одно
int   sum(int, int) proc near
arg_0 = dword ptr 8
arg_8 = dword ptr 10h
; Параметры из регистров копируем в память
mov   [rsp+arg_8],        edx
mov   [rsp+arg_0],        ecx
; Возвращаем их в более удобные регистры
mov   eax, [rsp+arg_8]
mov   ecx, [rsp+arg_0]
; Суммируем их
add   ecx, eax
mov   eax, ecx
```

Ко­пиру­ем резуль­тат сло­жения `arg_0` и `arg_8` обратно в `arg_0`. Неопыт­ные хакеры могут при­нять это за воз­вра­щение зна­чения через аргу­мент, одна­ко такое пред­положе­ние невер­но. Дело в том, что передан­ные фун­кции парамет­ры пос­ле ее завер­шения вытал­кива­ются из сте­ка и тут же «погиба­ют». Не забывай: парамет­ры, передан­ные по зна­чению, ведут себя так же, как и локаль­ные перемен­ные.

```asm
mov [rsp+arg_0], eax
; А вот сейчас в регистр EAX действительно копируется возвращаемое значение
mov eax, [rsp+arg_0]
; Следовательно, прототип функции выглядит так: int Sum(int a, int b);
retn
int sum(int, int) endp
```

## ВОЗВРАТ ЗНАЧЕНИЙ ЧЕРЕЗ ДИНАМИЧЕСКУЮ ПАМЯТЬ (КУЧУ)

Воз­вра­щение зна­чения через аргу­мент, передан­ный по ссыл­ке, не очень‑то укра­шает про­тотип фун­кции. Он вмиг перес­тает быть инту­итив­но понят­ным и тре­бует раз­верну­тых пояс­нений, что с этим аргу­мен­том ничего переда­вать не надо, нап­ротив, будь готов отсю­да при­нять. Но хвост с ней, с наг­лядностью и эсте­тикой (кто говорил, что быть прог­раммис­том лег­ко?), сущес­тву­ет и более серь­езная проб­лема — далеко не во всех слу­чаях раз­мер воз­вра­щаемых дан­ных известен наперед, час­тень­ко он выяс­няет­ся лишь в про­цес­се работы вызыва­емой фун­кции.

Вы­делить буфер с запасом? Нек­расиво и неэко­номич­но, даже в сис­темах с вир­туаль­ной памятью ее объ­ем небез­гра­ничен. Вот если бы вызыва­емая фун­кция самос­тоятель­но выделя­ла для себя память, как раз по пот­ребнос­ти, а потом воз­вра­щала на нее ука­затель... Ска­зано — сде­лано! Ошиб­ка мно­гих начина­ющих прог­раммис­тов как раз и зак­люча­ется в попыт­ке вер­нуть ука­затель на локаль­ные перемен­ные. Увы, они «уми­рают» вмес­те с завер­шени­ем фун­кции, и ука­затель ука­зыва­ет в «кос­мос». Пра­виль­ное решение зак­люча­ется в выделе­нии памяти из кучи (динами­чес­кой памяти), ска­жем, вызовом `malloc` или `new`, эта память «живет» вплоть до ее при­нуди­тель­ного осво­бож­дения фун­кци­ей `free` или `delete` соот­ветс­твен­но.

Для ана­лиза прог­раммы механизм выделе­ния памяти не сущес­твен, основную роль игра­ет тип воз­вра­щаемо­го зна­чения. Отли­чить ука­затель от осталь­ных типов дос­таточ­но лег­ко — толь­ко ука­затель может исполь­зовать­ся в качес­тве поад­ресно­го выраже­ния.

Раз­берем сле­дующий при­мер, демонс­три­рующий воз­вра­щение зна­чения через кучу:

```cpp
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

char* MyFunc(int a)
{
    char* x;
    x = (char*)malloc(100);
    _itoa_s(a, x, sizeof(x), 16);
    return x;
}

int main()
{
    char* x;
    x = MyFunc(0x666);
    printf("0x%s\n", x);
    free(x);
}
```

Ди­зас­сем­бли­рован­ный лис­тинг будет выг­лядеть так:

```asm
main proc near ; CODE XREF: __scrt_common_main_seh+107↓p
               ; DATA XREF: .pdata:0000000140004024↓o
Memory = qword ptr -18h
; Резервируем память для локальной переменной
sub    rsp, 38h
; Передаем функции MyFunc значение 0x666 типа int
mov    ecx, 666h ; a
```

Вы­зыва­ем `MyFunc`. Обра­ти вни­мание, что ни один аргу­мент не был передан фун­кции по ссыл­ке!

```cpp
call MyFunc(int)
```

```asm
; Копирование содержимого возвращенного функцией значения в переменную Memory
mov [rsp+38h+Memory], rax
; Супер! Загружаем в RDX возвращенное функцией значение
mov rdx, [rsp+38h+Memory]
```

Та­ким обра­зом переда­ем воз­вра­щен­ное фун­кци­ей `MyFunc` зна­чение фун­кции `printf`. Судя по спе­цифи­като­ру, тип воз­вра­щен­ного зна­чения — `char *`. Пос­коль­ку фун­кции `MyFunc` ни один из парамет­ров не переда­вал­ся по ссыл­ке, она явно выдели­ла память самос­тоятель­но и записа­ла туда получен­ную стро­ку.

А если бы фун­кции `MyFunc` переда­вались один или более парамет­ров по ссыл­ке? Тог­да не было бы никакой уве­рен­ности, что она не воз­вра­тила один из таких парамет­ров обратно, пред­варитель­но его модифи­циро­вав. Впро­чем, модифи­кация необя­затель­на. Нап­ример, переда­ем фун­кции ука­зате­ли на две стро­ки, и она воз­вра­щает ука­затель на ту из них, которая, ска­жем, короче или содер­жит боль­ше глас­ных букв. Поэто­му не вся­кое воз­вра­щение ука­зате­ля сви­детель­ству­ет о модифи­кации.

```asm
lea rcx, _Format ; "0x%s\n"
; Вызов printf — вывод на экран строки, возвращенной функцией MyFunc
call printf
```

Пе­реда­ем фун­кции `free` ука­затель, воз­вра­щен­ный фун­кци­ей `MyFunc`. Зна­чит, `MyFunc` дей­стви­тель­но самос­тоятель­но выдели­ла память вызовом `malloc`.

```asm
mov   rcx, [rsp+38h+Memory]    ; Memory
; Освобождаем память, выделенную MyFunc, для возвращения значения
call  cs:__imp_free
xor   eax, eax
; Восстанавливаем регистр
add   rsp, 38h
retn
; Таким образом, прототип MyFunc выглядит так: char* MyFunc(int a)
main  endp

char  *    MyFunc(int)         ; CODE XREF: main+9↓p

; DATA XREF: .pdata:0000000140004018↓o
DstBuf = qword ptr -18h
Value  = dword ptr 8
; Копируем параметр из регистра ECX в находящуюся в памяти переменную Value
mov   [rsp+Value],     ecx
; Резервируем память для локальных переменных
sub   rsp, 38h
; Готовим параметр для вызова функции, помещая его в регистр ECX
mov   ecx, 64h     ; 'd' ; Size
```

Вы­деля­ем `0x64` (100) байт памяти из кучи либо для собс­твен­ных нужд фун­кции, либо для воз­вра­щения резуль­тата. Пос­коль­ку из ана­лиза кода вызыва­ющей фун­кции нам уже извес­тно, что `MyFunc` воз­вра­щает ука­затель, очень веро­ятно, что вызов `malloc` выделя­ет память как раз для этой цели. Впро­чем, вызовов `malloc` может быть и нес­коль­ко, а ука­затель воз­вра­щает­ся толь­ко на один из них.

```asm
call  cs:__imp_malloc
; Запоминаем указатель в локальной переменной DstBuf
mov   [rsp+38h+DstBuf],    rax
; Готовим четыре параметра для функции _itoa_s:
; Крайний справа — 0x10 (16) — требуемая система исчисления для перевода числа
mov   r9d, 10h ; Radix
; Второй справа — размер буфера приема (строки)
mov   r8d, 8   ; Size
; Третий справа — указатель на буфер приема
mov   rdx, [rsp+38h+DstBuf]    ; DstBuf
; Последний параметр — значение типа int для перевода
mov   ecx, [rsp+38h+Value]     ; Value
```

Фун­кция `_itoa_s` перево­дит чис­ло в стро­ку и записы­вает ее в буфер по передан­ному ука­зате­лю.

```asm
call cs:__imp__itoa_s
```

Воз­вра­щаем ука­затель на реги­он памяти, выделен­ный самой `MyFunc` из кучи и содер­жащий резуль­тат работы `_itoa_s`.

```asm
mov rax, [rsp+38h+DstBuf]
; Восстанавливаем регистр
add rsp, 38h
retn
char * MyFunc(int) endp
```

## ВОЗВРАТ ЗНАЧЕНИЙ ЧЕРЕЗ ГЛОБАЛЬНЫЕ ПЕРЕМЕННЫЕ

«Мыль­ную опе­ру» перепе­вов с воз­вра­щени­ем ука­зате­лей про­дол­жает серия «Воз­вра­щение зна­чений через гло­баль­ные перемен­ные (и/или ука­зате­ля на гло­баль­ные перемен­ные)». Вооб­ще‑то гло­баль­ные перемен­ные — дур­ной тон, и такой стиль прог­рамми­рова­ния харак­терен в основном для прог­раммис­тов с мыш­лени­ем, необ­ратимо иска­лечен­ным иде­оло­гией бацика с его недораз­витым механиз­мом вызова под­прог­рамм.

Рань­ше мы уже под­робно обсужда­ли иден­тифика­цию гло­баль­ных перемен­ных, здесь же мы сос­редото­чим наши уси­лия имен­но на изу­чении механиз­мов воз­вра­щения зна­чений через гло­баль­ные перемен­ные.

Фак­тичес­ки все гло­баль­ные перемен­ные мож­но рас­смат­ривать как неяв­ные аргу­мен­ты каж­дой вызыва­емой фун­кции и в то же вре­мя как воз­вра­щаемые зна­чения. Любая фун­кция может про­изволь­ным обра­зом читать и модифи­циро­вать их, при­чем ни «переда­ча», ни «воз­вра­щение» гло­баль­ных перемен­ных не замет­ны при ана­лизе кода вызыва­ющей фун­кции. Для это­го необ­ходимо тща­тель­но иссле­довать вызыва­емую, манипу­лиру­ет ли она гло­баль­ными перемен­ными, и если да, то какими. Мож­но зай­ти и с обратной сто­роны: прос­мотром сег­мента дан­ных най­ти все гло­баль­ные перемен­ные, опре­делить их сме­щение и, прой­дясь кон­текс­тным поис­ком по все­му фай­лу, выявить фун­кции, которые на них ссы­лают­ся.

По­мимо гло­баль­ных, сущес­тву­ют еще и ста­тичес­кие перемен­ные. Они так­же рас­полага­ются в сег­менте дан­ных, но непос­редс­твен­но дос­тупны толь­ко объ­явив­шей их фун­кции. Точ­нее, огра­ниче­ние наложе­но не на сами перемен­ные, а на их име­на. Что­бы пре­дос­тавить дру­гим фун­кци­ям дос­туп к собс­твен­ным ста­тичес­ким перемен­ным, дос­таточ­но передать ука­затель. К счастью, этот трюк не соз­дает хакерам никаких проб­лем (хоть некото­рые зло­пыха­тели и объ­явля­ют его «про­рехой в защите»). Отсутс­твие непос­редс­твен­ного дос­тупа к «чужим» ста­тичес­ким перемен­ным и необ­ходимость вза­имо­дей­ство­вать с фун­кци­ей‑вла­дели­цей через пред­ска­зуемый интерфейс (воз­вра­щен­ный ука­затель) поз­воля­ет раз­бить прог­рамму на отдель­ные незави­симые модули, каж­дый из которых может быть про­ана­лизи­рован отдель­но. Что­бы не быть голос­ловным, про­демонс­три­руем это на сле­дующем при­мере, который иллюс­три­рует воз­врат зна­чения через гло­баль­ные ста­тичес­кие перемен­ные:

```cpp
#include <stdio.h>
char* MyFunc(int a)
{
    static char x[7][16] = { "Monday", "Tuesday", "Wednesday",
        "Thursday", "Friday", "Saturday", "Sunday" };

    return &x[a - 1][0];
}

int main()
{
    printf("%s\n", MyFunc(6));
}
```

| ![Вы­вод прог­раммы](./images/16/d8249810107730b981ad5143411a7c21.png)
|:--:|
| *Вы­вод прог­раммы* |

Ре­зуль­тат ком­пиляции ком­пилято­ром Visual C++ 2019 с вык­лючен­ной опти­миза­цией выг­лядит так:

```asm
char  * MyFunc(int) proc near
arg_0 = dword ptr 8
; Загружаем в аргумент arg_0 значение из регистра ECX,
; переданное из вызывающей функции
mov   [rsp+arg_0], ecx
; Помещаем это значение в регистр EAX
mov   eax, [rsp+arg_0]
```

Умень­шаем `EAX` на еди­ницу. Это кос­венно сви­детель­ству­ет о том, что `arg_0` не ука­затель, хотя матема­тичес­кие опе­рации над ука­зате­лями в C раз­решены и активно исполь­зуют­ся.

```asm
dec eax
; Расширение значения из двойного слова в четверное
cdqe
; Умножение помещенного в RAX значения (arg_0-1) на 0x10 (16)
imul rax, 10h
; В регистр RCX помещаем указатель на строку "Monday" из таблицы строк,
; расположенной в сегменте данных
lea rcx, x ; "Monday"
```

При­бав­ляем к базово­му ука­зате­лю на таб­лицу строк (`RCX`) зна­чение, получен­ное путем умно­жения `arg_0-1` на `0x10` (`RAX`). Таб­лица строк рас­полага­ется в сег­менте дан­ных. А в нем находят­ся либо ста­тичес­кие, либо гло­баль­ные перемен­ные. Пос­коль­ку зна­чение аргу­мен­та `arg_0` умно­жает­ся на некото­рую величи­ну (в дан­ном слу­чае на 16), мож­но пред­положить, что мы име­ем дело с дву­мер­ным мас­сивом. В дан­ном слу­чае — мас­сивом строк фик­сирован­ной дли­ны. Таким обра­зом, в `RAX` содер­жится ука­затель на стро­ку с индексом `arg_0-1`. Или, дру­гими сло­вами, с индексом `arg_0`, счи­тая с одно­го.

```asm
add  rcx, rax
; Перемещаем значение, чтобы освободить регистр RCX для следующей операции
mov  rax, rcx
; Определяем смещение в массиве
mov  ecx, 1
imul rcx, 0
```

Сле­дующее сло­жение не меня­ет зна­чение регис­тра — воз­вра­щает­ся ука­затель на соот­ветс­тву­ющий эле­мент мас­сива (с индексом `arg_0-1`). Как мы видим, нет никакой прин­ципи­аль­ной раз­ницы меж­ду воз­вра­щени­ем ука­зате­ля на реги­он памяти, выделен­ный из кучи, и воз­вра­щени­ем ука­зате­ля на ста­тичес­кие перемен­ные, рас­положен­ные в сег­менте дан­ных.

```asm
add rax, rcx
retn
char * MyFunc(int) endp

main proc near
sub   rsp, 28h
; Передаем функции MyFunc значение типа int (шестой день — Saturday)
mov   ecx, 6 ; a
; Вызываем MyFunc
call  MyFunc(int)
```

Пе­реда­ем воз­вра­щен­ное `MyFunc` зна­чение фун­кции `printf`. Судя по стро­ке спе­цифи­като­ров, это ука­затель на стро­ку.

```asm
mov  rdx, rax
lea  rcx, _Format ; "%s\n"
call printf
xor  eax, eax
add  rsp, 28h
retn
main endp
```

Сег­мент дан­ных:

```asm
.data:0000000140003040 ; char x[7][16]
; Наличие перекрестной ссылки только на одну функцию подсказывает, что тип
; этой переменной — static
.data:0000000140003040 x db 'Monday',0 ; DATA XREF: MyFunc(int)+10↑o
.data:0000000140003047 align 10h
.data:0000000140003050 aTuesday db 'Tuesday',0
.data:0000000140003058 align 20h
.data:0000000140003060 aWednesday db 'Wednesday',0
.data:000000014000306A align 10h
.data:0000000140003070 aThursday db 'Thursday',0
.data:0000000140003079 align 20h
.data:0000000140003080 aFriday db 'Friday',0
.data:0000000140003087 align 10h
.data:0000000140003090 aSaturday db 'Saturday',0
.data:0000000140003099 align 20h
.data:00000001400030A0 aSunday db 'Sunday',0
```

А теперь срав­ним пре­дыду­щий при­мер с нас­тоящи­ми гло­баль­ными перемен­ными:

```cpp
#include <stdio.h>

int a;
int b;
int c;

void Sum()
{
    c = a + b;
}

void MyFunc()
{
    a = 0x666;
    b = 0x777;
    Sum();
}

int main()
{
    MyFunc();
    printf("%x\n", c);
}
```

Ди­зас­сем­блер­ный лис­тинг:

```asm
main proc near
; Резервируем память для локальных переменных
sub rsp, 28h
; Вызываем MyFunc. Обрати внимание: функции явно ничего
; не передается и ничего не возвращается
call MyFunc(void)
```

Заг­ружа­ем в `EDX` зна­чение гло­баль­ной перемен­ной `c`. Смот­рим в сег­мент дан­ных: так‑так, вот она, перемен­ная `c`, рав­ная неоп­ределен­ному зна­чению. Одна­ко это­му зна­чению нель­зя доверять — быть может, ее уже успе­ли изме­нить ранее выз­ванные фун­кции. Пред­положе­ние о модифи­кации под­креп­ляет­ся парой перек­рес­тных ссы­лок, одна из которых ука­зыва­ет на фун­кцию `Sum` (см. ниже). Суф­фикс `w`, завер­шающий перек­рес­тную ссыл­ку, говорит о том, что `Sum` записы­вает в перемен­ную c какое‑то зна­чение. Какое? Это мож­но узнать из ана­лиза кода самой `Sum`.

```asm
mov edx, cs:int c
```

Пе­реда­ем воз­вра­щен­ное фун­кци­ей `Sum` зна­чение через гло­баль­ную перемен­ную c фун­кции `printf`. Судя по стро­ке спе­цифи­като­ров, аргу­мент име­ет тип `int`.

```asm
lea   rcx, _Format ; "%x\n"
; Выводим возвращенный Sum результат на терминал
call  printf
xor   eax, eax
; Восстанавливаем регистр
add   rsp, 28h
retn
main  endp
void  MyFunc(void) proc near
; Резервируем память для локальных переменных
sub   rsp, 28h
```

Прис­ваиваем перемен­ным `a` и `b` зна­чения `0x666` и `0x777`. Ука­зание сег­мента `cs` перед типом и име­нем перемен­ных говорит о том, что они находят­ся в сег­менте дан­ных.

```asm
mov   cs:int a, 666h
mov   cs:int b, 777h
; Вызываем функцию Sum, явно не принимающую и не возвращающую никаких значений
call  Sum(void)
; Восстанавливаем регистр
add   rsp, 28h
retn
void  MyFunc(void) endp
void  Sum(void) proc near
```

Заг­ружа­ем в `EAX` зна­чение гло­баль­ной перемен­ной `b`. Находим `b` в сег­менте дан­ных: ага, есть перек­рес­тная ссыл­ка на `MyFunc`, которая что‑то записы­вает в перемен­ную `b`. Пос­коль­ку вызов `MyFunc` пред­шес­тво­вал вызову `Sum`, мож­но ска­зать, что `MyFunc` помес­тила в `b` некото­рое зна­чение.

```asm
mov eax, cs:int b
```

Заг­ружа­ем в `ECX` зна­чение гло­баль­ной перемен­ной `a`. Находим `a` в сег­менте дан­ных, так­же есть перек­рес­тная ссыл­ка на `MyFunc`, которая, воз­можно, что‑то записы­вает в перемен­ную `a`.

```asm
mov ecx, cs:int a
```

Скла­дыва­ем `EAX` (хра­нящий зна­чение гло­баль­ной перемен­ной `a`) с `ECX` (хра­нящий зна­чение гло­баль­ной перемен­ной `b`) и записы­ваем резуль­тат в регистр `ECX`.

```asm
add ecx, eax
mov eax, ecx
; Помещаем результат сложения a + b в глобальную переменную c
mov cs:int c, eax
retn
void Sum(void) endp
```

Те­перь мы смог­ли разоб­рать­ся, как три фун­кции вза­имо­дей­ству­ют: `main()` вызыва­ет `MyFunc()`, та ини­циали­зиру­ет гло­баль­ные перемен­ные `a` и `b`. Затем уже `MyFunc()` вызыва­ет `Sum()`, помеща­ющую сум­му `a` и `b` в гло­баль­ную перемен­ную `c`. Наконец, `main()` берет эту перемен­ную `c` и переда­ет ее пос­редс­твом регис­тра фун­кции `printf` для вывода на экран.

Уф! Как все запута­но, а ведь это прос­тей­ший при­мер из трех фун­кций! Что же говорить о реаль­ной прог­рамме, в которой этих фун­кций тысячи, при­чем порядок вызова и поведе­ние каж­дой из них далеко не так оче­вид­ны!

Это еще раз доказы­вает, что зло­упот­ребле­ние гло­баль­ными перемен­ными прив­носит хаос.

Сег­мент дан­ных:

```asm
...
.data:000000014000362C ; int b
.data:000000014000362C int b dd ? ; DATA XREF: Sum(void)↑r
.data:000000014000362C ; MyFunc(void)+E↑w
.data:0000000140003630 ; int a
.data:0000000140003630 int a dd ? ; DATA XREF: Sum(void)+6↑r
.data:0000000140003630 ; MyFunc(void)+4↑w
.data:0000000140003634 ; int c
.data:0000000140003634 int c dd ? ; DATA XREF: Sum(void)+10↑w
.data:0000000140003634 ; main+9↑r
...
```

Су­дя по перек­рес­тным ссыл­кам, все три перемен­ные — гло­баль­ные, так как к каж­дой из них име­ет непос­редс­твен­ный дос­туп более одной фун­кции.

## ВОЗВРАТ ЗНАЧЕНИЙ ЧЕРЕЗ ФЛАГИ ПРОЦЕССОРА

Для боль­шинс­тва ассем­блер­ных фун­кций харак­терно исполь­зование регис­тра фла­гов про­цес­сора для воз­вра­щения резуль­тата успешнос­ти выпол­нения фун­кции. По общепри­нято­му сог­лашению уста­нов­ленный флаг перено­са (`CF`) сви­детель­ству­ет об ошиб­ке, вто­рое мес­то по популяр­ности занима­ет флаг нуля (`ZF`), а осталь­ные фла­ги прак­тичес­ки вооб­ще не исполь­зуют­ся.

Ус­танов­ка фла­га перено­са выпол­няет­ся коман­дой `STC` или любой матема­тичес­кой опе­раци­ей, при­водя­щей к обра­зова­нию перено­са (нап­ример, `CMP a, b`, где `a` < `b`), а сброс — коман­дой `CLC` или соот­ветс­тву­ющей матема­тичес­кой опе­раци­ей.

Флаг перено­са про­веря­ется обыч­но через условные перехо­ды `JC xxx` и `JNC xxx`, соот­ветс­твен­но исполня­ющиеся при наличии и отсутс­твии перено­са. Условные перехо­ды `JB xxx` и `JNB xxx` — их син­такси­чес­кие синони­мы, дающие при ассем­бли­рова­нии иден­тичный код.

```cpp
#include <stdio.h>

// Функция сообщения об ошибке деления
void Err()
{
    printf("-ERR: DIV by Zero\n");
}

// Вывод результата деления на экран
void Ok(int a)
{
    printf("%x\n", a);
}
```

Ни­же пред­став­лена ассем­блер­ная фун­кция деления. Делит `EAX` на `EBX`, воз­вра­щая час­тное в `EAX`, а оста­ток — в `EDX`. При попыт­ке деления на ноль уста­нав­лива­ет флаг перено­са.

```asm
__declspec(naked) int MyFunc()
{
    __asm {
xor  edx, edx; Обнуляем EDX, то есть команда div ожидает делимого в EAX
test ebx, ebx; Проверка делителя на равенство нулю
jz   _err ; Если делитель равен нулю, перейти к ветке _err
div  ebx ; Делим EAX на EBX (EBX заведомо не равен нулю)
ret
_err : ; Эта ветка получает управление при попытке деления на ноль
stc; Устанавливаем флаг переноса, сигнализируя об ошибке и...
ret; выходим
    }
}
```

А вот обер­тка для `MyFunc`. При­нима­ет два аргу­мен­та через стек: делимое и делитель. Выводит резуль­тат деления (или сооб­щение об ошиб­ке) на экран.

```cpp
__declspec(naked) void MyFunc_2(int a, int b)
{
    __asm
    {

mov   eax, [esp + 4]; Загружаем в EAX содержимое аргумента a
mov   ebx, [esp + 8]; Загружаем в EBX содержимое аргумента b
call  MyFunc ; Пытаемся делить a/b
jnc   _ok    ; Если флаг переноса сброшен, выводим результат, иначе...
call  Err    ; сообщение об ошибке.
ret          ; Возвращаемся
_ok :
push  eax    ; Передаем результат деления и...
call  Ok     ; выводим его на экран
add   esp, 4 ; Вычищаем за собой стек
ret          ; Возвращаемся
    }
}

int main()
{
    MyFunc_2(4, 0);
}
```

| ![Ре­зуль­тат работы прог­раммы](./images/16/e32b14fcb4ae964ce8ef8cc1a841c54f.png)
|:--:|
| *Ре­зуль­тат работы прог­раммы* |

## ВЫВОДЫ

Это было неп­росто — за один раз нас­коком охва­тить все спо­собы, которы­ми выз­ванная фун­кция может воз­вра­щать резуль­таты. Но мы спра­вились! Мы под­робно обсу­дили стан­дар­тные спо­собы воз­вра­та как целочис­ленных, так и вещес­твен­ных зна­чений опе­рато­ром `return`, уви­дели фун­дамен­таль­ную раз­ницу меж­ду ста­тичес­кими и гло­баль­ными перемен­ными. Заг­лянув под капот ком­пилято­ра, рас­смот­рели, каким обра­зом он воз­вра­щает огромные резуль­таты в куче. Так­же мы обсу­дили ассем­блер­ные inline-фун­кции, как с их помощью мож­но опти­мизи­ровать вычис­ления и вер­нуть получив­ший­ся резуль­тат. А в самом кон­це мы рас­смот­рели вари­ант переда­чи и воз­вра­та зна­чения пос­редс­твом фла­гов цен­траль­ного про­цес­сора.

В сле­дующей статье мы раз­берем не менее увле­катель­ную тему — как опе­раци­онная сис­тема выделя­ет память в момент запус­ка прог­раммы.
