# Ищем переменные разных типов в чужих программах

Вокруг переменных и с их непосредственным участием происходят главные события в любой программе. Поэтому каждый кодокопатель должен уметь находить переменные и определять их тип. Именно этим мы сегодня и займемся.

«Тип переменной» в рассматриваемом нами случае не имеет отношения к типу данных, имеется в виду совокупность способов работы с переменными со стороны программиста и компилятора, период жизни переменных и места их обитания. В идеальном случае для восстановления алгоритма исследуемой программы неплохо выяснить тип данных переменных, но это уже вопрос соотношения времени, трудозатрат и выгоды от полученной информации.

## Идентификация локальных стековых переменных

Локальные переменные размещаются в стеке (его также называют автоматической памятью), а удаляет их оттуда вызываемая функция, когда она завершится. Рассмотрим подробнее, как это происходит. Сначала в стек затягиваются аргументы, передаваемые функции (если они есть), а сверху на них кладется адрес возврата, помещаемый туда инструкцией `CALL`, которая вызывает эту функцию. Получив управление, функция открывает кадр стека — сохраняет прежнее значение регистра `RBP` и устанавливает его равным регистру `RSP` (регистр — указатель вершины стека). «Выше» (то есть в более младших адресах) `RBP` находится свободная область стека, «ниже» — служебные данные (сохраненный `RBP`, адрес возврата) и аргументы.

Сохранность области стека, расположенной выше указателя вершины стека (регистра `RSP`), не гарантирована от затирания и искажения. Ее беспрепятственно могут использовать, например, обработчики аппаратных прерываний, вызываемые в непредсказуемом месте в непредсказуемое время. Да и использование стека самой функцией (для сохранения регистров или передачи аргументов) приведет к его искажению. Какой выход? Принудительно переместить указатель вершины стека вверх, тем самым заняв данную область стека. Сохранность памяти, находящейся «ниже» `RSP`, гарантируется (имеется в виду гарантируется от непреднамеренных искажений). Очередной вызов инструкции `PUSH` занесет данные на вершину стека, не затирая локальные переменные.

По окончании работы функция обязана вернуть `RSP` на прежнее место, иначе инструкция `RET` снимет со стека отнюдь не адрес возврата, а вообще невесть что (значение самой «верхней» локальной переменной) и передаст управление «в космос»...

| ![Механизм размещения локальных переменных в стеке](./images/17/stack_local_vars.png)
|:--:|
| *Механизм размещения локальных переменных в стеке* |

На левой картинке показано состояние стека на момент вызова функции. Она открывает кадр стека, сохраняя прежнее значение регистра `RBP`, и устанавливает его равным `RSP`. На правой картинке изображено резервирование `0x14` байт стековой памяти под локальные переменные. Резервирование достигается перемещением регистра `RSP` «вверх» — в область младших адресов. Фактически локальные переменные размещаются в стеке так, как будто бы они были туда помещены командой `PUSH`. При завершении работы функция увеличивает значение регистра `RSP`, возвращая его на прежнюю позицию и освобождая тем самым память, занятую локальными переменными. Затем она стягивает со стека и восстанавливает значение `RBP`, закрывая тем самым кадр стека.

### Адресация локальных переменных

Адресация локальных переменных очень похожа на адресацию стековых аргументов, только аргументы располагаются «ниже» `RBP`, а локальные переменные — «выше». Другими словами, аргументы имеют положительные смещения относительно `RBP`, а локальные переменные — отрицательные, поэтому их очень легко отличить друг от друга. Например, `[RBP+xxx]` — аргумент, а `[RBP-xxx]` — локальная переменная.

Регистр — указатель кадра стека служит как бы барьером: по одну сторону от него аргументы функции, по другую — локальные переменные.

| ![Адресация локальных переменных](./images/17/local_vars_adr.png)
|:--:|
| *Адресация локальных переменных* |

Теперь понятно, почему при открытии кадра стека значение `RSP` копируется в `RBP`, иначе адресация локальных переменных и аргументов значительно усложнилась бы, а разработчики компиляторов (как это ни странно) тоже люди и не хотят без нужды осложнять себе жизнь. Впрочем, оптимизирующие компиляторы умеют адресовать локальные переменные и аргументы непосредственно через `RSP`, освобождая регистр `RBP` для более важных целей.

### Детали технической реализации

Существует множество вариантов реализации выделения и освобождения памяти под локальные переменные. Казалось бы, чем плохо очевидное `SUB` `RSP,xxx` на входе и `ADD` `RSP,` `xxx` на выходе? А вот некоторые компиляторы в стремлении отличаться ото всех остальных резервируют память не уменьшением, а увеличением `RSP`... Да, на отрицательное число, которое по умолчанию большинством дизассемблеров отображается как очень большое положительное. Оптимизирующие компиляторы при отводе небольшого количества памяти изменяют `SUB` на `PUSH` `reg`, что на несколько байтов короче. Последнее создает очевидные проблемы идентификации: попробуй разберись, то ли перед нами сохранение регистров в стеке, то ли передача аргументов, то ли резервирование памяти для локальных переменных (подробнее об этом см. раздел «Идентификация механизма выделения памяти»).

Алгоритм освобождения памяти также неоднозначен. Помимо увеличения регистра указателя вершины стека инструкцией `ADD` `RSP,` `xxx` (или в особо извращенных компиляторах увеличения его на отрицательное число), часто встречается конструкция `MOV` `RSP,` `RBP`. Мы ведь помним, что при открытии кадра стека `RSP` копировался в `RBP`, а сам `RBP` в процессе исполнения функции не изменялся. Наконец, память может быть освобождена инструкцией `POP`, выталкивающей локальные переменные одну за другой в какой-нибудь ненужный регистр (понятное дело, такой способ оправдывает себя лишь на небольшом количестве локальных переменных).

| ![Наиболее распространенные варианты реализации резервирования памяти под локальные переменные и ее освобождение](./images/17/local_vars_adr.png)
|:--:|
| *Наиболее распространенные варианты реализации резервирования памяти под локальные переменные и ее освобождение* |

### Идентификация механизма выделения памяти

Выделение памяти инструкциями `SUB` и `ADD` непротиворечиво и всегда интерпретируется однозначно. Если же оно выполняется командой `PUSH`, а освобождение — `POP`, эта конструкция становится неотличима от простого освобождения/сохранения регистров в стеке. Ситуация серьезно осложняется тем, что в функции присутствуют и настоящие команды сохранения регистров, сливаясь с командами выделения памяти. Как узнать, сколько байтов резервируется для локальных переменных и резервируются ли они вообще (может, в функции локальных переменных и нет вовсе)?

Ответить на этот вопрос позволяет поиск обращений к ячейкам памяти, лежащих «выше» регистра `RBP`, то есть с отрицательными относительными смещениями. Рассмотрим два примера.

![two_samples](./images/17/two_samples.png)

В левом из них никакого обращения к локальным переменным не происходит вообще, а в правом наличествует конструкция `MOV` `[RBP-4],0x666`, копирующая значение `0x666` в локальную переменную `var_4`. А раз есть локальная переменная, для нее кем-то должна быть выделена память. Поскольку инструкций `SUB` `RSP,` `xxx` и `ADD` `RSP,` `–` `xxx` в теле функций не наблюдается, подозрение падает на `PUSH` `RCX`, так как сохраненное содержимое регистра `RCX` располагается в стеке на четыре байта «выше» `RBP`. В данном случае подозревается лишь одна команда — `PUSH` `RCX`, поскольку `PUSH` `RBP` на роль «резерватора» не тянет. Но как быть, если подозреваемых несколько?
Определить количество выделенной памяти можно по смещению самой «высокой» локальной переменной, которую удается обнаружить в теле функции. То есть, отыскав все выражения типа `[RBP-xxx]`, выберем наибольшее смещение «xxx» — в общем случае оно равно количеству байтов выделенной под локальные переменные памяти. В частностях же встречаются объявленные, но не используемые локальные переменные. Им выделяется память (хотя оптимизирующие компиляторы просто выкидывают такие переменные за ненадобностью), но ни одного обращения к ним не происходит, и описанный выше алгоритм подсчета объема резервируемой памяти дает заниженный результат. Впрочем, эта ошибка никак не сказывается на результатах анализа программы.

### Инициализация локальных переменных

Существует два способа инициализации локальных переменных: присвоение необходимого значения инструкцией `MOV` (например, `MOV` `[RBP-04],` `0x666`) и непосредственное заталкивание значения в стек инструкцией `PUSH` (например, `PUSH` `0x777`). Последнее позволяет выгодно комбинировать выделение памяти под локальные переменные с их инициализацией (разумеется, только в том случае, если этих переменных немного).

Популярные компиляторы в подавляющем большинстве случаев выполняют операцию инициализации с помощью `MOV`, а `PUSH` более характерен

для ассемблерных извращений, которые встречаются, например, в защитах, имеющих задачу сбить хакеров с толку. Ну, если такой прием и смутит хакера, то только начинающего.

### Размещение массивов и структур

Массивы и структуры размещаются в стеке последовательно в смежных ячейках памяти, при этом меньший индекс массива (элемент структуры) лежит по меньшему адресу, но — внимание — адресуется большим модулем смещения относительно регистра указателя кадра стека. Это не покажется удивительным, если вспомнить, что локальные переменные адресуются отрицательными смещениями, следовательно, `[RBP-0x4]` > `[RBP-0x10]`.

Путаницу усиливает и то обстоятельство, что, давая локальным переменным имена, IDA опускает знак минус. Поэтому из двух имен, скажем `var_4` и `var_10`, по меньшему адресу лежит то, чей индекс больше! Если `var_4` и `var_10` — это два конца массива, то с непривычки возникает непроизвольное желание поместить `var_4` в голову, а `var_10` в хвост массива, хотя на самом деле все наоборот!

### Выравнивание в стеке

В некоторых случаях элементы структуры, массива и даже просто отдельные переменные требуется располагать по кратным адресам. Но ведь значение указателя вершины заранее не определено и неизвестно компилятору. Как же он, не зная фактического значения указателя, сможет выполнить это требование? Да очень просто — возьмет и откинет младшие биты `RSP`!

Легко доказать, что, если младший бит равен нулю, число четное. Чтобы быть уверенным, что значение указателя вершины стека делится на два без остатка, достаточно лишь сбросить его младший бит. Сбросив два бита, мы получим значение, заведомо кратное четырем, три — восьми и так далее.

Сбрасывает биты в подавляющем большинстве случаев инструкция `AND`.

Например, `AND` `RSP,` `FFFFFFF0` делает `RSP` кратным шестнадцати. Как было получено это значение? Переводим `0xFFFFFFF0` в двоичный вид, получаем `11111111` `11111111` `11111111` `11110000`. Видишь четыре нуля на конце? Значит, четыре младших бита любого числа будут маскированы и оно разделится без остатка на 24= 16.

Хотя с локальными переменными мы уже неоднократно встречались при изучении прошлых примеров, не помешает сделать это еще один раз:

```cpp
local_vars_identified
#include <stdio.h>
#include <stdlib.h>
int MyFunc(int a, int b)
{
    int c; // Локальная переменная типа int
    char x[50]; // Массив (демонстрирует схему размещения массивов в памяти)
    c = a + b; // Заносим в c сумму аргументов a и b
    _itoa_s(c, &x[0], sizeof(x), 0x10); // Переводим сумму a и b в строку
    printf("%x == %s == ", c, &x[0]); // Выводим строку на экран
    return c;
}

int main()
{
    // Объявляем локальные переменные a и b для того, чтобы
    // продемонстрировать механизм их инициализации компилятором.
    // Такие извращения понадобились для того, чтобы запретить
    // оптимизирующему компилятору помещать локальную переменную
    // в регистр (см. «Идентификация регистровых переменных») int a = 0x666;
    int b = 0x777;
    int c[1];
    // Так как функции printf передается указатель на с,
    // а указатель на регистр быть передан не может,
    // компилятор вынужден оставить переменную в памяти c[0] = MyFunc(a, b);
    printf("%x\n", &c[0]);
    return 0;
}
```

Результат выполнения программы local_vars_identified показан на картинке ниже.

| ![Обрати внимание, так как третье значение в выводе — адрес, при каждом запуске приложения его значение будет меняться](./images/17/cons_app_addr.png)
|:--:|
| *Обрати внимание, так как третье значение в выводе — адрес, при каждом запуске приложения его значение будет меняться* |

Результат компиляции компилятором Microsoft Visual C++ 2019 в режиме реального времени с отключенной оптимизацией должен выглядеть так:

```dump
int MyFunc(int, int) proc near
Value   = dword ptr -58h
DstBuf  = byte ptr -50h
var_18  = qword ptr -18h
```

Локальные переменные располагаются по отрицательному смещению относительно `RBP`, а аргументы функции — по положительному. Выше мы обсуждали, что большинство компиляторов (в том числе Visual C++) оптимизируют код, освобождая регистр `RBP`, являющийся указателем кадра стека, чтобы он исполнял роль дополнительного регистра общего назначения. При этом функцию `RBP` начинает выполнять `RSP` (указатель на вершину стека). Заметь также: чем выше расположена переменная, тем больше модуль ее смещения.

```asm
arg_0 = dword ptr 8
arg_8 = dword ptr 10h
; Инициализируем аргументы в памяти, загружая их значения из регистров
mov   [rsp+arg_8], edx
; О том, что это аргументы, а не нечто иное, говорит их положительное смещение относительно регистра RBP
mov   [rsp+arg_0], ecx
; Уменьшаем значение RSP на 0x78, резервируя 0x78 байт под локальные переменные
sub   rsp, 78h
mov   rax, cs: security_cookie xor rax, rsp
mov   [rsp+78h+var_18], rax
; Вновь копируем аргументы в регистры для удобства последующих операций
mov   eax, [rsp+78h+arg_8] mov ecx, [rsp+78h+arg_0]
; Складываем значения аргументов
add   ecx, eax
mov   eax, ecx
```

А вот и первая локальная переменная! На то, что это именно она и есть, указывает ее отрицательное смещение относительно регистра `RBP`. Почему отрицательное? А посмотри, как IDA определила `Value`. По моему личному мнению, было бы намного нагляднее, если бы отрицательные смещения локальных переменных подчеркивались более явно.

Копируем сумму в переменную `Value`.

```asm
mov   [rsp+78h+Value], eax
; Смещение
mov   eax, 1
imul  rax, 0
; В RAX загружаем указатель на буфер-приемник
lea   rax, [rsp+rax+78h+DstBuf]
; Передаем функции _itoa_s значение 0x10 (тип системы счисления)...
mov   r9d, 10h ; Radix
; ...затем передаем значение 0x32 — размер буфера-приемника...
mov   r8d, 32h ; '2' ; Size
; ...указатель на сам буфер-приемник...
mov   rdx, rax ; DstBuf
; ...наконец, само значение для преобразования
mov   ecx, [rsp+78h+Value] ; Value
```

Вызов `_itoa_s`: переводим содержимое `Value` в шестнадцатеричную систему счисления, возвращаем результат в строковой форме в буфере-приемнике.

```asm
call  cs:__imp__itoa_s
; Определяем смещение
mov   eax, 1
imul  rax, 0
lea   rax, [rsp+rax+78h+DstBuf]
; Подготавливаем параметры для вызова printf:
; указатель на строку
mov   r8,  rax
; первоначальное значение
mov   edx, [rsp+78h+Value]
; указатель на форматную строку
lea   rcx, _Format        ; "%x == %s == "
; вызов printf
call  printf
; Возвращаем в EAX значение локальной переменной Value
mov   eax, [rsp+78h+Value]
mov   rcx, [rsp+78h+var_18]
xor   rcx, rsp   ; StackCookie
call  __security_check_cookie
; Освобождаем память, занятую локальными переменными
add   rsp, 78h
retn
int MyFunc(int, int)  endp
main  proc near
b      = dword ptr -28h
a      = dword ptr -24h
var_20 = dword ptr -20h
var_18 = qword ptr -18h
; Резервируем 0x48 байт памяти для локальных переменных
sub   rsp, 48h
mov   rax, cs:__security_cookie
xor   rax, rsp
mov   [rsp+48h+var_18], rax
; Инициализируем локальную переменную a, присваивая ей значение 0x666
mov   [rsp+48h+a], 666h
; Инициализируем локальную переменную b, присваивая ей значение 0x777
mov   [rsp+48h+b], 777h
```

Смотри, локальные переменные расположены в памяти в обратном порядке обращения к ним! Не объявления, а именно обращения! Вообще-то порядок расположения не всегда бывает именно таким, это зависит от компилятора, поэтому полагаться на порядок никогда не стоит!

```asm
; Копируем в регистр EDX значение переменной b
mov   edx,     [rsp+48h+b]    ; b
; Копируем в регистр ECX значение переменной a
mov   ecx,     [rsp+48h+a]    ; a
; Вызываем MyFunc
call  MyFunc(int,int)
; Определяем смещение
mov   ecx,     4
imul  rcx,     0
; Присваиваем возвращенное функцией MyFunc значение переменной var_20
mov   [rsp+rcx+48h+var_20],      eax
; Снова определяем смещение
mov   eax,     4
imul  rax,     0
lea   rax,     [rsp+rax+48h+var_20]
; Передаем функции printf указатель на локальную переменную var_20...
mov   rdx,     rax
; ...передаем указатель на форматную строку
lea   rcx,     asc_140002260  ; "%x\n"
call  printf
; Возвращаем ноль
xor   eax,     eax
mov   rcx,     [rsp+48h+var_18]
xor   rcx,     rsp   ; StackCookie
call  __security_check_cookie
; Освобождаем память, занятую локальными переменными
add   rsp,     48h
retn
main  endp
```

Не очень сложно, правда? Что ж, тогда рассмотрим результат компиляции этого примера компилятором Embarcadero C++Builder 10.4 — это будет немного труднее!

```asm
MyFunc(int, int) proc near
var_50 = dword ptr -50h
var_4C = dword ptr -4Ch
var_48 = qword ptr -48h
Radix  = dword ptr -40h
var_C  = dword ptr -0Ch
var_8  = dword ptr -8
var_4  = dword ptr -4
; Открываем кадр стека
push   rbp
; Резервируем 0x70 байт под локальные переменные subrsp, 70h
; Используем RBP по его прямому назначению, загружая в него указатель на кадр стека
lea    rbp, [rsp+70h]
```

Можно предположить, что далее происходит подготовка параметров для вызова функции `itoa`. Третьим слева параметром передаем значение системы счисления.

```asm
mov r8d,   10h
; В RAX загружаем адрес первого элемента массива — строки-приемника
lea rax,   [rbp+Radix]
; Переданные в регистрах параметры размещаем в переменных в
оперативной  памяти
mov [rbp+var_4],    ecx
mov [rbp+var_8],    edx
; Возвращаем значение из памяти в регистр ECX
mov ecx,   [rbp+var_4]
; Складываем число, находящееся в регистре, со значением из переменной в памяти
add ecx,   [rbp+var_8]
; Сумму копируем в переменную var_C
mov [rbp+var_C],    ecx
mov ecx,   [rbp+var_C]
```

Здесь IDA делает неверное предположение, поскольку в `RAX` хранится не основание системы счисления, а указатель на строку — буфер приема для преобразованного значения. Помещаем этот указатель в регистр `RDX` — второй слева параметр.

```asm
movrdx, rax; Radix
```

Вызов функции `itoa` с тремя параметрами. Обрати внимание, что в программе на Visual C++ мы использовали «безопасную» версию данной функции — `_itoa_s`, принимающую четыре параметра. Дополнительный параметр представлял собой размер принимающего буфера.

```asm
call  itoa
; Подготавливаем параметры для вызова printf
; Третьим параметром выступает указатель строки для вывода
lea   r8,     [rbp+Radix]
; Вторым параметром приходится значение переменной var_C, где хранится сумма   аргументов
mov   edx,    [rbp+var_C]
; Первый параметр — форматная строка
lea   rcx,    aXS    ; "%x == %s == "
mov   [rbp+var_48],    rax
; Вызываем printf, передавая ей все подготовленные параметры
call  printf
; Возвращаем сумму аргументов в EDX
mov   edx,    [rbp+var_C]
mov   [rbp+var_4C],    eax
mov   [rbp+var_50],    edx
mov   eax,    [rbp+var_50]
; Освобождаем память, занятую локальными переменными
add   rsp,    70h
; Выталкиваем RBP из стека, восстанавливая его прежнее значение, + закрываем         кадр    стека
pop   rbp
retn
MyFunc(int, int)  endp
main  proc    near
var_24 = dword ptr -24h
var_20 = dword ptr -20h
var_1C = dword ptr -1Ch
var_18 = dword ptr -18h
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8  = dword ptr -8
var_4  = dword ptr -4
; Открываем кадр стека
push  rbp
; Выделяем 0x50 байт под локальные переменные
sub   rsp,    50h
; Копируем адрес указателя на вершину стека
lea   rbp,    [rsp+50h]
; Подготовка переменных
mov   [rbp+var_4],     0
mov   [rbp+var_8],     ecx
mov   [rbp+var_10],    rdx
mov   [rbp+var_14],    666h
mov   [rbp+var_18],    777h
```

Как видно, значения `0x666` и `0x777` из переменных копируются в регистры `ECX` и `EDX` явно для передачи в качестве параметров следующей функции.

```asm
mov ecx, [rbp+var_14]
mov edx, [rbp+var_18]
; Вызов функции
call MyFunc(int,int)
```

Проецируем указатель на переменную `var_1C` в регистр `RDX` с надеждой, что до вызова функции `printf` в эту переменную будет помещено значение, поскольку до сих пор она не инициализирована.

```asm
lea   rdx,    [rbp+var_1C]
; Вот и оно, размещение в переменной var_1C результата вызова функции MyFunc
mov   [rbp+var_1C],    eax
; Первый слева параметр для printf — форматная строка
lea   rcx,    asc_44E00D      ; "%x\n"
call  printf
lea   rcx,    aPause ; "pause"
mov   [rbp+var_20],    eax
; Вызов функции system с параметром "pause" для задержки приложения перед закрытием
call  system
mov   [rbp+var_4],     0
mov   [rbp+var_24],    eax
mov   eax,    [rbp+var_4]
; Освобождаем память, выделенную для локальных переменных
add   rsp,    50h
; Выталкиваем RBP из стека, восстанавливая его прежнее значение, + закрываем  кадр стека
pop   rbp
retn
main  endp
```

Надо признать, транслированный компилятором C++Builder код не сильно уступает результату Visual C++. Из глобальных отличий нам удалось увидеть использование регистра `RBP` по прямому назначению — в качестве указателя кадра стека, и на этом ощутимые различия заканчиваются.

### FPO — Frame Pointer Omission

Традиционно для адресации локальных переменных используется регистр `RBP`. Учитывая, что архитектура x64 включает в два раза больше регистров общего назначения, чем `x86` (16 против 8), казалось бы, одним регистром можно пренебречь. Между тем вычисления постоянно становятся все сложнее и сложнее, а требования к скорости выполнения — все выше и выше. Поэтому жертвовать даже одним драгоценным регистром, «насовсем» отдавая его локальным переменным, уж очень не хочется. Нельзя ли найти какое-нибудь другое, более элегантное решение?

Хорошенько подумав, мы придем к выводу, что отдельный регистр для адресации локальных переменных вообще не нужен. Можно (не без ухищрений, правда) обойтись одним лишь `RSP` — регистром — указателем стека. Подобный трюк мы повсеместно наблюдали в дизассемблерных листингах компилятора Visual C++.

Единственная проблема — плавающий кадр стека. Пусть после выделения памяти под локальные переменные `RSP` указывает на вершину выделенного региона. Тогда переменная `buff` окажется по адресу `RSP+0xC`. Но стоит занести что-нибудь в стек (аргумент вызываемой функции или регистр на временное хранение), как кадр «уползет» и `buff` окажется расположен уже не по адресу `RSP+0xC`, а по адресу `RSP+0x10`! Обрати внимание на следующее изображение.

| ![Адресация локальных переменных через регистр RSP приводит к образованию плавающего кадра стека](./images/17/local_vars_adr_rsp.png)
|:--:|
| *Адресация локальных переменных через регистр RSP приводит к образованию плавающего кадра стека* |

Современные компиляторы умеют адресовать локальные переменные через `RSP`, динамически отслеживая его значение (правда, при условии, что в теле функции нет хитрых ассемблерных вставок, изменяющих значение `RSP` непредсказуемым образом).

Это чрезвычайно затрудняет изучение кода, поскольку теперь невозможно, ткнув пальцем в произвольное место листинга, определить, к какой именно локальной переменной идет обращение. Приходится прочесывать всю функцию целиком, внимательно следя за значением `RSP` (и нередко совершая при этом грубые ошибки, пускающие всю работу насмарку). К счастью, дизассемблер IDA умеет обращаться с такими переменными, но хакер тем и отличается от простого смертного, что никогда всецело не полагается на автоматику, а сам стремится понять, как это работает!

Рассмотрим наш старый добрый `passCompare1`, откомпилировав его с ключом `/02` — оптимизация по скорости. Тогда компилятор будет стремиться использовать все регистры и адресовать локальные переменные через `RSP`, что нам и надо.

| ![passCompare1 — включение оптимизации по скорости](./images/17/pass_compare_on_opt_speed.png)
|:--:|
| *passCompare1 — включение оптимизации по скорости* |

Крохотную процедуру из десятка строк несложно проштудировать вручную с помощью того же HIEW, но вот на программе в миллион строк можно и лапти скинуть! Или... воспользоваться IDA. Посмотрим на результат ее работы:

```asm
main proc near   ; CODE XREF:scrt_common_main_seh+107↓p
                 ; DATA XREF: .pdata:000000014000400C↓o
Buf = byte ptr -88h
var_18 = qword ptr -18h
```

IDA обнаружила две локальные переменные. Вторая из них расположена по смещению `0x18` относительно кадра стека, оттого и названа она `var_18`.

```asm
; Использование регистра RBX
push  rbx
sub   rsp, 0A0h
mov   rax, cs:__security_cookie
xor   rax, rsp
```

IDA автоматически подставляет имя локальной переменной к ее смещению в кадре стека:

```asm
mov   [rsp+0A8h+var_18], rax
lea   rcx, _Format ; "Enter password:"
xor   ebx, ebx
call  printf
xor   ecx, ecx
call  cs:__imp___acrt_iob_func
mov   r8, rax ; File
lea   edx, [rbx+64h] ; MaxCount
lea   rcx, [rsp+0A8h+Buf] ; Buf
...
```

Обрати внимание: IDA правильно распознала обращение к нашей второй переменной.

## Идентификация регистровых и временных переменных

Стремясь минимизировать количество обращений к памяти, оптимизирующие компиляторы размещают наиболее интенсивно используемые локальные переменные в регистрах общего назначения, только по необходимости сохраняя их в стеке (а в идеальном случае не сохраняя их вовсе).

Какие трудности для анализа это создает? Во-первых, вводит контекстную зависимость в код. Так, увидев в любой точке функции команду типа `MOV` `RAX, [RBP+var_10]`, мы с уверенностью можем утверждать, что здесь в регистр `RAX` копируется содержимое переменной `var_10`. А что это за переменная? Это можно легко узнать, пройдясь по телу функции в поисках всех вхождений `var_10`, они-то и подскажут назначение переменной!

Однако с регистровыми переменными этот номер не пройдет! Положим, нам встретилась инструкция `MOV` `RAX,RSI` и мы хотим отследить все обращения к регистровой переменной `RSI`. Как быть, ведь поиск подстроки `RSI` в теле функции ничего не даст — вернее, напротив, выдаст множество ложных срабатываний? Один и тот же регистр (в нашем случае `RSI`) может использоваться (и используется) для временного хранения множества различных переменных! Поскольку регистров общего назначения 16, да к тому же `RBP` закреплен за указателем кадра стека, а `RAX` — за возвращаемым значением функции, остается 14 регистров, пригодных для хранения локальных переменных. А в программах на языке C++ и того меньше — один из этих регистров отведен под указатель на виртуальную таблицу, а другой — под указатель на экземпляр `this`. Остается 12 регистров. На первый взгляд — достаточно. Однако следует учитывать, что в типичной функции локальных переменных десятки! Вот компилятор и использует регистры как кеш, только в исключительных случаях каждая локальная переменная сидит в своем регистре, чаще всего переменные хаотично скачут по регистрам, временами сохраняются в стеке, зачастую выталкиваясь совсем в другой регистр (не в тот, в чье содержимое сохранялась).

Практически все распространенные дизассемблеры (в том числе и IDA) не в состоянии отслеживать «миграции» регистровых переменных, и эту операцию приходится выполнять «вручную». Определить содержимое интересующего регистра в произвольной точке программы достаточно просто, хотя и утомительно: достаточно прогнать программу с начала функции до этой точки на «эмуляторе процессора», работающего в «голове», отслеживая все операции пересылки. Гораздо сложнее выяснить, сколько локальных переменных хранится в таком-то регистре. Когда большое количество переменных отображается на небольшое число регистров, однозначно восстановить отображение становится невозможно. Вот, например, программист объявляет переменную `а` — компилятор помещает ее в регистр `X`. Затем некоторое время спустя программист объявляет переменную `b`, и, если переменная `а` больше не используется (что бывает довольно часто), компилятор может поместить в тот же самый регистр `X` переменную `b`, не заботясь о сохранении значения `а` (а зачем его сохранять, если оно не нужно). В результате мы теряем одну переменную. На первый взгляд здесь нет никаких проблем. Теряем, ну и ладно! Теоретически это мог сделать и сам программист. Спрашивается: зачем он вводил `b`, когда для работы вполне достаточно одной `а`? Если переменные `а` и `b` имеют один тип, то никаких проблем действительно не возникает, но в противном случае анализ программы будет чрезвычайно затруднен.

Перейдем к технике идентификации регистровых переменных. Во многих хакерских руководствах утверждается, что регистровая переменная отличается от остальных тем, что никогда не обращается к памяти вообще. Это неверно. Регистровые переменные могут временно сохраняться в стеке командой `PUSH` и восстанавливаться обратно командой `POP`. Конечно, в некотором, «высшем смысле» такая переменная перестает быть регистровой, но и не становится стековой. Чтобы не дробить типы переменных на множество классов, условимся считать, что (как утверждают другие хакерские руководства) регистровая переменная — это переменная, которая содержится в регистре общего назначения, возможно, сохраняется в стеке, но всегда на вершине, а не в кадре стека.

Другими словами, регистровые переменные никогда не адресуются через `RBP`. Если переменная адресуется через `RBP`, следовательно, она «прописана» в кадре стека и является стековой переменной. Правильно? Нет! Посмотри, что произойдет, если регистровой переменной `а` присвоить значение стековой переменной `b`. Компилятор сгенерирует приблизительно следующий код: `MOV` `REG,` `[RBP-xxx]`. Соответственно, присвоение стековой переменной значения регистровой будет выглядеть так: `MOV` `[RBP-xxx],` `REG`. Но, несмотря на явное обращение к кадру стека, переменная REG все же остается регистровой переменной. Рассмотрим следующий код:

```asm
...
MOV [RBP-0x4], 0x666
MOV RSI, [RBP-0x4]
MOV [RBP-0x8], RSI
MOV RSI, 0x777
SUB RSI, [RBP-0x8]
MOV [RBP-0xC], RSI
...
```

Его можно интерпретировать двояко — если действительно существует некая регистровая переменная `RSI`, исходный тест примера должен выглядеть так:

```asm
int var_4 = 0x666;
int var_8 = var_4;
int vac_C = 0x777 - var_8;
```

А может, регистр `RSI` используется как временная переменная для пересылки данных, тогда исходный текст примера должен выглядеть так:

```asm
int var_4 = 0x666; register
int RSI = var_4;
int var_8 = RSI;
RSI = 0x777 var_8;
int var_C = RSI;
```

> **Немного о регистрах**
>
> В языках C/C++ существует ключевое слово , предназначенное для принудительного размещения переменных в регистрах. И все было бы хорошо, только подавляющее большинство компиляторов втихомолку игно-рируют предписания программистов, размещая переменные там, где, по мнению компилятора, им будет «удобно». Разработчики компиляторов объясняют это тем, что компилятор лучше знает, как построить наиболее эффективный код. Не надо, говорят они, пытаться помогать ему.
> Напрашивается следующая аналогия: пассажир говорит, что ему надо в аэропорт, а таксист молча едет, «куда удобнее ». Ну не должна работа с ком-пилятором превращаться в войну с ним, никак не должна! Отказ разместить переменную в регистре вполне логичен, но в таком случае нужно прекращать компиляцию и выдать сообщение об ошибке, типа «убери , а то ком-пилить не буду!», или на худой конец — вывести предупреждение.

При том что алгоритм обоих листингов идентичен, первый из них заметно выигрывает в наглядности у второго. А главная цель дизассемблирования отнюдь не воспроизведение подлинного исходного текста программы, а реконструирование ее алгоритма. Совершенно безразлично, что представляет собой `RSI` — регистровую или временную переменную. Главное, чтобы костюмчик сидел. Грубо говоря, из нескольких вариантов интерпретации выбирай самый наглядный!

Вот мы и подошли к понятию временных переменных. Но, прежде чем заняться его изучением вплотную, завершим изучение регистровых переменных исследованием следующего примера, демонстрирующего идентификацию регистровых переменных:

```cpp

#include <stdio.h>
#include <stdlib.h>

int _tmain(int argc, _TCHAR* argv[])
{
    int a = 0x666;
    int b = 0x777;
    int c;

    c = a + b;

    printf("%x + %x = %x\n", a, b, c);
    c = b - a;

    printf("%x %x = %x\n", b, a, c);
    system("pause");

    return 0;
}
```

Откомпилируем этот код с помощью Embarcadero C++Builder с установленным флагом компиляции `-O2` (оптимизация по скорости).

| ![Настройка параметров оптимизации](./images/17/settings_param_opt.png)
|:--:|
| *Настройка параметров оптимизации* |

В итоге дизассемблированный листинг должен выглядеть приблизительно так:

```asm
; int __cdecl main(int argc, const char **argv, const char **envp)
public main
main proc near ; DATA XREF: __acrtused+29↑o
```

Обрати внимание, что IDA не распознала ни одной стековой переменной, хотя они объявлялись в программе. Выходит, компилятор разместил их в регистрах. Если бы мы откомпилировали этот пример с отключенной оптимизацией, в указанном месте было бы объявлено девять переменных. Можешь проверить это сам.

Сохраняем регистр в стеке или выделяем память для стековой переменной. Поскольку IDA не обнаружила ни одной стековой переменной, вероятнее всего, этот код сохраняет регистр.

```asm
push rbp
; Выделяем память в стеке
sub rsp, 20h
; Очевидно, инициализируем кадр стека
lea rbp, [rsp+20h]
```

Следующей командой в регистр `RCX` загружаем указатель на переменную `aXXX`. Дважды щелкаем по ней, переходим к ее определению... И видим строку: `'%x` `+` `%x` `=` `%x'`. Это явно строка формата. При вызове функции `printf` она всегда помещается в самый первый регистр для передачи. Следовательно, можно сделать вывод, что происходит подготовка параметров для вызова `printf`.

```asm
lea rcx, aXXX ; "%x + %x = %x\n"
```

Смотри, инициализируется регистр! Сравни это с примером `local_vars_identified`, приведенным в разделе «Идентификация локальных стековых переменных». Помнишь, там было `mov` `[rsp+48h+a],` `666h`?

```asm
mov edx, 666h
```

Следовательно, можно заподозрить, что `EDX` (являющийся 32-битной частью `RDX`) — это регистровая переменная. По существу, как нам известно, в регистре `RDX` передается второй параметр. Значит, наш предыдущий вывод о передаче параметров функции подтверждается. Таким образом, значение `0x666` передавалось через переменную. Реконструируя исходный текст, пишем: `int` `a` `=` `0x666`.

Аналогично `R8D` (32-битная часть `R8`, в котором передается третий параметр), скорее всего, представляет собой регистровую переменную: `int` `b` `=` `0x777`.

```asm
mov r8d, 777h
```

Судя по форматной строке, в последнем параметре передается сумма первых двух. Однако, как мы видим, компилятор опустил операцию сложения, выполнив ее в момент компиляции. В таком коде, где оба слагаемых «зашиты» в программу, сложение можно безболезненно выполнить на этапе компиляции, поскольку однозначно известно, что значения не будут изменены.

Зная, что `R9D` (32-битная часть `R9`, предназначенная для четвертого параметра) — регистровая переменная, просто помещаем в нее результат.

```asm
mov r9d, 0DDDh
```

Все четыре параметра находятся в целевых регистрах, можно восстановить прототип `printf`:

```cpp
printf("%x + %x = %x\n", a, b, a+b)
```

А теперь можно вызвать функцию:

```asm
call printf
```

В следующей части программы мы действуем похожим образом. Единственное, на что стоит обратить внимание: в отличие от предыдущей части листинга, в форматной строке стоит минус вместо плюса. Соответственно, четвертый параметр — разность второго и третьего. В таком порядке параметры и помещаются в регистры: сначала форматная строка, затем уменьшаемое и вычитаемое, последний параметр — разность.

```asm
lea rcx, aXXX_0 ; "%x %x = %x\n"
mov edx, 777h
mov r8d, 666h
mov r9d, 111h
call printf
```

Таким образом, исходный код для данного блока программы выглядит следующим образом:

```cpp
int a = 0x777;
int b = 0x666;
printf("%x + %x = %x", a, b, a - b);
```

```asm
lea   rcx, aPause    ; "pause"
; Вызов system("pause") для задержки программы перед закрытием
call  system
; Возвращение в EAX нулевого значения: "return 0;"
xor   eax, eax
; Очистка стека
add   rsp, 20h
; Восстановление регистра
pop   rbp
retn
main  endp
```

В итоге реконструированный исходный код выглядит так:

```cpp
int a = 0x666;
int b = 0x777;
printf("%x + %x = %x\n", a, b, a + b);
printf("%x + %x = %x\n", b, a, b a);
```

Сравнивая свой результат с оригинальным исходным текстом, с некоторой досадой обнаруживаем, что все-таки слегка ошиблись, не сохранив промежуточный результат выражения в переменную `c`. Однако эта ошибка отнюдь не загубила нашу работу, напротив, придала листингу более «причесанный» вид, облегчив его восприятие. Впрочем, о вкусах не спорят, и, если ты желаешь точнее следовать ассемблерному коду, что ж, твоя воля, вводи еще и переменную `c`.

### Временные переменные

Временными переменными мы будем называть локальные переменные, внедряемые в код программы самим компилятором. Для чего они нужны? Рассмотрим следующий пример: `int` `b` `=` `a`. Если `а` и `b` — стековые переменные, то непосредственное присвоение невозможно, поскольку в микропроцессорах серии 80x86 (к которой относится используемое ныне расширение x86_64) отсутствует адресация «память — память». Вот и приходится выполнять эту операцию в два этапа: «память → регистр» + «регистр → память». Поэтому, если скормить компилятору следующий код:

```cpp
int main()
{
    int a = 0x13;
    register int tmp = a;
    int b = tmp;
}
```

где `tmp` и есть временная переменная, создаваемая лишь на время выполнения операции b = a, а затем уничтожаемая за ненадобностью, с отключенной оптимизацией Visual C++ сгенерирует такой код:

```asm
...
mov [rsp+18h+var_18], 13h
mov eax,  [rsp+18h+var_18]
mov [rsp+18h+var_14], eax
mov eax, [rsp+18h+var_14]
mov [rsp+18h+var_10], eax
...
```

Компиляторы всегда стремятся размещать временные переменные в регистрах и только в крайних случаях заталкивают их в стек. Механизмы выделения памяти и способы чтения/записи временных переменных довольно разнообразны.

Сохранение переменных в стеке — обычная реакция компилятора на острый недостаток регистров. Целочисленные переменные чаще всего закидываются на вершину стека командой `PUSH`, а стягиваются оттуда командой `POP`. Встретив в тексте программы «тянитолкая» (инструкцию `PUSH` в паре с соответствующей ей `POP`), сохраняющего содержимое инициализированного регистра, но не стековый аргумент функции (см. «Идентификация аргументов функции»), можно достаточно уверенно утверждать, что мы имеем дело с целочисленной временной переменной.

В отличие от 32-разрядного процессора, в котором выделение памяти под вещественные переменные и их инициализация в большинстве случаев происходила раздельно, в 64-битном процессоре, благодаря наличию в нем регистров `XMM`* и набора команд `SSE`*, работа с вещественными числами практически не имеет разницы с выполнением целочисленных вычислений. Причина, по которой в отсутствии `SSE` приходилось сначала выделять память под вещественные значения, а только потом их инициализировать, заключалась в том, что команды, позволяющей перебрасывать числа с вершины стека сопроцессора на вершину стека основного процессора, не существовало, и эту операцию приходилось выполнять вручную.

Продвинутые компиляторы (например, Microsoft Visual C++) умеют располагать временные переменные в аргументах, оставшихся на вершине стека после завершения последней вызванной функции.

Временные переменные размером свыше 8 байт (строки, массивы, структуры, объекты) практически всегда размешаются в стеке, заметно выделяясь среди прочих типов своим механизмом инициализации — вместо традиционного `MOV` здесь используется одна из команд циклической пересылки `MOVS`*, при необходимости предваренная префиксом повторения `REP`. Так делают компиляторы Microsoft Visual C++, Embarcadero C++Builder. Тот же код с заменой целочисленной переменной на вещественную выглядит следующим образом:

```asm
...
movsd xmm0, cs: real@402b54fdf3b645a2
movsd [rsp+28h+var_28], xmm0
movsd xmm0, [rsp+28h+var_28]
movsd [rsp+28h+var_20], xmm0
movsd xmm0, [rsp+28h+var_20]
movsd [rsp+28h+var_18], xmm0
...
```

Механизм выделения памяти под временные переменные практически идентичен механизму выделения памяти стековым локальным переменным, однако никаких проблем идентификации не возникает. Во-первых, выделение памяти стековым переменным происходит сразу же после открытия кадра стека, а временным переменным — в любой точке функции. Во-вторых, временные переменные адресуются не через регистр — указатель кадра стека, а через указатель вершины стека. Однако таким же образом могут поступать оптимизирующие компиляторы со стековыми переменными, в чем мы не раз убеждались.

| ![Основной механизм манипуляции с временными переменными](./images/17/mem_loc_vars_manip.png)
|:--:|
| *Основной механизм манипуляции с временными переменными* |

В каких же случаях компилятор создает временные переменные? Вообще-то это зависит от «нрава» самого компилятора (чужая душа всегда потемки, а уж тем более душа компилятора). Однако можно выделить по крайней мере два случая, когда без создания временных переменных никак не обойтись:

1. При операциях присвоения, сложения, умножения.
2. В тех случаях, когда аргумент функции или член выражения — другая функция.

Рассмотрим оба случая подробнее.

Создание временных переменных при пересылках данных и вычислении выражений. Как уже отмечалось, микропроцессоры серии 80x86 не поддерживают непосредственную пересылку данных из памяти в память, поэтому присвоение одной переменной значения другой требует ввода временной регистровой переменной (при условии, что остальные переменные не регистровые).

Вычисление выражений (особенно сложных) также требует временных переменных для хранения промежуточных результатов. Вот, например, сколько, по-твоему, требуется временных переменных для вычисления следующего выражения?

```cpp
int a = 0x1;
int b = 0x2;
int c = 1 / ((1-a) / (1-b));
```

Начнем со скобок, переписав их таким образом:

```cpp
int tmp_d = 1;
tmp_d = tmp_d a;
```

и

```cpp
int tmp_e = 1;
tmp_e = tmp_e b;
```

затем

```cpp
int tmp_f = tmp_d / tmp_e;
```

и наконец

```cpp
int tmp_j = 1;
c = tmp_j / tmp_f;
```

Итого считаем... раз, два, три, четыре, ага, четыре временные переменные. Не слишком ли много? Давай попробуем записать это покороче:

```cpp
int tmp_d = 1;
tmp_d = tmp_d a; // (1 a);
int tmp_e = 1;
tmp_e = tmp_e b; // (1 b);
tmp_d = tmp_d / tmp_e; // (1 a) / (1 b);
tmp_e = 1;
tmp_e = tmp_e / tmp_d;
```

Как мы видим, вполне можно обойтись всего двумя временными переменными. Совсем другое дело! А что, если бы выражение было чуточку посложнее? Скажем, присутствовало бы десять пар скобок вместо трех. Сколько бы тогда потребовалось временных переменных? Нет, не соблазняйся искушением сразу же заглянуть в ответ, попробуй сосчитать сам! Уже сосчитал? Да что там считать, каким бы сложным выражение ни было, для его вычисления вполне достаточно всего двух временных переменных. А если раскрыть скобки, то можно ограничиться и одной, однако это потребует излишних вычислений.

А теперь посмотрим, какой код сгенерировал Visual C++ с отключенной оптимизацией (пример `time_expression`):

```asm
main proc near
; Объявление локальных переменных
var_18 = dword ptr -18h
var_14 = dword ptr -14h
var_10 = dword ptr -10h
var_C  = dword ptr -0Ch
; Смотри, в исходном коде мы объявили три локальные переменные, а
здесь  их   четыре
; Выделение памяти под стек
sub    rsp, 18h
; Инициализация локальных переменных
mov    [rsp+18h+var_18],    1
mov    [rsp+18h+var_14],    2
; Вот вводится первая временная переменная
mov    eax, 1
```

В нее записывается непосредственное значение, так как команда вычитания `SUB`, в силу архитектурных особенностей микропроцессоров серии 80x86, всегда записывает результат вычисления на место уменьшаемого. Поэтому уменьшаемое не может быть непосредственным значением, вот и приходится вводить временную переменную.

```asm
sub eax, [rsp+18h+var_18]
```

В регистре `EAX` теперь хранится вычисленное значение (`1-a`).

Вводится еще одна временная переменная, поскольку `EAX` трогать нельзя — он занят.

```asm
mov ecx, 1
sub ecx, [rsp+18h+var_14]
```

В регистре `ECX` теперь хранится вычисленное значение (`1-b`).

Преобразуем двойное слово, лежащее в `EAX`, в четверное слово, помещаемое в `EDX:EAX`. Машинная команда `idiv` всегда ожидает увидеть делимое именно в этих регистрах.

```asm
cdq
```

Делим (`1-a`) на (`1-b`), помещая частное в `EAX`. Прежнее значение временной переменной при этом неизбежно затирается, однако для дальнейших вычислений оно и не нужно. Вот и пускай себе затирается — не беда!

```asm
idiv  ecx
; Копируем значение (1-a)/(1-b) в локальную переменную var_10
mov   [rsp+18h+var_10],    eax
; Заносим в EAX непосредственное значение 1. Это еще одна временная переменная
mov   eax,                 1
; Обнуляем EDX
cdq
; Копируем результат (1-a)/(1-b) из переменной var_10 в ECX. Четвертая  временная            переменная
mov   ecx,                 [rsp+18h+var_10]
; Делим 1 на (1-a)/(1-b). Частное помещается в EAX
idiv  ecx
; Копируем результат в переменную var_C
mov   [rsp+18h+var_C],     eax
; Возвращаем ноль
xor   eax,                 eax
; Восстанавливаем стек
add   rsp,                 18h
retn
main  endp
```

Итак, для вычисления данного выражения потребовалось четыре временные переменные и всего два регистра общего назначения.

### Создание временных переменных для сохранения значения, возвращенного функцией, и результатов вычисления выражений

Большинство языков высокого уровня (в том числе и C/C++) допускают подстановку функций и выражений в качестве непосредственных аргументов. Например, `myfunc(a+b,` `myfunc_2(c))`. Прежде чем вызвать `myfunc`, компилятор должен вычислить значение выражения `a+b`. Это легко, но возникает вопрос: во что записать результат сложения?

#### Исходник time_expression2:

```cpp
int myfunc(int a, int b)
{
    return a + b;
}

int myfunc_2(int c)
{
    return c + 1;
}

int main()
{
    int a = 1;
    int b = 2;
    int c = 3;
    int d = myfunc(a + b, myfunc_2(c));
}
```

| ![Выполнение time_expression2 под дебаггером](./images/17/time_expression2_run_debug.png)
|:--:|
| *Выполнение time_expression2 под дебаггером* |

Посмотрим, как с этим справится компилятор Visual C++:

```asm
main `proc` `near`    ; CODE XREF: scrt_common_main_seh+107↓p
                      ; DATA XREF: .pdata:ExceptionDir↓o
; Объявляем локальные переменные c `=` dword `ptr` `-`18h
var_14 = dword ptr -14h
var_10 = dword ptr -10h
var_C  = dword ptr -0Ch
; Инициализируем стек
sub   rsp, 38h
; Определяем значения локальных переменных
mov   [rsp+38h+var_10], 1
mov   [rsp+38h+var_14], 2
mov   [rsp+38h+c], 3
```

Создается временная регистровая переменная `tEAX`, и в нее копируется значение локальной переменной `c` для передачи функции `myfunc_2` в качестве параметра.

```asm
mov ecx, [rsp+38h+c] ; c
```

Функция `myfunc_2` возвращает свое значение в регистре `EAX`. Его можно рассматривать как своего рода еще одну временную регистровую переменную.

```asm
call myfunc_2(int)
```

Копируем в `ECX` значение локальной переменной `var_14`. `ECX` — еще одна временная переменная. Обрати внимание, что `EAX` до сих пор занята.

```asm
mov  ecx, [rsp+38h+var_14]
; Создается еще одна (третья) временная регистровая переменная tEDX
mov  edx, [rsp+38h+var_10]
; Складываем var_10 и var_14, заменяя значение в tEDX
add  edx, ecx
; Копируем результат предыдущего сложения во временную регистровую переменную tECX
mov  ecx, edx ; a
; Наконец, используем результат функции myfunc_2, копируя его в tEDX
mov  edx, eax ; b
```

Параметры для передачи следующей функции подготовлены путем размещения в нужных регистрах.

```asm
call  myfunc(int,int)
; Результат myfunc копируем в локальную переменную var_C
mov   [rsp+38h+var_C], eax
; Возвращаем ноль
xor   eax,    eax
; Деинициализируем стек
add   rsp,    38h
retn
main  endp
```

### Область видимости временных переменных

Временные переменные — это в некотором роде очень локальные переменные. Область их видимости в большинстве случаев ограничена несколькими строками кода, вне контекста которых временная переменная не имеет никакого смысла. По большому счету временная переменная не имеет смысла вообще и только загромождает код. В самом деле, `myfunc(a+b)` намного короче и понятнее, чем `int` `tmp=a+b;` `myfunc(tmp)`. Поэтому, чтобы не засорять дизассемблерный листинг, стремись не употреблять в комментариях временные переменные, а подставляй вместо них их фактические значения. Сами же временные переменные разумно предварять каким-нибудь характерным префиксом, например `tmp_` (или `t`, если ты патологический любитель краткости). Например:

```asm
MOV EAX, [RBP+var_4] ; // var_8 = var_4
             ; ^ tEAX = var_4
ADD EAX, [RBP+var_8] ;
               ^ tEAX += var_8
MOV ECX, EAX; // MyFunc(var_4+var_8)
CALL MyFunc
```

## Идентификация глобальных переменных

Программа, нашпигованная глобальными переменными, едва ли не самое страшное проклятие хакеров: нет древа строгой иерархии, компоненты программы тесно переплетаются друг с другом, и, чтобы понять алгоритм одного из них, приходится прочесывать весь листинг в поисках перекрестных ссылок. А в совершенстве восстанавливать перекрестные ссылки не умеет ни один дизассемблер, даже IDA!

Идентифицировать глобальные переменные очень просто, гораздо проще, чем все остальные конструкции языков высокого уровня. Глобальные переменные сразу же выдают себя непосредственной адресацией памяти, то есть обращение к ним выглядит приблизительно так: `MOV` `RAX,` `[401066]`, где `0x401066` и есть адрес глобальной переменной.

Сложнее понять, для чего эта переменная, собственно, нужна и каково ее содержимое на данный момент. В отличие от локальных переменных, глобальные переменные контекстно зависимы. В самом деле, каждая локальная переменная инициализируется «своей» функцией и не зависит от того, какие функции были вызваны до нее. Напротив, глобальные переменные может модифицировать кто угодно и когда угодно, значение глобальной переменной в произвольной точке программы не определено. Чтобы его выяснить, необходимо проанализировать все манипулирующие с ней функции и — более того — восстановить порядок их вызова. Подробнее этот вопрос будет рассмотрен в разделе «Построение дерева вызовов», пока же разберемся с техникой восстановления перекрестных ссылок.

### Техника восстановления перекрестных ссылок

В большинстве случаев с восстановлением перекрестных ссылок прекрасно справляется автоматический анализатор IDA, и делать это вручную практически не приходится. Однако иногда IDA ошибается, да и не всегда (и не у всех!) она бывает под рукой. Поэтому совсем нелишне уметь справляться с глобальными переменными самому.

#### Отслеживание обращений к глобальным переменным контекстным поиском их смещения в сегменте кода данных

Непосредственная адресация глобальных переменных чрезвычайно облегчает поиск манипулирующих с ними машинных команд. Рассмотрим, например, такую конструкцию: `MOV` `EAX,[0x41B904]`. После ассемблирования она будет выглядеть так: `A1` `04` `B9` `41` `00`. Смещение глобальной переменной записывается, «как есть» (естественно, с соблюдением обратного порядка следования байтов — старшие располагаются по большему адресу, а младшие — по меньшему).

Тривиальный контекстный поиск позволит выявить все обращения к интересующей нас глобальной переменной, достаточно лишь узнать ее смещение, переписать его справа налево и... вместе с полезной информацией получить какое-то количество «мусора». Ведь не каждое число, совпадающее по значению со смещением глобальной переменной, обязано быть указателем на эту переменную. К тому же `04` `B9` `41` `00` удовлетворяет, например, следующий контекст:

```dump
83EC04 sub esp, 004
B941000000 mov ecx, 000000041
```

Ошибка очевидна — искомое значение не является операндом инструкции, более того, оно «захватило» сразу две инструкции! Отбросив все вхождения, пересекающие границы инструкции, мы сразу же избавляемся от значительной части «мусора». Единственная проблема: как определить границы инструкций, по части инструкции о ней самой сказать ничего нельзя.

Вот, например, встречается нам следующее: `...8D` `81` `<b>04` `B9` `41` `00`

`b>` `00...` Эту последовательность, за вычетом последнего нуля, можно интерпретировать так: `lea` `eax,[ecx+0x41B904]`. Но если предположить, что `0x8D` принадлежит «хвосту» предыдущей команды, то получится следующее: `add` `d,[ecx][edi]*4,000000041`, а может быть, здесь и вовсе несколько команд...

Самый надежный способ определения границ машинных команд — трассированное дизассемблирование, но, к сожалению, это чрезвычайно ресурсоемкая операция, и далеко не всякий дизассемблер умеет трассировать код. Поэтому приходится идти другим путем...

Образно машинный код можно представить в виде машинописного текста, напечатанного без пробелов. Если попробовать читать с произвольной позиции, мы, скорее всего, попадем на середину слова и ничего не поймем. Может быть, волей случая первые несколько слогов и сложатся в осмысленное слово (а то и в два!), но дальше пойдет сплошная чепуха. Например: мамылараму. Ага, «мамы» — множественное число от «мама», подходит? Подходит. Дальше — лараму. «Лараму» — это что, народный индийский герой такой со множеством родительниц? Скажешь тоже, вот ерунда какая!

Смещаемся на одну букву вперед, оставляя «м» предыдущему слову. «А» — что ж, вполне возможно, это и есть союз «А», тем более что за ним идет осмысленное местоимение «мы», получается «А мы Лараму» или «А мы Лара Му». Кто такой этот Лараму?

Сдвигаемся еще на одну букву и читаем «мыла», а за ним «раму». Заработало! А «ма», стало быть, хвост от «мама».

Вот примерно так читается и машинный код, причем эта аналогия весьма адекватна. Слово (русское) не может начинаться с некоторых букв (например, с ы, мягкого и твердого знаков), существуют характерные суффиксы и окончания с сочетанием букв, практически не встречающихся в других частях предложения. Соответственно, видя в конце несколько идущих подряд нулей, можно с высокой степенью уверенности утверждать, что это непосредственное значение, а непосредственные значения располагаются в конце.

#### Отличия констант от указателей, или продолжаем разгребать мусор

Вот наконец мы избавились от ложных срабатываний, бессмысленность которых очевидна с первого взгляда. Куча мусора заметно приуменьшилась, но... в ней все еще продолжают встречаться такие штучки, как `PUSH` `0x401010`. Что такое `0x401010` — константа или смещение? С равным успехом может быть и то и другое. Пока не доберемся до манипулирующего с ней кода, мы вообще не сможем сказать ничего вразумительного. Если манипулирующий код обращается к `0x401010` по значению — это константа (выражающая, например, скорость улепетывания Пятачка от Слонопотама), а если по ссылке — это указатель (в данном контексте смещение).

Подробнее эту проблему мы еще обсудим в разделе «Идентификация констант и смещений», пока же заметим, что минимальный адрес загрузки файла в Windows NT равен `0x10000`.

### Косвенная адресация глобальных переменных

Довольно часто приходится слышать утверждение, что глобальные переменные всегда адресуются непосредственно (исключая, конечно, ассемблерные вставки, на ассемблере программист может обращаться к переменным, как захочет). На самом же деле все далеко не так... Если глобальная переменная передается функции по ссылке (а почему бы программисту не передать глобальную переменную по ссылке?), она будет адресоваться косвенно — через указатель.

Автору могут возразить: а зачем вообще явно передавать глобальную переменную функции? Любая функция и без этого может к ней обратиться. Автор не спорит. Да, может, но только если знает об этом заранее. Вот, скажем, есть у нас функция `xchg`, обменивающая свои аргументы местами, и есть две глобальные переменные, которые нам позарез приспичило обменять. Функции `xchg` доступны все глобальные переменные, но она не знает, какие из них необходимо обменивать (и необходимо ли это вообще), вот и приходится ей явно передавать глобальные переменные как аргументы. А это значит, что всех обращений к глобальным переменным простым контекстным поиском мы не найдем. Самое печальное — не найдет их и IDA Pro. Да и как бы она могла их найти? Для этого ей потребовался бы полноценный эмулятор процессора или хотя бы основных команд, в чем мы и убедимся в следующем примере, который демонстрирует явную передачу глобальных переменных:

```cpp
#include <stdio.h>

int a; int b; // Глобальные переменные a и b

// Функция, обменивающая значения аргументов
void xchg(int* a, int* b)
{
    int c;
    // Косвенное обращение к аргументам по указателю
    c = *a;
    *a = *b;
    *b = c;
    // Если аргументы функции — глобальные переменные,
    // то они будут адресоваться не прямо, а косвенно
}

int main()
{
    a = 0x666;
    b = 0x777; // Здесь непосредственное обращение к глобальным переменным

    printf("%x <> %x\n", a, b); // Первоначальные значения

    xchg(&a, &b); // Передача глобальных переменных по ссылке

    printf("%x <> %x", a, b); // Значения после обмена
}
```

| ![Вывод программы pass_global_vars](./images/17/pass_global_vars_con_out.png)
|:--:|
| *Вывод программы pass_global_vars* |

Результат компиляции Microsoft Visual C++ должен выглядеть так:

```asm
main proc near
; Заметь, ни одной локальной переменной
; Инициализация стека
sub rsp, 28h
; В двух последующих командах выполняется инициализация двух
глобальных переменных
mov cs:int a, 666h
mov cs:int b, 777h
```

На то, что это действительно глобальные переменные, указывает наличие сегмента перед их именами.

Следующие три инструкции размещают параметры в регистрах для передачи их функции: число1, число2, указатель на форматную строку.

```asm
mov r8d, cs:int b
mov edx, cs:int a
lea rcx, _Format ; "%x <> %x\n"
; Вывод на эмулятор терминала начальных значений
call printf
```

В регистры `RCX` и `RDX` для передачи параметров функции `xchg` помещаются адреса числовых глобальных переменных, другими словами — передаются ссылки на них. Это значит, что вызываемая функция будет обращаться к переменным косвенно, через указатели, точно так же, как она обращается с локальными переменными.

```asm
lea rdx, int b ; b
lea rcx, int a ; a
; Вызов функции обмена значениями
call xchg(int *,int *)
```

Так как результаты возвратились по тем же самым адресам, значения копируются из тех же самых переменных в регистры для передачи функции `printf`...

```asm
mov   r8d, cs:int b
mov   edx, cs:int a
; ...форматная строка
lea   rcx, aXX_0   ; "%x <> %x"
; Вывод на эмулятор терминала значений после обмена
call  printf
; Возвращение нуля
xor   eax, eax
; Деинициализация стека
add   rsp, 28h
retn
main  endp
void  xchg(int *, int *) proc near
; Переменная и два параметра
var_18 = dword ptr -18h
arg_0  = qword ptr 8
arg_8  = qword ptr 10h
; Инициализация параметров arg_0 и arg_8 значениями из регистров
mov   [rsp+arg_8], rdx
mov   [rsp+arg_0], rcx
; Выделение памяти для стека
sub   rsp, 18h
; Загрузка в RAX содержимого параметра arg_0
mov   rax, [rsp+18h+arg_0]
```

Смотри — косвенное обращение к глобальной переменной! А еще говорят, будто бы таких не бывает! Разумеется, определить, что обращение происходит именно к глобальной переменной (и какой именно глобальной переменной), можно только анализом кода вызывающей функции.

Адрес параметра `arg_0` помещается в `EAX`, заменяя собой значение `arg_0` в `RAX`. При этом, заметь, оригинальное значение `arg_0` все еще хранится в памяти.

```asm
mov  eax, [rax]
; Копируем адрес arg_0 в локальную переменную var_18 (единственную локальную переменную)
mov  [rsp+18h+var_18],    eax
; Загружаем в RAX содержимое параметра arg_0
mov  rax, [rsp+18h+arg_0]
; Загружаем в RCX содержимое параметра arg_8
mov  rcx, [rsp+18h+arg_8]
; Помещаем в ECX адрес arg_8, тем самым затирая сам arg_8
mov  ecx, [rcx]
; На место адреса параметра arg_0 помещаем адрес параметра arg_8
mov  [rax],               ecx
; Загружаем в RAX значение arg_8
mov  rax, [rsp+18h+arg_8]
; Загружаем в ECX содержимое локальной переменной var_18
mov  ecx, [rsp+18h+var_18]
```

По адресу `arg_8` размещаем значение локальной переменной `var_18`, а в ней находится адрес параметра `arg_0`.

```asm
mov [rax], ecx
; Очистка стека
add rsp, 18h
retn
void xchg(int *, int *) endp
```

Непростой примерчик получился, что неудивительно, с указателями всегда надо быть начеку. Одна ошибка, и все полетит в тартарары!

IDA нашла все ссылки на обе глобальные переменные:

```asm
.data:000000014000362C ; int b
.data:000000014000362C int b  dd ? ; DATA XREF: main+Ew
.data:000000014000362C ; main+18r ...
.data:0000000140003630 ; int a
.data:0000000140003630 int a  dd ? ; DATA XREF: main+4w
.data:0000000140003630 ; main+1Fr ...
```

* Первые две: `main+4w` и `main+Ew` — на код инициализации (`w`rite).
* Вторые две: `main+1Fr` и `main+18r` — на код чтения значения (`r`ead).

Если среди перекрестных ссылок на глобальную переменную присутствуют ссылки с суффиксом `о`, обозначающие взятие смещения (аналог ассемблерной директивы `offset`), то сразу же вскидывай ушки на макушку — раз присутствует `offset`, значит, имеет место передача глобальной переменной по ссылке. А ссылка — это косвенная адресация. А косвенная адресация — это ласты: утомительный «ручной» анализ и никаких чудес прогресса.

#### Статические переменные

Статические переменные — это разновидность глобальных переменных, но с ограниченной областью видимости: они доступны только из той функции, в которой и были объявлены. Во всем остальном статические и глобальные переменные полностью совпадают — обе размещаются в сегменте данных, обе непосредственно адресуются (исключая случаи обращения через ссылку), обе...

В общем, есть лишь одна существенная разница — к глобальной переменной могут обращаться любые функции, а к статической — только одна. А как насчет глобальных переменных, используемых лишь одной функцией? Да какие же это глобальные переменные?! Это не глобальность, это кривость исходного кода программы. Если переменная используется лишь одной функцией, нет никакой необходимости объявлять ее глобальной! Всякая непосредственно адресуемая ячейка памяти — глобальная (статическая) переменная, но не всякая статическая (глобальная) переменная

всегда адресуется непосредственно.

## Заключение

Целью сегодняшних упражнений с дизассемблером было научиться определять всевозможные типы переменных, узнать разнообразие их поведения, чтобы при «боевом» анализе сразу отнести найденные переменные к определенным категориям. Учитывая срок жизни переменной, места ее обитания, способы передачи, можно существенно упростить свою работу по восстановлению алгоритма ломаемой программы, выделив только те переменные, которые используются в конкретный момент ее выполнения. Выбери сам, хочется ли тебе следить за доброй сотней переменных или, разделив их на классы, следить только за малой частью из них. Ответ напрашивается сам.
