# Разбираем самодеятельность компиляторов при трансляции оператора выбора

В этой статье мы изучим оператор выбора `switch`. Давай разберемся, какие формы он может принять в двоичном коде, как разные компиляторы транслируют его и как его найти в дизассемблированном коде по характерным признакам.

## ИЩЕМ ОПЕРАТОРЫ SWITCH-CASE-BREAK В БИНАРНОМ КОДЕ

Для улучшения читабельности программ в язык C был введен оператор множественного выбора — switch. В Delphi с той же самой задачей справляется оператор CASE, более гибкий, чем его C-аналог, но об их различиях мы поговорим позднее.

Легко показать, что switch эквивалентен такой конструкции:

```c
IF (a == x1) THEN оператор1
ELSE IF (a == X2) THEN оператор2
IF (a == X2) THEN оператор2
IF (a == X2) THEN оператор2
ELSE ... оператор по умолчанию
```

Если изобразить это ветвление в виде логического дерева, то образуется характерная «косичка».

| ![Тран­сля­ция опе­рато­ра switch в общем слу­чае](./images/22/sch_switch.png)
|:--:|
| *Тран­сля­ция опе­рато­ра switch в общем слу­чае* |

Казалось бы, идентифицировать switch никакого труда не составит — даже не строя дерева, невозможно не обратить внимание на длинную цепочку гнезд, проверяющих истинность условия равенства некоторой переменной с серией непосредственных значений (сравнения переменной с другой переменной оператор switch не допускает).

Однако в реальной жизни все происходит совсем не так. Компиляторы (даже неоптимизирующие) транслируют switch в настоящий «мясной рулет», доверху нашпигованный всевозможными операциями отношений. Давай откомпилируем следующий код компилятором Microsoft Visual C++ 2022:

```c
#include <stdio.h>
int main()
{
    int a = 0x666;
    switch (a)
    {
        case 0:

            printf("a == 0");
            break;
        case 1:
            printf("a == 1");
            break;
        case 2:
            printf("a == 2");
            break;
        case 0x666:
            printf("a == 666h");
            break;
        default:
            printf("Default");
    }
}
```

| ![Вывод приложения switch_cases](./images/22/switch_cases_result.png)
|:--:|
| *Вывод приложения switch_cases* |

Теперь посмотрим в IDA на результат дизассемблирования.

| ![Дерево распустило ветки во все стороны. Можно сделать однозначный вывод: в дизассемблируемой программе присутствует оператор множественного выбора switch-case](./images/22/tree_from_ida.png)
|:--:|
| *Дерево распустило ветки во все стороны. Можно сделать однозначный вывод: в дизассемблируемой программе присутствует оператор множественного выбора switch-case* |

```asm
main proc near ; CODE XREF: scrt_common_main_seh+107↓p
; DATA XREF: .pdata:0000000140004018↓o
; Объявляем две локальные переменные,
; но почему две, если в исходном коде объявлена только одна?
var_18 = dword ptr -18h
var_14 = dword ptr -14h
; Резервируем место для локальных переменных
sub rsp, 38h
; Инициализируем локальные переменные:
; var_14 присваиваем значение 0х666, следовательно, это переменная a
mov [rsp+38h+var_14], 666h
mov eax, [rsp+38h+var_14]
```

Переменной `var_18` присваиваем это же значение. Обрати внимание: ее создает оператор `switch` для собственных нужд. Значит, мы определили, для чего в программе объявлена вторая локальная переменная! Она нужна для хранения первоначального значения. Таким образом, даже если значение сравниваемой переменной `var_14` в каком-то ответвлении `CASE` будет изменено, это не повлияет на результат выборов, поскольку значение переменной `var_18` не поменяется!

```asm
mov [rsp+38h+var_18], eax
; Сравниваем значение var_18 с нулем
cmp [rsp+38h+var_18], 0
; Если сравнение успешно, переходим в блок кода, выводящий в консоль "a == 0"
; Этот код получен трансляцией ветки case 0: printf("a == 0");
; Иначе продолжаем выполнение
jz short loc_140001115
; Сравниваем значение var_18 с 1
cmp [rsp+38h+var_18], 1
; В случае успеха прыгаем внутрь блока кода для вывода "a == 1"
; Этот код получен трансляцией ветки case 1: printf("a == 1");
; Иначе продолжаем выполнение
jz short loc_140001123
; Сравниваем значение var_18 с 2
cmp [rsp+38h+var_18], 2
; В случае равенства выводим "a == 2"
; Этот код получен трансляцией ветки case 2: printf("a == 2");
; Иначе продолжаем выполнение
jz short loc_140001131
; Сравниваем var_18 и 0x666
cmp [rsp+38h+var_18], 666h
; Если равно, выводим "a == 666h"
; Этот код получен трансляцией ветки case 0x666: printf("a == 666h");
jz short loc_14000113F
```

Если мы досюда добрались, значит, ни одно условие не сработало, поэтому выполняем дефолтное действие: делаем безусловный переход в блок кода для вывода строчки Default.

Этот код получен трансляцией ветки default: printf("Default");:

```asm
jmp short loc_14000114D
; ------------------------------------------------
loc_140001115: ; CODE XREF: main+19↑j
; printf("a == 0");
lea rcx, _Format ; "a == 0"
call printf
```

А вот этот безусловный переход, выносящий управление за пределы switch — в конец программы, есть оператор break, находящийся в конце каждой ветки. Если бы его не было, то начали бы выполняться все остальные ветки case, независимо от того, к какому значению var_18 они принадлежат!

```asm
jmp short loc_140001159 ; break
; ------------------------------------------------
loc_140001123: ; CODE XREF: main+20↑j
; printf("a == 1");
lea rcx, aA1 ; "a == 1" call printf
jmp short loc_140001159 ; break
; ------------------------------------------------
loc_140001131: ; CODE XREF: main+27↑j
; printf("a == 2");
lea rcx, aA2 ; "a == 2" call printf
jmp short loc_140001159 ; break
; ------------------------------------------------
loc_14000113F: ; CODE XREF: main+31↑j
; printf("a == 666h");
lea rcx, aA666h ; "a == 666h" call printf
jmp short loc_140001159 ; break
; ------------------------------------------------
loc_14000114D: ; CODE XREF: main+33↑j
; printf("Default");
lea rcx, aDefault ; "Default" call printf
loc_140001159: ; Конец SWITCH ; CODE XREF: main+41↑j
; main+4F↑j ...
; Возвращаем 0
xor eax, eax
; Восстанавливаем стек
add rsp, 38h retn
main endp
```

Выглядит довольно прямолинейно. Дизассемблерный листинг можно условно разделить на две части: первая часть — сам оператор выбора, откуда каждое условие передает управление во вторую часть — конкретную ветку, соответствующую этому условию.

Для сравнения взглянем, какой код построит C++Builder 10 на основе этой же программы:

```asm
public main
main proc near ; DATA XREF: acrtused+29↑o
; Как много локальных переменных!
var_38 = dword ptr -38h
var_34 = dword ptr -34h
var_30 = dword ptr -30h
var_2C = dword ptr -2Ch
var_28 = dword ptr -28h
var_24 = dword ptr -24h
var_20 = dword ptr -20h
var_1C = dword ptr -1Ch
var_18 = dword ptr -18h
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8 = dword ptr -8
var_4 = dword ptr -4
; Открываем кадр стека
push rbp
; Резервируем место для локальных переменных sub rsp, 60h
; В RBP сохраняем указатель на дно стека
lea rbp, [rsp+60h]
; Инициализируем локальные переменные:
mov [rbp+var_4], 0
mov [rbp+var_8], ecx
mov [rbp+var_10], rdx
; var_14 присваиваем значение 0х666, следовательно, это переменная a
mov [rbp+var_14], 666h
; В ECX помещаем значение var_14
mov ecx, [rbp+var_14]
; Следующим элегантным образом сравниваем значение var_14 с нулем
test ecx, ecx
```

Команда `TEST` не меняет значение операндов, поэтому присваиваем переменной `var_18` значение 0х666. Выходит, `var_18` — автоматическая переменная, созданная switch для своей работы, чтобы при изменении `var_14` внутри какой-либо ветки кода это не повлияло на дальнейший выбор пути выполнения.

```asm
mov [rbp+var_18], ecx
```

Совершаем прыжок следующей командой только в том случае, когда в результате выполнения предыдущего сравнения флаг `ZF` стал равен единице, а это могло произойти, только когда операнд равен нулю. Но в рассматриваемом примере это не так, поэтому пропускаем прыжок.

```asm
jz short loc_401454
```

Судя по месту назначения этого перехода, код получен трансляцией ветки

`case 0: printf("a == 0");`.

По идее, от следующей инструкции можно избавиться. Напрасно компилятор ее вставил, поскольку она ничего полезного не делает, а только занимает место. Посуди сам: `$` указывает на текущую позицию в программе, а команда `jmp` занимает два байта — получается, она только переводит выполнение через себя.

```asm
jmp short $+2
; ------------------------------------------------
loc_40142B: ; CODE XREF: main+29↑j
; В var_18 находится число 0х666, помещаем его в EAX
mov eax, [rbp+var_18]
; Уменьшаем EAX на единицу! Что это может значить?
; Никакого вычитания в нашей программе не было!
sub eax, 1
; Сохраняем значение разности в переменной var_1C
mov [rbp+var_1C], eax
```

Совершаем следующий переход на основе флага, если `ZF == 1`, но кто же его установил? Вспоминаем: `SUB` — та же `CMP`, только изменяющая значение операнда-приемника. Другими словами, в него помещается разность. Но в нашем случае `ZF` не равен единице, так как результат вычитания не равен нулю. Выходит, этот код получен трансляцией ветки `case 1: printf("a == 1")`, что подтверждает просмотр области, куда указывает переход.

Таким образом, данная конструкция просто завуалированная проверка `EAX` на равенство нулю! Ох и хитрый же этот C++Builder-компилятор!

```asm
jz short loc_401465
; Делаем принудительный шаг вперед, хотя можно было бы обойтись без дополнительной команды

jmp short $+2
; ------------------------------------------------
loc_401438: ; CODE XREF: main+36↑j
; В var_18 находится число 0х666, помещаем его в EAX для выполнения дальнейших вычислений
mov eax, [rbp+var_18]
```

Вычитаем из значения `var_18` 2, иными словами, исходя из нашего анализа выше, сравниваем значение переменной `var_18` с числом 2. Выходит, этот код получен трансляцией ветки `case 2: printf("a == 2")`.

```asm
sub eax, 2
mov [rbp+var_20], eax
```

Переход выполняется, если установлен флаг нуля. А он будет установлен, когда после вычитания двойки командой `SUB` в `EAX` останется ноль, то есть исходное значение `EAX` должно быть равно двум. Но этого не случилось. Поэтому продолжаем выполнение.

```asm
jz short loc_401476
jmp short $+2
; ------------------------------------------------
loc_401445: ; CODE XREF: main+43↑j
mov eax, [rbp+var_18]
```

Сравниваем неизменную `var_18` с `0x666`. Бинго! После вычитания получился ноль! Следовательно, возводим флаг `ZF`.

```asm
sub eax, 666h
mov [rbp+var_24], eax
```

Так как флаг `ZF` равен единице, совершаем переход к ветке `loc_401487`, где происходит вывод строки `"a == 666h"`. Значит, этот код получен после трансляции ветки `case 0x666: printf("a == 666h");`.

```asm
jz short loc_401487
```

Если бы предыдущий переход не был выполнен, то с помощью следующего безусловного перехода мы бы сделали прыжок на ветку `Default`.

```asm
jmp short loc_401498
; ------------------------------------------------
; Сами ветки
loc_401454: ; CODE XREF: main+27↑j
; printf("a == 0");
lea rcx, aA0 ; "a == 0"
call printf
mov [rbp+var_28], eax
jmp short loc_4014A7 ; break
; ------------------------------------------------
loc_401465: ; CODE XREF: main+34↑j
; printf("a == 1");
lea rcx, aA1 ; "a == 1"
call printf
mov [rbp+var_2C], eax
jmp short loc_4014A7 ; break
; ------------------------------------------------
loc_401476: ; CODE XREF: main+41↑j
; printf("a == 2");
lea rcx, aA2 ; "a == 2"
call printf
mov [rbp+var_30], eax
jmp short loc_4014A7 ; break
; ------------------------------------------------
loc_401487: ; CODE XREF: main+50↑j
; printf("a == 666h");
lea rcx, aA666h ; "a == 666h"
call printf
mov [rbp+var_34], eax
jmp short loc_4014A7 ; break
; ------------------------------------------------
loc_401498: ; CODE XREF: main+52↑j
; printf("Default");
lea rcx, aDefault ; "Default"
call printf
mov [rbp+var_38], eax
loc_4014A7: ; конец switch ; CODE XREF: main+63↑j
; main+74↑j ...
; Возвращаем 0
mov [rbp+var_4], 0
mov eax, [rbp+var_4]
; Восстанавливаем стек
add rsp, 60h
; Закрываем кадр стека
pop rbp retn
main endp
```

C++Builder проявил большую креативность. Сравнил переменную с нулем он при помощи инструкции `TEST`, а для всех остальных случаев вместо инструкции `CMP`, как это сделал Visual C++, C++Builder воспользовался командой вычитания `SUB`. Оптимизатор на основе своих глубинных знаний посчитал, что такой код будет быстрее выполняться!

Вот только нам, хакерам, от этого решения ничуть не легче. Ведь прямая ретрансляция кода дает конструкцию вроде

```c
if (a == 0)
    printf("a == 0");
else if (a == 1)
    printf("a == 1");
else if (a == 2)
    printf("a == 2");
else if (a == 0x664)
    printf("a == 666h");
else
    printf("Default");
```

В ней совсем не угадывается оператор `switch`! Впрочем, почему это не угадывается? Угадывается, еще как! Где есть длинная цепочка `IF — THEN — ELSE — IF — THEN — ELSE…`, там и до `switch` недалеко! Узнать оператор множественного выбора будет еще легче, если изобразить его в виде логического дерева.

Другая характерная деталь — case-обработчики, точнее оператор `break`, традиционно замыкающий каждый из них. Они-то и образуют правую половину «косички», сходясь все вместе в точке `Z`. Правда, многие программисты питают патологическую любовь к case-обработчикам размером в два-три экрана, включая в них, помимо всего прочего, и циклы, и ветвления, и даже вложенные операторы множественного выбора! В результате правая часть «косички» превращается в непроходимый таежный лес, сквозь который не проберется и стадо слонопотамов. Но даже если и так, левая часть «косички» все равно останется достаточно простой и легко распознаваемой.

В более сложных деревьях, имеющих большее количество условий — несколько десятков или сотен (как мы знаем, программисты любят нашпиговывать операторы `switch` условиями), компилятор может вставить дополнительное условие, которого не было в исходной программе. Зачем же он занимается такой самодеятельностью?

Когда в программе много условий, процессор совсем запаривается проверять их все (а по закону подлости нужный `case` будет в самом конце). Поэтому компилятор с помощью дополнительного условия разделяет дерево на два или большее количество в зависимости от добавленных условий, уменьшая его высоту. Вместо одной ветви транслятор построит две, поместив в левую часть только числа, меньшие определенного значения, а в правую — все остальные. Благодаря этому условия из конца дерева будут перенесены в его начало. Данный метод оптимизации поиска значений называют методом вилки, но не будем сейчас на нем останавливаться, а лучше разберем его в разделе «Обрезка длинных деревьев».

Изменение порядка сравнений — право компилятора. Стандарт ничего об этом не говорит, и каждая реализация вольна поступать так, как ей заблагорассудится. Другое дело — case-обработчики (то есть тот код, которому `case` передает управление в случае истинности отношения). Они обязаны располагаться так, как были объявлены в программе, поскольку при отсутствии закрывающего оператора `break` они должны выполняться строго в задуманном программистом порядке, хотя эта возможность языка C используется крайне редко.

Таким образом, идентификация оператора `switch` не сильно усложняется: если после уничтожения узлового гнезда и прививки правой ветки к левой (или наоборот) мы получаем эквивалентное дерево и это дерево образует характерную «косичку», мы имеем дело с оператором множественного выбора или его аналогом.

## ОТЛИЧИЯ SWITCH ОТ ОПЕРАТОРА CASE ЯЗЫКА PASCAL

Оператор `CASE` языка Pascal практически идентичен своему С-собрату — оператору `switch`, хотя и близнецами их не назовешь: `CASE` выгодно отличается поддержкой наборов и диапазонов значений.

Если обработку наборов можно реализовать и через `switch`, правда не так элегантно, как на Pascal (смотри следующий листинг), то проверить вхождение значения в диапазон на C получится исключительно с помощью конструкции `IF — THEN — ELSE`. Зато в Pascal каждый case-обработчик принудительно завершается неявным `break`, а С-программист волен ставить (или не ставить) его по своему усмотрению.

| ![Сравнение case со switch](./images/22/switch_c_pascal.png)
|:--:|
| *Сравнение case со switch* |

Однако оба языка накладывают жесткое ограничение на выбор сравниваемой переменной: она должна принадлежать к численному типу, а все наборы (диапазоны) значений представлять собой константы или константные выражения, вычисляемые на стадии компиляции. Подстановка переменных или вызовов функций не допускается.

Любопытно было бы посмотреть, как Pascal транслирует проверку диапазонов, и сравнить его с компиляторами языка C. Рассмотрим следующий пример:

```pas
program many_cases_d;

{$APPTYPE CONSOLE}
{$R *.res}

uses
    System.SysUtils;

var
    a : LongInt;
    begin
        a := 6;
        case a of
            2 : WriteLn('a == 2');
            4, 6 : WriteLn('a == 4 | 6 ');
            10..100 : WriteLn('a == [10,100]');
    end;
end.
```

Результат его компиляции компилятором Delphi 10.4 должен выглядеть так:

```asm
public _ZN12Many_cases_d14initializationEv
_ZN12Many_cases_d14initializationEv proc near
; DATA XREF: HEADER: 0000000000400128↑o
; .pdata:0000000000448328↓o
; unwind { //_ZN6System23_DelphiExceptionHandlerEPNS_16TExceptionRecordEyPvS2_
; Открываем кадр стека
push rbp
; Резервируем 0х20 байт в стеке
sub rsp, 20h
mov rbp, rsp
mov rax, cs:off_432120
; В RAX помещаем значение 1
mov byte ptr [rax], 1
nop
; В RCX помещаем указатель на область памяти размером в 0х12 (18) байт
lea rcx, qword_429088
; Вызываем процедуру для инициализации исполнительного модуля операционной среды
; В качестве параметра передается указатель на область памяти
call _ZN7Sysinit8_InitExeEPv
; Значение 0х6 помещается в память, все указывает на то, что это наша переменная a...
mov cs:_ZN12Many_cases_d1aE, 6
; Затем значение перемещается в регистр
mov eax, cs:_ZN12Many_cases_d1aE
; Вычитаем из EAX значение 0х2
sub eax, 2
; Проверяем на равенство нулю
test eax, eax
; Если флаг нуля возведен, делаем переход на ветку, где выводится "a== 0"
; В нашем случае продолжаем выполнение
jz short loc_428F8E
; Еще вычитаем из EAX 0x2
sub eax, 2
; Снова проверяем на равенство нулю
test eax, eax
```

Если флаг нуля возведен, делаем переход на ветку, где выводится `"a == 4 | 6 "`, это похоже на наш случай, но предыдущий `test` не возвел флаг `ZF`, так как в результате вычитания получилось число 2.

```asm
jz short loc_428FB0
; Снова такое же вычитание...
sub eax, 2
; ...и проверка
test eax, eax
```

На этот раз результат разности удовлетворяет условию перехода, поэтому совершаем прыжок на ту же самую ветку, на которую был переход выше, где выводится `"a == 4 | 6 "`!

```asm
jz short loc_428FB0
; Предположим, в параллельной вселенной выполнение продолжилось
; Тогда после нахождения очередной разности
sub eax, 4
; проверяем EAX
cmp eax, 5Bh ; '['
```

Если значение в нем больше или равно (то есть не меньше) `0х5B (91)`, значит, все условные переходы прошли лесом и мы совершаем прыжок в конец `CASE`.

```asm
jnb short loc_428FF2 ; Эта метка играет роль оператора break
```

Если предыдущий переход не сработал, значит, наше число попадает в диапазон, поэтому переходим на следующую метку, чтобы вывести соответствующую надпись.

```asm
jmp short loc_428FD2
; ------------------------------------------------
loc_428F8E: ; CODE XREF: _ZN12Many_cases_d14initializationEv+34↑j
mov rcx, cs:off_431F40
lea rdx, aA2 ; "a == 2"
call _ZN6System14_Write0UStringERNS_8TTextRecENS_13UnicodeStringE
mov rcx, rax
call _ZN6System8_WriteLnERNS_8TTextRecE
call _ZN6System8__IOTestEv
jmp short loc_428FF2 ; break
; ------------------------------------------------
loc_428FB0: ; CODE XREF: _ZN12Many_cases_d14initializationEv+3B↑j
; _ZN12Many_cases_d14initializationEv+42↑j
mov rcx, cs:off_431F40
lea rdx, aA46 ; "a == 4 | 6 "
call _ZN6System14_Write0UStringERNS_8TTextRecENS_13UnicodeStringE
mov rcx, rax
call _ZN6System8_WriteLnERNS_8TTextRecE
call _ZN6System8__IOTestEv
jmp short loc_428FF2 ; break
; ------------------------------------------------
loc_428FD2: ; CODE XREF: _ZN12Many_cases_d14initializationEv+4C↑j
mov rcx, cs:off_431F40
lea rdx, aA10100 ; "a == [10,100]"
call _ZN6System14_Write0UStringERNS_8TTextRecENS_13UnicodeStringE
mov rcx, rax
call _ZN6System8_WriteLnERNS_8TTextRecE
call _ZN6System8__IOTestEv
loc_428FF2: ; CODE XREF: _ZN12Many_cases_d14initializationEv+4A↑j
; _ZN12Many_cases_d14initializationEv+6E↑j ...
; Конец CASE
; Процедура останавливает выполнение приложения
call _ZN6System6_Halt0Ev
; ------------------------------------------------
; Если никакое исключение не возникло, делаем безусловный прыжок в обход перехватчика исключения
jmp short loc_429001
; ------------------------------------------------
db 2 dup(90h)
; ------------------------------------------------
; Если возникла исключительная ситуация, ловим ее, чтобы не дать ей вмешаться в работу операционной системы
call _ZN6System19_UnhandledExceptionEv
; ------------------------------------------------
db 90h
; ------------------------------------------------
loc_429001: ; CODE XREF: _ZN12Many_cases_d14initializationEv+B7↑j
; Конец приложения:
; восстанавливаем стек
lea rsp, [rbp+20h]
; и закрываем кадр стека
pop rbp
retn
; } // starts at 428F40
_ZN12Many_cases_d14initializationEv endp
```

Почему же компилятор, чтобы найти попадание в диапазон, в качестве верхнего предела взял не `0х64 (100)`, как прописано в исходном коде, а `0х5B ( 91)`? Дело в том, что, учитывая все предыдущие вычитания, это будет наибольшее из возможных значений.

Условные операторы, в частности операторы выбора, предоставляют компилятору широчайший творческий простор для генерации двоичного кода! К примеру, если взглянуть на исходный текст нашей программы, никакого вычитания там и в помине не было!

## ОБРЕЗКА ДЛИННЫХ ДЕРЕВЬЕВ

В некоторых (хотя и редких) случаях операторы множественного выбора содержат сотни (а то и тысячи) наборов значений, и если решать задачу сравнения в лоб, то высота логического дерева окажется гигантской до неприличия, а его прохождение займет весьма длительное время, что не лучшим образом скажется на производительности программы.

Но задумайся: чем, собственно, занимается оператор `switch`? Если отвлечься от устоявшего мнения, что `switch` дает специальный способ выбора одного из многих вариантов (который заключается в проверке совпадения значения данного выражения с одной из заданных констант и соответствующем ветвлении), то можно сказать, что `switch` — оператор «поиска соответствующего значения». В таком случае каноническое switch-дерево представляет собой тривиальный алгоритм «последовательного поиска» — самый неэффективный алгоритм из всех.

Пусть, например, исходный текст программы выглядел так:

```C
switch (a)
{
case 98:
    …;
case 4:
    …;
case 3:
    …;
case 9:
    …;
case 22:
    …;
case 0:
    …;
case 11:
    …;
case 666:
    …;
case 096:
    …;
case 777:
    …;
case 7:
    …;
}
```

Тогда соответствующее ему неоптимизированное логическое дерево будет достигать в высоту одиннадцать гнезд.

| ![Логическое дерево до утрамбовки](./images/22/log_tree_b.png)
|:--:|
| *Логическое дерево до утрамбовки* |

Причем на левой ветке корневого гнезда окажется аж десять других гнезд, а на правой вообще ни одного (только соответствующий ему case-обработчик).

Исправить «перекос» можно, разрезав одну ветку на две и привив образовавшиеся половинки к новому гнезду, которое содержит условие, определяющее, в какой из веток следует искать сравниваемую переменную. Например, левая ветка может содержать гнезда с четными значениями, а правая — с нечетными. Но это плохой критерий: четных и нечетных значений редко бывает поровну, и вновь образуется перекос.

Гораздо надежнее поступить так: берем наименьшее из всех значений и бросаем его в кучу А, затем берем наибольшее из всех значений и бросаем его в кучу B. Так повторяем до тех пор, пока не рассортируем все имеющиеся значения.

Поскольку оператор множественного выбора требует уникальности каждого значения, то есть каждое число может встречаться в наборе (диапазоне) значений лишь однажды, легко показать, что:

* в обеих кучах будет содержаться равное количество чисел (в худшем случае — в одной куче окажется на число больше);
* все числа кучи A меньше наименьшего из чисел кучи B.

Следовательно, достаточно выполнить только одно сравнение, чтобы определить, в какой из двух куч следует искать сравниваемые значения.

Высота нового дерева будет равна `((N + 1)/2) + 1`, где N — количество гнезд старого дерева. Действительно, мы же делим ветвь дерева надвое и добавляем новое гнездо — отсюда и берется `N/2`  и `+1`, а `N + 1` необходимо для округления результата деления в большую сторону. К примеру, если высота неоптимизированного дерева достигала 100 гнезд, то теперь она уменьшилась до 51.

Что? Говоришь, 51 все равно много? А что нам мешает разбить каждую из двух ветвей еще на две? Это уменьшит высоту дерева до 27 гнезд! Аналогично последующее уплотнение даст 16 → 12 → 11 → 9 → 8... и все! Более плотная упаковка дерева невозможна (подумай почему, на худой конец — построй само дерево). Но согласись, восемь гнезд — это не сто! Полное прохождение оптимизированного дерева потребует менее девяти сравнений!

«Трамбовать» логические деревья оператора множественного выбора умеют практически все компиляторы, даже неоптимизирующие. Это увеличивает производительность, но затрудняет анализ откомпилированной программы. Взгляни еще раз на рисунок «Логическое дерево до утрамбовки» — несбалансированное дерево наглядно и интуитивно понятно. После балансировки же (схема «до утрамбовки» выше) — в нем сам Тигра хвост обломит.

| ![Логическое дерево после утрамбовки](./images/22/log_tree_a.png)
|:--:|
| *Логическое дерево после утрамбовки* |

К счастью, балансировка дерева допускает эффективное обращение. Но прежде чем засучить рукава и приготовиться к лазанью по деревьям (а Тигра лазает по деревьям лучше всех!), введем понятие балансировочного узла.

**Балансировочный узел** не изменяет логики работы двоичного дерева и является факультативным узлом, единственная функция которого — укорачивание длины ветвей. Балансировочный узел без потери функциональности дерева может быть замещен любой из своих ветвей. Причем каждая ветвь балансировочного узла должна содержать одно или больше гнезд.

Рассуждая от противного: если все узлы логического дерева, правая ветка которых содержит одно или более гнезд, могут быть замещены этой самой правой веткой без потери функциональности дерева, то данная конструкция представляет собой оператор `switch`. Почему именно правая ветка? Так ведь оператор множественного выбора в развернутом состоянии представляет цепочку гнезд, соединенных левыми ветвями друг с другом, а на правых держатся case-обработчики. Вот мы и пытаемся подцепить все правые гнезда на левую ветвь. Если это удается, мы имеем дело с оператором множественного выбора, а нет — с чем-то другим.

Рассмотрим обращение балансировки на примере следующего дерева.

| ![Обращение балансировки логического дерева — шаг 1](./images/22/log_tree_balance_s1.png)
|:--:|
| *Обращение балансировки логического дерева — шаг 1* |

Двигаясь от левой нижней ветви, мы будем продолжать взбираться на дерево до тех пор, пока не встретим узел, держащий на своей правой ветви одно или более гнезд. В нашем случае это узел `(а > 5)`. Смотри: если данный узел заменить гнездами `(а == 7)` и `(а == 9)`, функциональность дерева не нарушится!

| ![Обращение балансировки логического дерева — шаг 2](./images/22/log_tree_balance_s2.png)
|:--:|
| *Обращение балансировки логического дерева — шаг 2* |

Аналогично узел `(а > 10)` может быть безболезненно заменен гнездами `(а > 96)`, `(а == 96)`, `(а == 22)` и `(а == 11)`, а узел `(а > 96)`, в свою очередь, гнездами `(а == 98)`, `(а == 666)` и `(а == 777)`. В конце концов образуется классическое switch-дерево, в котором оператор множественного выбора распознается с первого взгляда.

| ![Обращение балансировки логического дерева — шаг 3](./images/22/log_tree_balance_s3.png)
|:--:|
| *Обращение балансировки логического дерева — шаг 3* |

### Сложные случаи балансировки, или оптимизирующая балансировка

Для уменьшения высоты «утрамбовываемого» дерева хитрые трансляторы стремятся замещать гнезда балансировочными узлами. Рассмотрим следующий пример.

| ![Хитрый случай балансировки](./images/22/log_tree_tricly_case.png)
|:--:|
| *Хитрый случай балансировки* |

Для уменьшения высоты дерева транслятор разбивает его на две половины — в левую идут гнезда со значениями, меньшими единицы или равными ей, а в правую — все остальные. Казалось бы, на правой ветке узла `(а > 1)` должно висеть гнездо `(а == 2)`, ан нет! Здесь мы видим узел `(а > 2)`, к левой ветке которого прицеплен case-обработчик :2! А что, вполне логично: если `(а>1)` и `!(а>2)`, то `а == 2`!

Легко видеть, что узел `(а > 2)` жестко связан с узлом `(а > 1)` и работает на пару с последним. Нельзя выкинуть один из них, не нарушив работоспособности другого! Обратить балансировку дерева по описанному выше алгоритму без нарушения его функциональности невозможно. Отсюда может создаться мнение, что мы имеем дело вовсе не с оператором множественного выбора, а с чем-то другим.

Чтобы развеять это заблуждение, придется предпринять ряд дополнительных шагов. Первое — у switch-дерева все case-обработчики всегда находятся на правой ветви. Смотрим, можно ли трансформировать наше дерево так, чтобы case-обработчик 2 оказался на левой ветви балансировочного узла? Да, можно: изменив `(а > 2)` на `(а < 3)` и поменяв ветви местами (другими словами, выполнив инверсию). Второе — все гнезда switch-дерева содержат в себе условия равенства. Смотрим, можем ли мы заменить неравенство `(а < 3)` аналогичным ему равенством? Конечно же, можем — `(а == 2)`!

После всех этих преобразований обращение балансировки дерева удается выполнить без труда!

### Ветвления в case-обработчиках

В реальной жизни case-обработчики прямо-таки кишат ветвлениями, циклами и прочими условными переходами всех мастей. Как следствие — логическое дерево приобретает вид, ничуть не напоминающий оператор множественного выбора, а скорее смахивающий на заросли чертополоха. Понятное дело, идентифицировав case-обработчики, мы могли бы решить эту проблему, но как их идентифицировать?

Очень просто: за редкими клиническими исключениями, case-обработчики содержат ветвления относительно сравниваемой переменной. Действительно, конструкции `switch(a) .... case 666 : if (а == 666) ...` или `switch(a) ... case 666 : if (a > 66) ...` абсолютно лишены смысла. Таким образом, мы можем смело удалить из логического дерева все гнезда с условиями, касающимися сравниваемой переменной (переменной корневого гнезда).

Хорошо, а если программист в порыве собственной глупости или стремлении затруднить анализ программы впаяет в case-обработчики ветвления относительно сравниваемой переменной? Оказывается, это ничуть не затруднит анализ! Впаянные ветвления элементарно распознаются и обрезаются либо как избыточные, либо как никогда не выполняющиеся. Например, если к правой ветке гнезда `(a == 3)` прицепить гнездо `(a > 0)` — его можно удалить, как не несущее в себе никакой информации. Если же к правой ветке того же самого гнезда прицепить гнездо `(a == 2)`, его можно удалить, как никогда не выполняющееся: если `a == 3`, то заведомо `a != 2`!

## ИТОГИ

Сперва мы разобрали, чем конструкция оператора выбора `switch — case — break` отличается от обычного условного оператора `if — then — else`, как различаются их двоичные представления. Рассмотрели, какой бинарный код создают два самых популярных компилятора C++ на основе одной и той же программы, внутри которой присутствует оператор выбора. Убедились, что разница значительна.

Далее наше внимание захватил аналог оператора выбора из языка Pascal. Его рассмотрение с точки зрения программиста выявило бесспорное превосходство над C++-аналогом, однако, с точки зрения ассемблерщика, код, подготовленный Delphi-компилятором, особо не отличается от кода того же C++Builder.

Когда оператор условного выбора имеет разумное количество исполняемых веток, разобраться в условиях и вытекающих из них действий не составляет особого труда, загадка решается с использованием головного мозга без применения дополнительных инструментов. Тем не менее, если в операторе выбора содержатся десятки условий с ветками, приходится чертить логическое дерево. Как мы обсуждали в одной из прошлых статей, для построения логического дерева можно написать чертежную программу или воспользоваться графопостроителем из состава IDA Pro в том случае, если она есть среди твоего хакерского инструментария.

Но если дерево состоит из сотен или тысяч веток, а программисты с бодуна всякое могут написать, никакого мыслительного процесса для анализа такого кода не хватит. И приходится снова прибегать к помощи логических деревьев. А если дерево получится заросшим, вскарабкаться на него будет очень сложно. Поэтому его придется обрезать, сокращать, исключать не имеющие смысла узлы. Эти шаги мы подробно рассмотрели в третьем разделе статьи.

В следующей статье мы разберем устройство циклов — еще одного крайне важного механизма всех языков программирования.
